@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Main
 -- Copyright   :  (c) 2015-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -10,8 +10,8 @@
 {-# LANGUAGE ScopedTypeVariables #-}
 module Main where
 
-import qualified Data.Text.Lazy     as T
-import qualified Data.Text.Lazy.IO  as T
+import qualified Data.Text    as T
+import qualified Data.Text.IO as T
 import           System.FilePath ((</>))
 import qualified System.Directory   as Dir
 
@@ -23,6 +23,7 @@ import qualified Cryptol.ModuleSystem.Base      as M
 import qualified Cryptol.ModuleSystem.Env       as M
 import qualified Cryptol.ModuleSystem.Monad     as M
 import qualified Cryptol.ModuleSystem.NamingEnv as M
+import           Cryptol.ModuleSystem.Interface (noIfaceParams)
 
 import qualified Cryptol.Parser           as P
 import qualified Cryptol.Parser.AST       as P
@@ -35,6 +36,7 @@ import qualified Cryptol.TypeCheck     as T
 import qualified Cryptol.TypeCheck.AST as T
 
 import qualified Cryptol.Utils.Ident as I
+import           Cryptol.Utils.Logger(quietLogger)
 
 import qualified Data.SBV.Dynamic as SBV
 
@@ -46,7 +48,6 @@ main = do
   defaultMain [
     bgroup "parser" [
         parser "Prelude" "lib/Cryptol.cry"
-      , parser "Extras"  "lib/Cryptol/Extras.cry"
       , parser "PreludeWithExtras" "bench/data/PreludeWithExtras.cry"
       , parser "BigSequence" "bench/data/BigSequence.cry"
       , parser "BigSequenceHex" "bench/data/BigSequenceHex.cry"
@@ -55,7 +56,6 @@ main = do
       ]
    , bgroup "typechecker" [
         tc cd "Prelude" "lib/Cryptol.cry"
-      , tc cd "Extras"  "lib/Cryptol/Extras.cry"
       , tc cd "PreludeWithExtras" "bench/data/PreludeWithExtras.cry"
       , tc cd "BigSequence" "bench/data/BigSequence.cry"
       , tc cd "BigSequenceHex" "bench/data/BigSequenceHex.cry"
@@ -74,6 +74,13 @@ main = do
       ]
    ]
 
+-- | Evaluation options, mostly used by `trace`.
+-- Since the benchmarks likely do not use base, these don't matter very much
+evOpts :: E.EvalOpts
+evOpts = E.EvalOpts { E.evalLogger = quietLogger
+                    , E.evalPPOpts = E.defaultPPOpts
+                    }
+
 -- | Make a benchmark for parsing a Cryptol module
 parser :: String -> FilePath -> Benchmark
 parser name path =
@@ -100,7 +107,7 @@ tc cd name path =
                 }
             Right pm = P.parseModule cfg bytes
         menv <- M.initialModuleEnv
-        (Right ((prims, scm, tcEnv), menv'), _) <- M.runModuleM menv $ withLib $ do
+        (Right ((prims, scm, tcEnv), menv'), _) <- M.runModuleM (evOpts,menv) $ withLib $ do
           -- code from `loadModule` and `checkModule` in
           -- `Cryptol.ModuleSystem.Base`
           let pm' = M.addPrelude pm
@@ -114,19 +121,19 @@ tc cd name path =
           return (prims, scm, tcEnv)
         return (prims, scm, tcEnv, menv')
   in env setup $ \ ~(prims, scm, tcEnv, menv) ->
-    bench name $ nfIO $ M.runModuleM menv $ withLib $ do
+    bench name $ nfIO $ M.runModuleM (evOpts,menv) $ withLib $ do
       let act = M.TCAction { M.tcAction = T.tcModule
                            , M.tcLinter = M.moduleLinter (P.thing (P.mName scm))
                            , M.tcPrims  = prims
                            }
-      M.typecheck act scm tcEnv
+      M.typecheck act scm noIfaceParams tcEnv
 
 ceval :: String -> String -> FilePath -> T.Text -> Benchmark
 ceval cd name path expr =
   let withLib = M.withPrependedSearchPath [cd </> "lib"] in
   let setup = do
         menv <- M.initialModuleEnv
-        (Right (texpr, menv'), _) <- M.runModuleM menv $ withLib $ do
+        (Right (texpr, menv'), _) <- M.runModuleM (evOpts,menv) $ withLib $ do
           m <- M.loadModuleByPath path
           M.setFocusedModule (T.mName m)
           let Right pexpr = P.parseExpr expr
@@ -134,7 +141,7 @@ ceval cd name path expr =
           return texpr
         return (texpr, menv')
   in env setup $ \ ~(texpr, menv) ->
-    bench name $ nfIO $ E.runEval $ do
+    bench name $ nfIO $ E.runEval evOpts $ do
       env' <- E.evalDecls (S.allDeclGroups menv) mempty
       (e :: E.Value) <- E.evalExpr env' texpr
       E.forceValue e
@@ -145,7 +152,7 @@ seval cd name path expr =
   let withLib = M.withPrependedSearchPath [cd </> "lib"] in
   let setup = do
         menv <- M.initialModuleEnv
-        (Right (texpr, menv'), _) <- M.runModuleM menv $ withLib $ do
+        (Right (texpr, menv'), _) <- M.runModuleM (evOpts,menv) $ withLib $ do
           m <- M.loadModuleByPath path
           M.setFocusedModule (T.mName m)
           let Right pexpr = P.parseExpr expr
@@ -153,7 +160,7 @@ seval cd name path expr =
           return texpr
         return (texpr, menv')
   in env setup $ \ ~(texpr, menv) ->
-    bench name $ nfIO $ E.runEval $ do
+    bench name $ nfIO $ E.runEval evOpts $ do
       env' <- E.evalDecls (S.allDeclGroups menv) mempty
       (e :: S.Value) <- E.evalExpr env' texpr
       E.io $ SBV.generateSMTBenchmark False $
