@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  OptParser
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -10,6 +10,7 @@
 module OptParser where
 
 import Data.Monoid (Endo(..))
+import Data.Semigroup
 
 import Prelude ()
 import Prelude.Compat
@@ -18,14 +19,17 @@ data OptParser opt
   = OptSuccess (Endo opt)
   | OptFailure [String]
 
-instance Monoid (OptParser opt) where
-  mempty = OptSuccess mempty
-  l `mappend` r = case (l,r) of
+instance Semigroup (OptParser opt) where
+  l <> r = case (l,r) of
     (OptSuccess f,OptSuccess g) -> OptSuccess (f `mappend` g)
     (OptFailure a,OptFailure b) -> OptFailure (a `mappend` b)
     (OptFailure _,_)            -> l
     (_,OptFailure _)            -> r
 
+instance Monoid (OptParser opt) where
+  mempty = OptSuccess mempty
+  mappend = (<>)
+
 runOptParser :: opt -> OptParser opt -> Either [String] opt
 runOptParser def parse = case parse of
   OptSuccess update -> Right (appEndo update def)
