@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  REPL.Haskeline
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -12,10 +12,13 @@
 
 module REPL.Haskeline where
 
+import           Cryptol.Prims.Syntax (primTyList, primTyIdent)
 import           Cryptol.REPL.Command
 import           Cryptol.REPL.Monad
 import           Cryptol.REPL.Trie
 import           Cryptol.Utils.PP
+import           Cryptol.Utils.Logger(stdoutLogger)
+import           Cryptol.Utils.Ident(modNameToText, interactiveName, isInfixIdent)
 
 import qualified Control.Exception as X
 import           Control.Monad (guard, join)
@@ -25,6 +28,8 @@ import           Data.Char (isAlphaNum, isSpace)
 import           Data.Maybe(isJust)
 import           Data.Function (on)
 import           Data.List (isPrefixOf,nub,sortBy,sort)
+import qualified Data.Set as Set
+import qualified Data.Text as T (unpack)
 import           System.IO (stdout)
 import           System.Console.ANSI (setTitle, hSupportsANSI)
 import           System.Console.Haskeline
@@ -37,7 +42,7 @@ import           Prelude ()
 import           Prelude.Compat
 
 
--- | One REPL invocation, either form a file or from the terminal.
+-- | One REPL invocation, either from a file or from the terminal.
 crySession :: Maybe FilePath -> Bool -> REPL CommandExitCode
 crySession mbBatch stopOnError =
   do settings <- io (setHistoryFile (replSettings isBatch))
@@ -111,7 +116,7 @@ loadCryRC cryrc =
 -- | Haskeline-specific repl implementation.
 repl :: Cryptolrc -> Maybe FilePath -> Bool -> REPL () -> IO CommandExitCode
 repl cryrc mbBatch stopOnError begin =
-  runREPL (isJust mbBatch) $
+  runREPL (isJust mbBatch) stdoutLogger $
   do status <- loadCryRC cryrc
      case status of
        CommandOk -> begin >> crySession mbBatch stopOnError
@@ -213,11 +218,7 @@ cmdComp :: String -> CommandDescr -> [Completion]
 cmdComp prefix c = do
   cName <- cNames c
   guard (prefix `isPrefixOf` cName)
-  return $ Completion
-    { replacement = drop (length prefix) cName
-    , display     = cName
-    , isFinished  = True
-    }
+  return $ nameComp prefix cName
 
 -- | Dispatch to a completion function based on the kind of completion the
 -- command is expecting.
@@ -226,9 +227,11 @@ cmdArgument ct cursor@(l,_) = case ct of
   ExprArg     _ -> completeExpr cursor
   DeclsArg    _ -> (completeExpr +++ completeType) cursor
   ExprTypeArg _ -> (completeExpr +++ completeType) cursor
+  ModNameArg _  -> completeModName cursor
   FilenameArg _ -> completeFilename cursor
   ShellArg _    -> completeFilename cursor
   OptionArg _   -> completeOption cursor
+  HelpArg     _ -> completeHelp cursor
   NoArg       _ -> return (l,[])
   FileExprArg _ -> completeExpr cursor
 
@@ -259,6 +262,37 @@ completeType (l,_) = do
       vars = filter (n `isPrefixOf`) ns
   return (l,map (nameComp n) vars)
 
+-- | Complete a name for which we can show REPL help documentation.
+completeHelp :: CompletionFunc REPL
+completeHelp (l, _) = do
+  ns1 <- getExprNames
+  ns2 <- getTypeNames
+  let ns3 = primTyNames
+  let ns4 = concatMap cNames (nub (findCommand ":"))
+  let ns = Set.toAscList (Set.fromList (ns1 ++ ns2 ++ ns3)) ++ ns4
+  let n    = reverse l
+  case break isSpace n of
+    (":set", _ : n') ->
+      do let n'' = dropWhile isSpace n'
+         let vars = map optName (lookupTrie (dropWhile isSpace n') userOptions)
+         return (l, map (nameComp n'') vars)
+    _                ->
+      do let vars = filter (n `isPrefixOf`) ns
+         return (l, map (nameComp n) vars)
+
+primTyNames :: [String]
+primTyNames = map (showIdent . primTyIdent) primTyList
+  where showIdent i = show (optParens (isInfixIdent i) (pp i))
+
+-- | Complete a name from the list of loaded modules.
+completeModName :: CompletionFunc REPL
+completeModName (l, _) = do
+  ms <- getModNames
+  let ns   = map (T.unpack . modNameToText) (interactiveName : ms)
+      n    = reverse (takeWhile (not . isSpace) l)
+      vars = filter (n `isPrefixOf`) ns
+  return (l, map (nameComp n) vars)
+
 -- | Generate a completion from a prefix and a name.
 nameComp :: String -> String -> Completion
 nameComp prefix c = Completion
