@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Eval
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -17,6 +17,9 @@
 module Cryptol.Eval (
     moduleEnv
   , runEval
+  , EvalOpts(..)
+  , PPOpts(..)
+  , defaultPPOpts
   , Eval
   , EvalEnv
   , emptyEnv
@@ -41,29 +44,30 @@ import qualified Data.Sequence as Seq
 import           Data.List
 import           Data.Maybe
 import qualified Data.Map.Strict as Map
+import           Data.Semigroup
 
 import Prelude ()
 import Prelude.Compat
 
-type EvalEnv = GenEvalEnv Bool BV
+type EvalEnv = GenEvalEnv Bool BV Integer
 
 -- Expression Evaluation -------------------------------------------------------
 
 -- | Extend the given evaluation environment with all the declarations
 --   contained in the given module.
-moduleEnv :: EvalPrims b w
-          => Module         -- ^ Module containing declarations to evaluate
-          -> GenEvalEnv b w -- ^ Environment to extend
-          -> Eval (GenEvalEnv b w)
+moduleEnv :: EvalPrims b w i
+          => Module           -- ^ Module containing declarations to evaluate
+          -> GenEvalEnv b w i -- ^ Environment to extend
+          -> Eval (GenEvalEnv b w i)
 moduleEnv m env = evalDecls (mDecls m) =<< evalNewtypes (mNewtypes m) env
 
 -- | Evaluate a Cryptol expression to a value.  This evaluator is parameterized
 --   by the `EvalPrims` class, which defines the behavior of bits and words, in
 --   addition to providing implementations for all the primitives.
-evalExpr :: EvalPrims b w
-         => GenEvalEnv b w     -- ^ Evaluation environment
+evalExpr :: EvalPrims b w i
+         => GenEvalEnv b w i   -- ^ Evaluation environment
          -> Expr               -- ^ Expression to evaluate
-         -> Eval (GenValue b w)
+         -> Eval (GenValue b w i)
 evalExpr env expr = case expr of
 
   -- Try to detect when the user has directly written a finite sequence of
@@ -72,25 +76,28 @@ evalExpr env expr = case expr of
     -- NB, even if the list cannot be packed, we must use `VWord`
     -- when the element type is `Bit`.
     | isTBit tyv -> {-# SCC "evalExpr->Elist/bit" #-}
-        return $ VWord len $ return $
+        return $ VWord len $
           case tryFromBits vs of
-            Just w  -> WordVal w
-            Nothing -> BitsVal $ Seq.fromList $ map (fromVBit <$>) vs
-    | otherwise -> {-# SCC "evalExpr->EList" #-}
-        return $ VSeq len $ finiteSeqMap vs
+            Just w  -> return $ WordVal w
+            Nothing -> do xs <- mapM (delay Nothing) vs
+                          return $ BitsVal $ Seq.fromList $ map (fromVBit <$>) xs
+    | otherwise -> {-# SCC "evalExpr->EList" #-} do
+        xs <- mapM (delay Nothing) vs
+        return $ VSeq len $ finiteSeqMap xs
    where
     tyv = evalValType (envTypes env) ty
     vs  = map (evalExpr env) es
     len = genericLength es
 
   ETuple es -> {-# SCC "evalExpr->ETuple" #-} do
-     let xs = map eval es
+     xs <- mapM (delay Nothing . eval) es
      return $ VTuple xs
 
   ERec fields -> {-# SCC "evalExpr->ERec" #-} do
-     let xs = [ (f, eval e)
-              | (f,e) <- fields
-              ]
+     xs <- sequence [ do thk <- delay Nothing (eval e)
+                         return (f, thk)
+                    | (f, e) <- fields
+                    ]
      return $ VRecord xs
 
   ESel e sel -> {-# SCC "evalExpr->ESel" #-} do
@@ -129,7 +136,7 @@ evalExpr env expr = case expr of
       val     -> do vdoc <- ppV val
                     panic "[Eval] evalExpr"
                       ["expected a polymorphic value"
-                      , show vdoc, show e, show ty
+                      , show vdoc, show (pp e), show (pp ty)
                       ]
 
   EApp f x -> {-# SCC "evalExpr->EApp" #-} do
@@ -159,17 +166,17 @@ evalExpr env expr = case expr of
 
 -- Newtypes --------------------------------------------------------------------
 
-evalNewtypes :: EvalPrims b w
+evalNewtypes :: EvalPrims b w i
              => Map.Map Name Newtype
-             -> GenEvalEnv b w
-             -> Eval (GenEvalEnv b w)
+             -> GenEvalEnv b w i
+             -> Eval (GenEvalEnv b w i)
 evalNewtypes nts env = foldM (flip evalNewtype) env $ Map.elems nts
 
 -- | Introduce the constructor function for a newtype.
-evalNewtype :: EvalPrims b w
+evalNewtype :: EvalPrims b w i
             => Newtype
-            -> GenEvalEnv b w
-            -> Eval (GenEvalEnv b w)
+            -> GenEvalEnv b w i
+            -> Eval (GenEvalEnv b w i)
 evalNewtype nt = bindVar (ntName nt) (return (foldr tabs con (ntParams nt)))
   where
   tabs _tp body = tlam (\ _ -> body)
@@ -180,16 +187,16 @@ evalNewtype nt = bindVar (ntName nt) (return (foldr tabs con (ntParams nt)))
 
 -- | Extend the given evaluation environment with the result of evaluating the
 --   given collection of declaration groups.
-evalDecls :: EvalPrims b w
+evalDecls :: EvalPrims b w i
           => [DeclGroup]         -- ^ Declaration groups to evaluate
-          -> GenEvalEnv b w      -- ^ Environment to extend
-          -> Eval (GenEvalEnv b w)
+          -> GenEvalEnv b w i    -- ^ Environment to extend
+          -> Eval (GenEvalEnv b w i)
 evalDecls dgs env = foldM evalDeclGroup env dgs
 
-evalDeclGroup :: EvalPrims b w
-              => GenEvalEnv b w
+evalDeclGroup :: EvalPrims b w i
+              => GenEvalEnv b w i
               -> DeclGroup
-              -> Eval (GenEvalEnv b w)
+              -> Eval (GenEvalEnv b w i)
 evalDeclGroup env dg = do
   case dg of
     Recursive ds -> do
@@ -219,13 +226,13 @@ evalDeclGroup env dg = do
 --
 --   In order to faithfully evaluate the nonstrict semantics of Cryptol, we have to take some
 --   care in this process.  In particular, we need to ensure that every recursive definition
---   binding is indistinguishable from it's eta-expanded form.  The straightforward solution
+--   binding is indistinguishable from its eta-expanded form.  The straightforward solution
 --   to this is to force an eta-expansion procedure on all recursive definitions.
---   However, for the so-called 'Value' types we can instead optimisticly use the 'delayFill'
+--   However, for the so-called 'Value' types we can instead optimistically use the 'delayFill'
 --   operation and only fall back on full eta expansion if the thunk is double-forced.
-fillHole :: BitWord b w
-         => GenEvalEnv b w
-         -> (Name, Schema, Eval (GenValue b w), Eval (GenValue b w) -> Eval ())
+fillHole :: BitWord b w i
+         => GenEvalEnv b w i
+         -> (Name, Schema, Eval (GenValue b w i), Eval (GenValue b w i) -> Eval ())
          -> Eval ()
 fillHole env (nm, sch, _, fill) = do
   case lookupVar nm env of
@@ -240,7 +247,7 @@ fillHole env (nm, sch, _, fill) = do
 --   be implemented rather more efficently than general types because we can
 --   rely on the 'delayFill' operation to build a thunk that falls back on performing
 --   eta-expansion rather than doing it eagerly.
-isValueType :: GenEvalEnv b w -> Schema -> Bool
+isValueType :: GenEvalEnv b w i -> Schema -> Bool
 isValueType env Forall{ sVars = [], sProps = [], sType = t0 }
    = go (evalValType (envTypes env) t0)
  where
@@ -254,28 +261,28 @@ isValueType _ _ = False
 
 
 -- | Eta-expand a word value.  This forces an unpacked word representation.
-etaWord  :: BitWord b w
+etaWord  :: BitWord b w i
          => Integer
-         -> Eval (GenValue b w)
-         -> Eval (WordValue b w)
+         -> Eval (GenValue b w i)
+         -> Eval (WordValue b w i)
 etaWord n x = do
   w <- delay Nothing (fromWordVal "during eta-expansion" =<< x)
   return $ BitsVal $ Seq.fromFunction (fromInteger n) $ \i ->
     do w' <- w; indexWordValue w' (toInteger i)
 
 
--- | Given a simulator value and it's type, fully eta-expand the value.  This
+-- | Given a simulator value and its type, fully eta-expand the value.  This
 --   is a type-directed pass that always produces a canonical value of the
 --   expected shape.  Eta expansion of values is sometimes necessary to ensure
 --   the correct evaluation semantics of recursive definitions.  Otherwise,
 --   expressions that should be expected to produce well-defined values in the
 --   denotational semantics will fail to terminate instead.
-etaDelay :: BitWord b w
+etaDelay :: BitWord b w i
          => String
-         -> GenEvalEnv b w
+         -> GenEvalEnv b w i
          -> Schema
-         -> Eval (GenValue b w)
-         -> Eval (GenValue b w)
+         -> Eval (GenValue b w i)
+         -> Eval (GenValue b w i)
 etaDelay msg env0 Forall{ sVars = vs0, sType = tp0 } = goTpVars env0 vs0
   where
   goTpVars env []     x = go (evalValType (envTypes env) tp0) x
@@ -289,8 +296,9 @@ etaDelay msg env0 Forall{ sVars = vs0, sType = tp0 } = goTpVars env0 vs0
 
   go tp (Ready x) =
     case x of
-      VBit _    -> return x
-      VWord _ _ -> return x
+      VBit _     -> return x
+      VInteger _ -> return x
+      VWord _ _  -> return x
       VSeq n xs
         | TVSeq _nt el <- tp
         -> return $ VSeq n $ IndexSeqMap $ \i -> go el (lookupSeqMap xs i)
@@ -320,6 +328,8 @@ etaDelay msg env0 Forall{ sVars = vs0, sType = tp0 } = goTpVars env0 vs0
   go tp x =
     case tp of
       TVBit -> x
+      TVInteger -> x
+      TVIntMod _ -> x
 
       TVSeq n TVBit ->
           do w <- delayFill (fromWordVal "during eta-expansion" =<< x) (etaWord n x)
@@ -358,7 +368,7 @@ etaDelay msg env0 Forall{ sVars = vs0, sType = tp0 } = goTpVars env0 vs0
 
 
 declHole :: Decl
-         -> Eval (Name, Schema, Eval (GenValue b w), Eval (GenValue b w) -> Eval ())
+         -> Eval (Name, Schema, Eval (GenValue b w i), Eval (GenValue b w i) -> Eval ())
 declHole d =
   case dDefinition d of
     DPrim   -> evalPanic "Unexpected primitive declaration in recursive group"
@@ -379,14 +389,14 @@ declHole d =
 --   handle the subtle name-binding issues that arise from recursive
 --   definitions.  The 'read only' environment is used to bring recursive
 --   names into scope while we are still defining them.
-evalDecl :: EvalPrims b w
-         => GenEvalEnv b w  -- ^ A 'read only' environment for use in declaration bodies
-         -> GenEvalEnv b w  -- ^ An evaluation environment to extend with the given declaration
-         -> Decl            -- ^ The declaration to evaluate
-         -> Eval (GenEvalEnv b w)
+evalDecl :: EvalPrims b w i
+         => GenEvalEnv b w i  -- ^ A 'read only' environment for use in declaration bodies
+         -> GenEvalEnv b w i  -- ^ An evaluation environment to extend with the given declaration
+         -> Decl              -- ^ The declaration to evaluate
+         -> Eval (GenEvalEnv b w i)
 evalDecl renv env d =
   case dDefinition d of
-    DPrim   -> bindVarDirect (dName d) (evalPrim d) env
+    DPrim   -> return $ bindVarDirect (dName d) (evalPrim d) env
     DExpr e -> bindVar (dName d) (evalExpr renv e) env
 
 
@@ -395,11 +405,11 @@ evalDecl renv env d =
 -- | Apply the the given "selector" form to the given value.  This function pushes
 --   tuple and record selections pointwise down into other value constructs
 --   (e.g., streams and functions).
-evalSel :: forall b w
-         . EvalPrims b w
-        => GenValue b w
+evalSel :: forall b w i
+         . EvalPrims b w i
+        => GenValue b w i
         -> Selector
-        -> Eval (GenValue b w)
+        -> Eval (GenValue b w i)
 evalSel val sel = case sel of
 
   TupleSel n _  -> tupleSel n val
@@ -410,9 +420,6 @@ evalSel val sel = case sel of
   tupleSel n v =
     case v of
       VTuple vs       -> vs !! n
-      VSeq w vs       -> VSeq w <$> mapSeqMap (tupleSel n) vs
-      VStream vs      -> VStream <$> mapSeqMap (tupleSel n) vs
-      VFun f          -> return $ VFun (\x -> tupleSel n =<< f x)
       _               -> do vdoc <- ppValue defaultPPOpts v
                             evalPanic "Cryptol.Eval.evalSel"
                               [ "Unexpected value in tuple selection"
@@ -421,9 +428,6 @@ evalSel val sel = case sel of
   recordSel n v =
     case v of
       VRecord {}      -> lookupRecord n v
-      VSeq w vs       -> VSeq w <$> mapSeqMap (recordSel n) vs
-      VStream vs      -> VStream <$> mapSeqMap (recordSel n) vs
-      VFun f          -> return $ VFun (\x -> recordSel n =<< f x)
       _               -> do vdoc <- ppValue defaultPPOpts v
                             evalPanic "Cryptol.Eval.evalSel"
                               [ "Unexpected value in record selection"
@@ -447,28 +451,31 @@ evalSel val sel = case sel of
 -- | Evaluation environments for list comprehensions: Each variable
 -- name is bound to a list of values, one for each element in the list
 -- comprehension.
-data ListEnv b w = ListEnv
-  { leVars   :: !(Map.Map Name (Integer -> Eval (GenValue b w)))
+data ListEnv b w i = ListEnv
+  { leVars   :: !(Map.Map Name (Integer -> Eval (GenValue b w i)))
       -- ^ Bindings whose values vary by position
-  , leStatic :: !(Map.Map Name (Eval (GenValue b w)))
+  , leStatic :: !(Map.Map Name (Eval (GenValue b w i)))
       -- ^ Bindings whose values are constant
   , leTypes  :: !TypeEnv
   }
 
-instance Monoid (ListEnv b w) where
+instance Semigroup (ListEnv b w i) where
+  l <> r = ListEnv
+    { leVars   = Map.union (leVars  l)  (leVars  r)
+    , leStatic = Map.union (leStatic l) (leStatic r)
+    , leTypes  = Map.union (leTypes l)  (leTypes r)
+    }
+
+instance Monoid (ListEnv b w i) where
   mempty = ListEnv
     { leVars   = Map.empty
     , leStatic = Map.empty
     , leTypes  = Map.empty
     }
 
-  mappend l r = ListEnv
-    { leVars   = Map.union (leVars  l)  (leVars  r)
-    , leStatic = Map.union (leStatic l) (leStatic r)
-    , leTypes  = Map.union (leTypes l)  (leTypes r)
-    }
+  mappend l r = l <> r
 
-toListEnv :: GenEvalEnv b w -> ListEnv b w
+toListEnv :: GenEvalEnv b w i -> ListEnv b w i
 toListEnv e =
   ListEnv
   { leVars   = mempty
@@ -479,7 +486,7 @@ toListEnv e =
 -- | Evaluate a list environment at a position.
 --   This choses a particular value for the varying
 --   locations.
-evalListEnv :: ListEnv b w -> Integer -> GenEvalEnv b w
+evalListEnv :: ListEnv b w i -> Integer -> GenEvalEnv b w i
 evalListEnv (ListEnv vm st tm) i =
     let v = fmap ($i) vm
      in EvalEnv{ envVars = Map.union v st
@@ -487,21 +494,21 @@ evalListEnv (ListEnv vm st tm) i =
                }
 
 bindVarList :: Name
-            -> (Integer -> Eval (GenValue b w))
-            -> ListEnv b w
-            -> ListEnv b w
+            -> (Integer -> Eval (GenValue b w i))
+            -> ListEnv b w i
+            -> ListEnv b w i
 bindVarList n vs lenv = lenv { leVars = Map.insert n vs (leVars lenv) }
 
 -- List Comprehensions ---------------------------------------------------------
 
 -- | Evaluate a comprehension.
-evalComp :: EvalPrims b w
-         => GenEvalEnv b w  -- ^ Starting evaluation environment
-         -> Nat'            -- ^ Length of the comprehension
-         -> TValue          -- ^ Type of the comprehension elements
-         -> Expr            -- ^ Head expression of the comprehension
-         -> [[Match]]       -- ^ List of parallel comprehension branches
-         -> Eval (GenValue b w)
+evalComp :: EvalPrims b w i
+         => GenEvalEnv b w i -- ^ Starting evaluation environment
+         -> Nat'             -- ^ Length of the comprehension
+         -> TValue           -- ^ Type of the comprehension elements
+         -> Expr             -- ^ Head expression of the comprehension
+         -> [[Match]]        -- ^ List of parallel comprehension branches
+         -> Eval (GenValue b w i)
 evalComp env len elty body ms =
        do lenv <- mconcat <$> mapM (branchEnvs (toListEnv env)) ms
           mkSeq len elty <$> memoMap (IndexSeqMap $ \i -> do
@@ -509,17 +516,17 @@ evalComp env len elty body ms =
 
 -- | Turn a list of matches into the final environments for each iteration of
 -- the branch.
-branchEnvs :: EvalPrims b w
-           => ListEnv b w
+branchEnvs :: EvalPrims b w i
+           => ListEnv b w i
            -> [Match]
-           -> Eval (ListEnv b w)
+           -> Eval (ListEnv b w i)
 branchEnvs env matches = foldM evalMatch env matches
 
 -- | Turn a match into the list of environments it represents.
-evalMatch :: EvalPrims b w
-          => ListEnv b w
+evalMatch :: EvalPrims b w i
+          => ListEnv b w i
           -> Match
-          -> Eval (ListEnv b w)
+          -> Eval (ListEnv b w i)
 evalMatch lenv m = case m of
 
   -- many envs
