@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Eval.Env
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -14,7 +14,7 @@
 {-# LANGUAGE DeriveGeneric #-}
 module Cryptol.Eval.Env where
 
-import Cryptol.Eval.Monad( Eval, delay, ready )
+import Cryptol.Eval.Monad( Eval, delay, ready, PPOpts )
 import Cryptol.Eval.Type
 import Cryptol.Eval.Value
 import Cryptol.ModuleSystem.Name
@@ -24,6 +24,7 @@ import Cryptol.Utils.PP
 
 
 import qualified Data.Map.Strict as Map
+import Data.Semigroup
 
 import GHC.Generics (Generic)
 import Control.DeepSeq
@@ -33,37 +34,40 @@ import Prelude.Compat
 
 -- Evaluation Environment ------------------------------------------------------
 
-data GenEvalEnv b w = EvalEnv
-  { envVars       :: !(Map.Map Name (Eval (GenValue b w)))
+data GenEvalEnv b w i = EvalEnv
+  { envVars       :: !(Map.Map Name (Eval (GenValue b w i)))
   , envTypes      :: !TypeEnv
   } deriving (Generic, NFData)
 
-instance Monoid (GenEvalEnv b w) where
+instance Semigroup (GenEvalEnv b w i) where
+  l <> r = EvalEnv
+    { envVars     = Map.union (envVars     l) (envVars     r)
+    , envTypes    = Map.union (envTypes    l) (envTypes    r)
+    }
+
+instance Monoid (GenEvalEnv b w i) where
   mempty = EvalEnv
     { envVars       = Map.empty
     , envTypes      = Map.empty
     }
 
-  mappend l r = EvalEnv
-    { envVars     = Map.union (envVars     l) (envVars     r)
-    , envTypes    = Map.union (envTypes    l) (envTypes    r)
-    }
+  mappend l r = l <> r
 
-ppEnv :: BitWord b w => PPOpts -> GenEvalEnv b w -> Eval Doc
+ppEnv :: BitWord b w i => PPOpts -> GenEvalEnv b w i -> Eval Doc
 ppEnv opts env = brackets . fsep <$> mapM bind (Map.toList (envVars env))
   where
    bind (k,v) = do vdoc <- ppValue opts =<< v
                    return (pp k <+> text "->" <+> vdoc)
 
 -- | Evaluation environment with no bindings
-emptyEnv :: GenEvalEnv b w
+emptyEnv :: GenEvalEnv b w i
 emptyEnv  = mempty
 
 -- | Bind a variable in the evaluation environment.
 bindVar :: Name
-        -> Eval (GenValue b w)
-        -> GenEvalEnv b w
-        -> Eval (GenEvalEnv b w)
+        -> Eval (GenValue b w i)
+        -> GenEvalEnv b w i
+        -> Eval (GenEvalEnv b w i)
 bindVar n val env = do
   let nm = show $ ppLocName n
   val' <- delay (Just nm) val
@@ -72,24 +76,24 @@ bindVar n val env = do
 -- | Bind a variable to a value in the evaluation environment, without
 --   creating a thunk.
 bindVarDirect :: Name
-              -> GenValue b w
-              -> GenEvalEnv b w
-              -> Eval (GenEvalEnv b w)
+              -> GenValue b w i
+              -> GenEvalEnv b w i
+              -> GenEvalEnv b w i
 bindVarDirect n val env = do
-  return $ env{ envVars = Map.insert n (ready val) (envVars env) }
+  env{ envVars = Map.insert n (ready val) (envVars env) }
 
 -- | Lookup a variable in the environment.
 {-# INLINE lookupVar #-}
-lookupVar :: Name -> GenEvalEnv b w -> Maybe (Eval (GenValue b w))
+lookupVar :: Name -> GenEvalEnv b w i -> Maybe (Eval (GenValue b w i))
 lookupVar n env = Map.lookup n (envVars env)
 
 -- | Bind a type variable of kind *.
 {-# INLINE bindType #-}
-bindType :: TVar -> Either Nat' TValue -> GenEvalEnv b w -> GenEvalEnv b w
+bindType :: TVar -> Either Nat' TValue -> GenEvalEnv b w i -> GenEvalEnv b w i
 bindType p ty env = env { envTypes = Map.insert p ty (envTypes env) }
 
 -- | Lookup a type variable.
 {-# INLINE lookupType #-}
-lookupType :: TVar -> GenEvalEnv b w -> Maybe (Either Nat' TValue)
+lookupType :: TVar -> GenEvalEnv b w i -> Maybe (Either Nat' TValue)
 lookupType p env = Map.lookup p (envTypes env)
 
