@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Eval.Type
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -21,12 +21,15 @@ import Cryptol.Utils.Ident (Ident)
 import Data.Maybe(fromMaybe)
 import qualified Data.Map.Strict as Map
 import GHC.Generics (Generic)
+import GHC.Stack(HasCallStack)
 import Control.DeepSeq
 
 -- | An evaluated type of kind *.
 -- These types do not contain type variables, type synonyms, or type functions.
 data TValue
   = TVBit                    -- ^ @ Bit @
+  | TVInteger                -- ^ @ Integer @
+  | TVIntMod Integer         -- ^ @ Z n @
   | TVSeq Integer TValue     -- ^ @ [n]a @
   | TVStream TValue          -- ^ @ [inf]t @
   | TVTuple [TValue]         -- ^ @ (a, b, c )@
@@ -39,6 +42,8 @@ tValTy :: TValue -> Type
 tValTy tv =
   case tv of
     TVBit       -> tBit
+    TVInteger   -> tInteger
+    TVIntMod n  -> tIntMod (tNum n)
     TVSeq n t   -> tSeq (tNum n) (tValTy t)
     TVStream t  -> tSeq tInf (tValTy t)
     TVTuple ts  -> tTuple (map tValTy ts)
@@ -76,7 +81,7 @@ type TypeEnv = Map.Map TVar (Either Nat' TValue)
 
 
 -- | Evaluation for types (kind * or #).
-evalType :: TypeEnv -> Type -> Either Nat' TValue
+evalType :: HasCallStack => TypeEnv -> Type -> Either Nat' TValue
 evalType env ty =
   case ty of
     TVar tv ->
@@ -89,6 +94,10 @@ evalType env ty =
     TCon (TC c) ts ->
       case (c, ts) of
         (TCBit, [])     -> Right $ TVBit
+        (TCInteger, []) -> Right $ TVInteger
+        (TCIntMod, [n]) -> case num n of
+                             Inf   -> evalPanic "evalType" ["invalid type Z inf"]
+                             Nat m -> Right $ TVIntMod m
         (TCSeq, [n, t]) -> Right $ tvSeq (num n) (val t)
         (TCFun, [a, b]) -> Right $ TVFun (val a) (val b)
         (TCTuple _, _)  -> Right $ TVTuple (map val ts)
@@ -105,14 +114,14 @@ evalType env ty =
     num = evalNumType env
 
 -- | Evaluation for value types (kind *).
-evalValType :: TypeEnv -> Type -> TValue
+evalValType :: HasCallStack => TypeEnv -> Type -> TValue
 evalValType env ty =
   case evalType env ty of
     Left _ -> evalPanic "evalValType" ["expected value type, found numeric type"]
     Right t -> t
 
 -- | Evaluation for number types (kind #).
-evalNumType :: TypeEnv -> Type -> Nat'
+evalNumType :: HasCallStack => TypeEnv -> Type -> Nat'
 evalNumType env ty =
   case evalType env ty of
     Left n -> n
@@ -120,7 +129,7 @@ evalNumType env ty =
 
 
 -- | Reduce type functions, raising an exception for undefined values.
-evalTF :: TFun -> [Nat'] -> Nat'
+evalTF :: HasCallStack => TFun -> [Nat'] -> Nat'
 evalTF f vs
   | TCAdd           <- f, [x,y]   <- vs  =      nAdd x y
   | TCSub           <- f, [x,y]   <- vs  = mb $ nSub x y
@@ -131,6 +140,8 @@ evalTF f vs
   | TCExp           <- f, [x,y]   <- vs  =      nExp x y
   | TCMin           <- f, [x,y]   <- vs  =      nMin x y
   | TCMax           <- f, [x,y]   <- vs  =      nMax x y
+  | TCCeilDiv       <- f, [x,y]   <- vs  = mb $ nCeilDiv x y
+  | TCCeilMod       <- f, [x,y]   <- vs  = mb $ nCeilMod x y
   | TCLenFromThen   <- f, [x,y,z] <- vs  = mb $ nLenFromThen x y z
   | TCLenFromThenTo <- f, [x,y,z] <- vs  = mb $ nLenFromThenTo x y z
   | otherwise  = evalPanic "evalTF"
