@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.ModuleSystem
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -16,7 +16,7 @@ module Cryptol.ModuleSystem (
   , ModuleCmd, ModuleRes
   , findModule
   , loadModuleByPath
-  , loadModule
+  , loadModuleByName
   , checkExpr
   , evalExpr
   , checkDecls
@@ -28,10 +28,11 @@ module Cryptol.ModuleSystem (
   , renameType
 
     -- * Interfaces
-  , Iface(..), IfaceDecls(..), genIface
+  , Iface(..), IfaceParams(..), IfaceDecls(..), genIface
   , IfaceTySyn, IfaceDecl(..)
   ) where
 
+import qualified Cryptol.Eval as E
 import qualified Cryptol.Eval.Value        as E
 import           Cryptol.ModuleSystem.Env
 import           Cryptol.ModuleSystem.Interface
@@ -48,7 +49,7 @@ import qualified Cryptol.Utils.Ident as M
 
 -- Public Interface ------------------------------------------------------------
 
-type ModuleCmd a = ModuleEnv -> IO (ModuleRes a)
+type ModuleCmd a = (E.EvalOpts,ModuleEnv) -> IO (ModuleRes a)
 
 type ModuleRes a = (Either ModuleError (a,ModuleEnv), [ModuleWarning])
 
@@ -60,25 +61,20 @@ findModule :: P.ModName -> ModuleCmd FilePath
 findModule n env = runModuleM env (Base.findModule n)
 
 -- | Load the module contained in the given file.
-loadModuleByPath :: FilePath -> ModuleCmd T.Module
-loadModuleByPath path env = runModuleM (resetModuleEnv env) $ do
-  -- unload the module if it already exists
-  unloadModule path
-
+loadModuleByPath :: FilePath -> ModuleCmd (FilePath,T.Module)
+loadModuleByPath path (evo,env) = runModuleM (evo,resetModuleEnv env) $ do
+  unloadModule ((path ==) . lmFilePath)
   m <- Base.loadModuleByPath path
   setFocusedModule (T.mName m)
-  return m
+  return (path,m)
 
 -- | Load the given parsed module.
-loadModule :: FilePath -> P.Module PName -> ModuleCmd T.Module
-loadModule path m env = runModuleM env $ do
-  -- unload the module if it already exists
-  unloadModule path
-
-  let n = P.thing (P.mName m)
-  m' <- loadingModule n (Base.loadModule path m)
+loadModuleByName :: P.ModName -> ModuleCmd (FilePath,T.Module)
+loadModuleByName n env = runModuleM env $ do
+  unloadModule ((n ==) . lmName)
+  (path,m') <- Base.loadModuleFrom (FromModule n)
   setFocusedModule (T.mName m')
-  return m'
+  return (path,m')
 
 -- Extended Environments -------------------------------------------------------
 
