@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.ModuleSystem.Env
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -26,12 +26,12 @@ import qualified Cryptol.TypeCheck as T
 import qualified Cryptol.TypeCheck.AST as T
 import Cryptol.Utils.PP (NameDisp)
 
-import Control.Monad (guard)
+import Control.Monad (guard,mplus)
 import qualified Control.Exception as X
-import Data.Foldable (fold)
 import Data.Function (on)
 import qualified Data.Map as Map
-import Data.Monoid ((<>))
+import Data.Maybe(fromMaybe)
+import Data.Semigroup
 import System.Directory (getAppUserDataDirectory, getCurrentDirectory)
 import System.Environment(getExecutablePath)
 import System.FilePath ((</>), normalise, joinPath, splitPath, takeDirectory)
@@ -45,19 +45,50 @@ import Prelude.Compat
 
 -- Module Environment ----------------------------------------------------------
 
+-- | This is the current state of the interpreter.
 data ModuleEnv = ModuleEnv
   { meLoadedModules :: LoadedModules
-  , meNameSeeds     :: T.NameSeeds
-  , meEvalEnv       :: EvalEnv
-  , meFocusedModule :: Maybe ModName
-  , meSearchPath    :: [FilePath]
-  , meDynEnv        :: DynamicEnv
-  , meMonoBinds     :: !Bool
-  , meSolverConfig  :: T.SolverConfig
-  , meCoreLint      :: CoreLint
-  , meSupply        :: !Supply
-  } deriving (Generic, NFData)
+    -- ^ Information about all loaded modules.  See 'LoadedModule'.
+    -- Contains information such as the file where the module was loaded
+    -- from, as well as the module's interface, used for type checking.
 
+  , meNameSeeds     :: T.NameSeeds
+    -- ^ A source of new names for the type checker.
+
+  , meSolverConfig  :: T.SolverConfig
+    -- ^ Configuration settings for the SMT solver used for type-checking.
+
+  , meEvalEnv       :: EvalEnv
+    -- ^ The evaluation environment.  Contains the values for all loaded
+    -- modules, both public and private.
+
+  , meCoreLint      :: CoreLint
+    -- ^ Should we run the linter to ensure sanity.
+
+  , meMonoBinds     :: !Bool
+    -- ^ Are we assuming that local bindings are monomorphic.
+    -- XXX: We should probably remove this flag, and set it to 'True'.
+
+
+
+  , meFocusedModule :: Maybe ModName
+    -- ^ The "current" module.  Used to decide how to print names, for example.
+
+  , meSearchPath    :: [FilePath]
+    -- ^ Where we look for things.
+
+  , meDynEnv        :: DynamicEnv
+    -- ^ This contains additional definitions that were made at the command
+    -- line, and so they don't reside in any module.
+
+  , meSupply        :: !Supply
+    -- ^ Name source for the renamer
+
+  } deriving Generic
+
+instance NFData ModuleEnv
+
+-- | Should we run the linter?
 data CoreLint = NoCoreLint        -- ^ Don't run core lint
               | CoreLint          -- ^ Run core lint
   deriving (Generic, NFData)
@@ -72,7 +103,7 @@ resetModuleEnv env = env
   }
 
 initialModuleEnv :: IO ModuleEnv
-initialModuleEnv  = do
+initialModuleEnv = do
   curDir <- getCurrentDirectory
 #ifndef RELOCATABLE
   dataDir <- getDataDir
@@ -84,37 +115,40 @@ initialModuleEnv  = do
   let handle :: X.IOException -> IO String
       handle _e = return ""
   userDir <- X.catch (getAppUserDataDirectory "cryptol") handle
+  let searchPath = [ curDir
+                   -- something like $HOME/.cryptol
+                   , userDir
+#if defined(mingw32_HOST_OS) || defined(__MINGW32__)
+                   -- ../cryptol on win32
+                   , instDir </> "cryptol"
+#else
+                   -- ../share/cryptol on others
+                   , instDir </> "share" </> "cryptol"
+#endif
+
+#ifndef RELOCATABLE
+                   -- Cabal-defined data directory. Since this
+                   -- is usually a global location like
+                   -- /usr/local, search this one last in case
+                   -- someone has multiple Cryptols
+                   , dataDir
+#endif
+                   ]
+
   return ModuleEnv
     { meLoadedModules = mempty
     , meNameSeeds     = T.nameSeeds
     , meEvalEnv       = mempty
     , meFocusedModule = Nothing
       -- we search these in order, taking the first match
-    , meSearchPath    = [ curDir
-                          -- something like $HOME/.cryptol
-                        , userDir
-#if defined(mingw32_HOST_OS) || defined(__MINGW32__)
-                          -- ../cryptol on win32
-                        , instDir </> "cryptol"
-#else
-                          -- ../share/cryptol on others
-                        , instDir </> "share" </> "cryptol"
-#endif
-
-#ifndef RELOCATABLE
-                          -- Cabal-defined data directory. Since this
-                          -- is usually a global location like
-                          -- /usr/local, search this one last in case
-                          -- someone has multiple Cryptols
-                        , dataDir
-#endif
-                        ]
+    , meSearchPath    = searchPath
     , meDynEnv        = mempty
     , meMonoBinds     = True
     , meSolverConfig  = T.SolverConfig
                           { T.solverPath = "z3"
                           , T.solverArgs = [ "-smt2", "-in" ]
                           , T.solverVerbose = 0
+                          , T.solverPreludePath = searchPath
                           }
     , meCoreLint      = NoCoreLint
     , meSupply        = emptySupply
@@ -137,18 +171,28 @@ loadedModules = map lmModule . getLoadedModules . meLoadedModules
 --
 -- XXX This could really do with some better error handling, just returning
 -- mempty when one of the imports fails isn't really desirable.
-focusedEnv :: ModuleEnv -> (IfaceDecls,R.NamingEnv,NameDisp)
-focusedEnv me = fold $
+--
+-- XXX: This is not quite right.   For example, it does not take into
+-- account *how* things were imported in a module (e.g., qualified).
+-- It would be simpler to simply store the naming environment that was
+-- actually used when we renamed the module.
+focusedEnv :: ModuleEnv -> (IfaceParams,IfaceDecls,R.NamingEnv,NameDisp)
+focusedEnv me =
+  fromMaybe (noIfaceParams, mempty, mempty, mempty) $
   do fm   <- meFocusedModule me
      lm   <- lookupModule fm me
      deps <- mapM loadImport (T.mImports (lmModule lm))
      let (ifaces,names) = unzip deps
          Iface { .. }   = lmInterface lm
          localDecls     = ifPublic `mappend` ifPrivate
-         localNames     = R.unqualifiedEnv localDecls
+         localNames     = R.unqualifiedEnv localDecls `mappend`
+                                              R.modParamsNamingEnv ifParams
          namingEnv      = localNames `R.shadowing` mconcat names
 
-     return (mconcat (localDecls:ifaces), namingEnv, R.toNameDisp namingEnv)
+     return ( ifParams
+            , mconcat (localDecls:ifaces)
+            , namingEnv
+            , R.toNameDisp namingEnv)
   where
   loadImport imp =
     do lm <- lookupModule (iModule imp) me
@@ -163,51 +207,66 @@ dynamicEnv me = (decls,names,R.toNameDisp names)
   decls = deIfaceDecls (meDynEnv me)
   names = R.unqualifiedEnv decls
 
--- | Retrieve all 'IfaceDecls' referenced by a module, as well as all of its
--- public and private declarations, checking expressions
-qualifiedEnv :: ModuleEnv -> IfaceDecls
-qualifiedEnv me = fold $
-  do fm   <- meFocusedModule me
-     lm   <- lookupModule fm me
-     deps <- mapM loadImport (T.mImports (lmModule lm))
-     let Iface { .. } = lmInterface lm
-     return (mconcat (ifPublic : ifPrivate : deps))
-  where
-  loadImport imp =
-    do lm <- lookupModule (iModule imp) me
-       return (ifPublic (lmInterface lm))
-
 
 -- Loaded Modules --------------------------------------------------------------
 
-newtype LoadedModules = LoadedModules
-  { getLoadedModules :: [LoadedModule]
+data LoadedModules = LoadedModules
+  { lmLoadedModules      :: [LoadedModule]
+    -- ^ Invariants:
+    -- 1) All the dependencies of any module `m` must precede `m` in the list.
+    -- 2) Does not contain any parameterized modules.
+
+  , lmLoadedParamModules :: [LoadedModule]
+    -- ^ Loaded parameterized modules.
+
   } deriving (Show, Generic, NFData)
--- ^ Invariant: All the dependencies of any module `m` must precede `m` in the list.
+
+getLoadedModules :: LoadedModules -> [LoadedModule]
+getLoadedModules x = lmLoadedParamModules x ++ lmLoadedModules x
+
+instance Semigroup LoadedModules where
+  l <> r = LoadedModules
+    { lmLoadedModules = List.unionBy ((==) `on` lmName)
+                                      (lmLoadedModules l) (lmLoadedModules r)
+    , lmLoadedParamModules = lmLoadedParamModules l ++ lmLoadedParamModules r }
 
 instance Monoid LoadedModules where
-  mempty        = LoadedModules []
-  mappend l r   = LoadedModules
-                $ List.unionBy ((==) `on` lmName) (getLoadedModules l) (getLoadedModules r)
+  mempty = LoadedModules { lmLoadedModules = []
+                         , lmLoadedParamModules = []
+                         }
+  mappend l r = l <> r
 
 data LoadedModule = LoadedModule
   { lmName      :: ModName
-  , lmFilePath  :: FilePath -- ^ The file path used to load this module (may not be canonical)
-  , lmCanonicalPath :: FilePath -- ^ The canonical version of the path of this module
+  , lmFilePath  :: FilePath
+    -- ^ The file path used to load this module (may not be canonical)
+  , lmCanonicalPath :: FilePath
+    -- ^ The canonical version of the path of this module
   , lmInterface :: Iface
   , lmModule    :: T.Module
   } deriving (Show, Generic, NFData)
 
+-- | Has this module been loaded already.
 isLoaded :: ModName -> LoadedModules -> Bool
 isLoaded mn lm = any ((mn ==) . lmName) (getLoadedModules lm)
 
+-- | Try to find a previously loaded module
 lookupModule :: ModName -> ModuleEnv -> Maybe LoadedModule
-lookupModule mn env = List.find ((mn ==) . lmName) (getLoadedModules (meLoadedModules env))
+lookupModule mn me = search lmLoadedModules `mplus` search lmLoadedParamModules
+  where
+  search how = List.find ((mn ==) . lmName) (how (meLoadedModules me))
 
-addLoadedModule :: FilePath -> FilePath -> T.Module -> LoadedModules -> LoadedModules
+
+-- | Add a freshly loaded module.  If it was previously loaded, then
+-- the new version is ignored.
+addLoadedModule ::
+  FilePath -> FilePath -> T.Module -> LoadedModules -> LoadedModules
 addLoadedModule path canonicalPath tm lm
-  | isLoaded (T.mName tm) lm = lm
-  | otherwise                = LoadedModules (getLoadedModules lm ++ [loaded])
+  | isLoaded (T.mName tm) lm  = lm
+  | T.isParametrizedModule tm = lm { lmLoadedParamModules = loaded :
+                                                lmLoadedParamModules lm }
+  | otherwise                = lm { lmLoadedModules =
+                                          lmLoadedModules lm ++ [loaded] }
   where
   loaded = LoadedModule
     { lmName      = T.mName tm
@@ -217,15 +276,14 @@ addLoadedModule path canonicalPath tm lm
     , lmModule    = tm
     }
 
-removeLoadedModule :: FilePath -> LoadedModules -> LoadedModules
-removeLoadedModule path (LoadedModules ms) = LoadedModules (remove ms)
-  where
+-- | Remove a previously loaded module.
+removeLoadedModule :: (LoadedModule -> Bool) -> LoadedModules -> LoadedModules
+removeLoadedModule rm lm =
+  LoadedModules
+    { lmLoadedModules = filter (not . rm) (lmLoadedModules lm)
+    , lmLoadedParamModules = filter (not . rm) (lmLoadedParamModules lm)
+    }
 
-  remove (lm:rest)
-    | lmFilePath lm == path = rest
-    | otherwise             = lm : remove rest
-
-  remove [] = []
 
 -- Dynamic Environments --------------------------------------------------------
 
@@ -240,17 +298,20 @@ data DynamicEnv = DEnv
   , deEnv   :: EvalEnv
   } deriving (Generic, NFData)
 
+instance Semigroup DynamicEnv where
+  de1 <> de2 = DEnv
+    { deNames = deNames de1 <> deNames de2
+    , deDecls = deDecls de1 <> deDecls de2
+    , deEnv   = deEnv   de1 <> deEnv   de2
+    }
+
 instance Monoid DynamicEnv where
   mempty = DEnv
     { deNames = mempty
     , deDecls = mempty
     , deEnv   = mempty
     }
-  mappend de1 de2 = DEnv
-    { deNames = deNames de1 <> deNames de2
-    , deDecls = deDecls de1 <> deDecls de2
-    , deEnv   = deEnv   de1 <> deEnv   de2
-    }
+  mappend de1 de2 = de1 <> de2
 
 -- | Build 'IfaceDecls' that correspond to all of the bindings in the
 -- dynamic environment.
