@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.ModuleSystem.Name
 -- Copyright   :  (c) 2015-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -20,6 +20,7 @@
 module Cryptol.ModuleSystem.Name (
     -- * Names
     Name(), NameInfo(..)
+  , NameSource(..)
   , nameUnique
   , nameIdent
   , nameInfo
@@ -33,6 +34,9 @@ module Cryptol.ModuleSystem.Name (
     -- ** Creation
   , mkDeclared
   , mkParameter
+  , toParamInstName
+  , asParamName
+  , paramModRecParam
 
     -- ** Unique Supply
   , FreshM(..), nextUniqueM
@@ -45,17 +49,20 @@ module Cryptol.ModuleSystem.Name (
   , lookupPrimType
   ) where
 
-import           Cryptol.Parser.AST( Fixity(..) )
-import           Cryptol.Parser.Position (Range,Located(..))
+import           Cryptol.Parser.Fixity
+import           Cryptol.Parser.Position (Range,Located(..),emptyRange)
 import           Cryptol.Utils.Ident
 import           Cryptol.Utils.Panic
 import           Cryptol.Utils.PP
 
+
 import           Control.DeepSeq
 import           Control.Monad.Fix (MonadFix(mfix))
 import qualified Data.Map as Map
 import qualified Data.Monoid as M
 import           Data.Ord (comparing)
+import qualified Data.Text as Text
+import           Data.Char(isAlpha,toUpper)
 import           GHC.Generics (Generic)
 import           MonadLib
 import           Prelude ()
@@ -64,7 +71,7 @@ import           Prelude.Compat
 
 -- Names -----------------------------------------------------------------------
 -- | Information about the binding site of the name.
-data NameInfo = Declared !ModName
+data NameInfo = Declared !ModName !NameSource
                 -- ^ This name refers to a declaration from this module
               | Parameter
                 -- ^ This name is a parameter (function or type)
@@ -88,7 +95,11 @@ data Name = Name { nUnique :: {-# UNPACK #-} !Int
 
                  , nLoc :: !Range
                    -- ^ Where this name was defined
-                 } deriving (Show, Generic, NFData)
+                 } deriving (Generic, NFData, Show)
+
+
+data NameSource = SystemName | UserName
+                    deriving (Generic, NFData, Show, Eq)
 
 instance Eq Name where
   a == b = compare a b == EQ
@@ -102,15 +113,17 @@ cmpNameLexical :: Name -> Name -> Ordering
 cmpNameLexical l r =
   case (nameInfo l, nameInfo r) of
 
-    (Declared nsl,Declared nsr) ->
+    (Declared nsl _,Declared nsr _) ->
       case compare nsl nsr of
         EQ  -> comparing nameIdent l r
         cmp -> cmp
 
     (Parameter,Parameter) -> comparing nameIdent l r
 
-    (Declared nsl,Parameter) -> compare nsl (identText (nameIdent r))
-    (Parameter,Declared nsr) -> compare (identText (nameIdent l)) nsr
+    (Declared nsl _,Parameter) -> compare (modNameToText nsl)
+                                          (identText (nameIdent r))
+    (Parameter,Declared nsr _) -> compare (identText (nameIdent l))
+                                          (modNameToText nsr)
 
 
 -- | Compare two names by the way they would be displayed.
@@ -118,28 +131,44 @@ cmpNameDisplay :: NameDisp -> Name -> Name -> Ordering
 cmpNameDisplay disp l r =
   case (nameInfo l, nameInfo r) of
 
-    (Declared nsl, Declared nsr) ->
+    (Declared nsl _, Declared nsr _) -> -- XXX: uses system name info?
       let pfxl = fmtModName nsl (getNameFormat nsl (nameIdent l) disp)
           pfxr = fmtModName nsr (getNameFormat nsr (nameIdent r) disp)
-       in case compare pfxl pfxr of
-            EQ  -> compare (nameIdent l) (nameIdent r)
+       in case cmpText pfxl pfxr of
+            EQ  -> cmpName l r
             cmp -> cmp
 
-    (Parameter,Parameter) ->
-      compare (nameIdent l) (nameIdent r)
+    (Parameter,Parameter) -> cmpName l r
 
-    (Declared nsl,Parameter) ->
+    (Declared nsl _,Parameter) ->
       let pfxl = fmtModName nsl (getNameFormat nsl (nameIdent l) disp)
-       in case compare pfxl (identText (nameIdent r)) of
+       in case cmpText pfxl (identText (nameIdent r)) of
             EQ  -> GT
             cmp -> cmp
 
-    (Parameter,Declared nsr) ->
+    (Parameter,Declared nsr _) ->
       let pfxr = fmtModName nsr (getNameFormat nsr (nameIdent r) disp)
-       in case compare (identText (nameIdent l)) pfxr of
+       in case cmpText (identText (nameIdent l)) pfxr of
             EQ  -> LT
             cmp -> cmp
 
+  where
+  cmpName xs ys  = cmpIdent (nameIdent xs) (nameIdent ys)
+  cmpIdent xs ys = cmpText (identText xs) (identText ys)
+
+  -- Note that this assumes that `xs` is `l` and `ys` is `r`
+  cmpText xs ys =
+    case (Text.null xs, Text.null ys) of
+      (True,True)   -> EQ
+      (True,False)  -> LT
+      (False,True)  -> GT
+      (False,False) -> compare (cmp (fx l) xs) (cmp (fx r) ys)
+        where
+        fx a = fLevel <$> nameFixity a
+        cmp a cs = (ordC (Text.index cs 0), a, cs)
+        ordC a | isAlpha a  = fromEnum (toUpper a)
+               | a == '_'   = 1
+               | otherwise  = 0
 
 -- | Figure out how the name should be displayed, by referencing the display
 -- function in the environment. NOTE: this function doesn't take into account
@@ -148,11 +177,11 @@ ppName :: Name -> Doc
 ppName Name { .. } =
   case nInfo of
 
-    Declared m -> withNameDisp $ \disp ->
+    Declared m _ -> withNameDisp $ \disp ->
       case getNameFormat m nIdent disp of
-        Qualified m' -> pp m' <> text "::" <> pp nIdent
-        UnQualified  ->                       pp nIdent
-        NotInScope   -> pp m  <> text "::" <> pp nIdent
+        Qualified m' -> pp m' <.> text "::" <.> pp nIdent
+        UnQualified  ->                         pp nIdent
+        NotInScope   -> pp m  <.> text "::" <.> pp nIdent
 
     Parameter -> pp nIdent
 
@@ -188,10 +217,21 @@ nameLoc  = nLoc
 nameFixity :: Name -> Maybe Fixity
 nameFixity = nFixity
 
+
 asPrim :: Name -> Maybe Ident
-asPrim Name { .. }
-  | nInfo == Declared preludeName = Just nIdent
-  | otherwise                     = Nothing
+asPrim Name { .. } =
+  case nInfo of
+    Declared p _ | p == preludeName -> Just nIdent
+    _ -> Nothing
+
+toParamInstName :: Name -> Name
+toParamInstName n =
+  case nInfo n of
+    Declared m s -> n { nInfo = Declared (paramInstModName m) s }
+    Parameter   -> n
+
+asParamName :: Name -> Name
+asParamName n = n { nInfo = Parameter }
 
 
 -- Name Supply -----------------------------------------------------------------
@@ -269,7 +309,10 @@ data Supply = Supply !Int
 -- plenty of room for names that the compiler needs to know about (wired-in
 -- constants).
 emptySupply :: Supply
-emptySupply  = Supply 0
+emptySupply  = Supply 0x1000
+-- For one such name, see paramModRecParam
+-- XXX: perhaps we should simply not have such things, but that's the way
+-- for now.
 
 nextUnique :: Supply -> (Int,Supply)
 nextUnique (Supply n) = s' `seq` (n,s')
@@ -280,10 +323,10 @@ nextUnique (Supply n) = s' `seq` (n,s')
 -- Name Construction -----------------------------------------------------------
 
 -- | Make a new name for a declaration.
-mkDeclared :: ModName -> Ident -> Maybe Fixity -> Range -> Supply -> (Name,Supply)
-mkDeclared m nIdent nFixity nLoc s =
+mkDeclared :: ModName -> NameSource -> Ident -> Maybe Fixity -> Range -> Supply -> (Name,Supply)
+mkDeclared m sys nIdent nFixity nLoc s =
   let (nUnique,s') = nextUnique s
-      nInfo        = Declared m
+      nInfo        = Declared m sys
    in (Name { .. }, s')
 
 -- | Make a new parameter name.
@@ -293,6 +336,13 @@ mkParameter nIdent nLoc s =
       nFixity      = Nothing
    in (Name { nInfo = Parameter, .. }, s')
 
+paramModRecParam :: Name
+paramModRecParam = Name { nInfo = Parameter
+                        , nFixity = Nothing
+                        , nIdent  = packIdent "$modParams"
+                        , nLoc    = emptyRange
+                        , nUnique = 0x01
+                        }
 
 -- Prim Maps -------------------------------------------------------------------
 
