@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.ModuleSystem.NamingEnv
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -20,14 +20,17 @@ module Cryptol.ModuleSystem.NamingEnv where
 import Cryptol.ModuleSystem.Interface
 import Cryptol.ModuleSystem.Name
 import Cryptol.Parser.AST
+import Cryptol.Parser.Name(isGeneratedName)
 import Cryptol.Parser.Position
+import qualified Cryptol.TypeCheck.AST as T
 import Cryptol.Utils.PP
 import Cryptol.Utils.Panic (panic)
 
 import Data.List (nub)
-import Data.Maybe (catMaybes,fromMaybe)
+import Data.Maybe (catMaybes,fromMaybe,mapMaybe)
 import qualified Data.Map.Strict as Map
 import qualified Data.Set as Set
+import Data.Semigroup
 import MonadLib (runId,Id)
 
 import GHC.Generics (Generic)
@@ -49,6 +52,22 @@ data NamingEnv = NamingEnv { neExprs :: !(Map.Map PName [Name])
                              -- ^ Expression-level fixity environment
                            } deriving (Show, Generic, NFData)
 
+-- | Return a list of value-level names to which this parsed name may refer.
+lookupValNames :: PName -> NamingEnv -> [Name]
+lookupValNames qn ro = Map.findWithDefault [] qn (neExprs ro)
+
+-- | Return a list of type-level names to which this parsed name may refer.
+lookupTypeNames :: PName -> NamingEnv -> [Name]
+lookupTypeNames qn ro = Map.findWithDefault [] qn (neTypes ro)
+
+
+
+instance Semigroup NamingEnv where
+  l <> r   =
+    NamingEnv { neExprs  = Map.unionWith merge (neExprs  l) (neExprs  r)
+              , neTypes  = Map.unionWith merge (neTypes  l) (neTypes  r)
+              , neFixity = Map.union           (neFixity l) (neFixity r) }
+
 instance Monoid NamingEnv where
   mempty        =
     NamingEnv { neExprs  = Map.empty
@@ -57,10 +76,7 @@ instance Monoid NamingEnv where
 
   -- NOTE: merging the fixity maps is a special case that just prefers the left
   -- entry, as they're already keyed by a name with a unique
-  mappend l r   =
-    NamingEnv { neExprs  = Map.unionWith merge (neExprs  l) (neExprs  r)
-              , neTypes  = Map.unionWith merge (neTypes  l) (neTypes  r)
-              , neFixity = Map.union           (neFixity l) (neFixity r) }
+  mappend l r   = l <> r
 
   mconcat envs  =
     NamingEnv { neExprs  = Map.unionsWith merge (map neExprs  envs)
@@ -98,11 +114,11 @@ toNameDisp NamingEnv { .. } = NameDisp display
   names = Map.fromList
      $ [ mkEntry pn mn (nameIdent n) | (pn,ns)     <- Map.toList neExprs
                                      , n           <- ns
-                                     , Declared mn <- [nameInfo n] ]
+                                     , Declared mn _ <- [nameInfo n] ]
 
     ++ [ mkEntry pn mn (nameIdent n) | (pn,ns)     <- Map.toList neTypes
                                      , n           <- ns
-                                     , Declared mn <- [nameInfo n] ]
+                                     , Declared mn _ <- [nameInfo n] ]
 
   mkEntry pn mn i = ((mn,i),fmt)
     where
@@ -176,16 +192,25 @@ data InModule a = InModule !ModName a
 namingEnv' :: BindsNames a => a -> Supply -> (NamingEnv,Supply)
 namingEnv' a supply = runId (runSupplyT supply (runBuild (namingEnv a)))
 
+newTop :: FreshM m => ModName -> PName -> Maybe Fixity -> Range -> m Name
+newTop ns thing fx rng = liftSupply (mkDeclared ns src (getIdent thing) fx rng)
+  where src = if isGeneratedName thing then SystemName else UserName
+
+newLocal :: FreshM m => PName -> Range -> m Name
+newLocal thing rng = liftSupply (mkParameter (getIdent thing) rng)
 
 newtype BuildNamingEnv = BuildNamingEnv { runBuild :: SupplyT Id NamingEnv }
 
+instance Semigroup BuildNamingEnv where
+  BuildNamingEnv a <> BuildNamingEnv b = BuildNamingEnv $
+    do x <- a
+       y <- b
+       return (mappend x y)
+
 instance Monoid BuildNamingEnv where
   mempty = BuildNamingEnv (pure mempty)
 
-  mappend (BuildNamingEnv a) (BuildNamingEnv b) = BuildNamingEnv $
-    do x <- a
-       y <- b
-       return (mappend x y)
+  mappend = (<>)
 
   mconcat bs = BuildNamingEnv $
     do ns <- sequence (map runBuild bs)
@@ -258,6 +283,28 @@ unqualifiedEnv IfaceDecls { .. } =
               | d    <- Map.elems ifDecls ]
 
 
+-- | Compute an unqualified naming environment, containing the various module
+-- parameters.
+modParamsNamingEnv :: IfaceParams -> NamingEnv
+modParamsNamingEnv IfaceParams { .. } =
+  NamingEnv { neExprs = Map.fromList $ map fromFu $ Map.keys ifParamFuns
+            , neTypes = Map.fromList $ map fromTy $ Map.elems ifParamTypes
+            , neFixity = Map.fromList $ mapMaybe toFix $ Map.elems ifParamFuns
+            }
+
+  where
+  toPName n = mkUnqual (nameIdent n)
+
+  fromTy tp = let nm = T.mtpName tp
+              in (toPName nm, [nm])
+
+  fromFu f  = (toPName f, [f])
+
+  toFix x = do d <- T.mvpFixity x
+               return (T.mvpName x, d)
+
+
+
 data ImportIface = ImportIface Import Iface
 
 -- | Produce a naming environment from an interface file, that contains a
@@ -271,7 +318,7 @@ instance BindsNames ImportIface where
 instance BindsNames (InModule (Bind PName)) where
   namingEnv (InModule ns b) = BuildNamingEnv $
     do let Located { .. } = bName b
-       n <- liftSupply (mkDeclared ns (getIdent thing) (bFixity b) srcRange)
+       n <- newTop ns thing (bFixity b) srcRange
 
        let fixity = case bFixity b of
              Just f  -> mempty { neFixity = Map.singleton n f }
@@ -283,7 +330,7 @@ instance BindsNames (InModule (Bind PName)) where
 instance BindsNames (TParam PName) where
   namingEnv TParam { .. } = BuildNamingEnv $
     do let range = fromMaybe emptyRange tpRange
-       n <- liftSupply (mkParameter (getIdent tpName) range)
+       n <- newLocal tpName range
        return (singletonT tpName n)
 
 -- | The naming environment for a single module.  This is the mapping from
@@ -296,17 +343,33 @@ instance BindsNames (Module PName) where
 instance BindsNames (InModule (TopDecl PName)) where
   namingEnv (InModule ns td) =
     case td of
-      Decl d      -> namingEnv (InModule ns (tlValue d))
-      TDNewtype d -> namingEnv (InModule ns (tlValue d))
+      Decl d           -> namingEnv (InModule ns (tlValue d))
+      TDNewtype d      -> namingEnv (InModule ns (tlValue d))
+      DParameterType d -> namingEnv (InModule ns d)
+      DParameterConstraint {} -> mempty
+      DParameterFun  d -> namingEnv (InModule ns d)
       Include _   -> mempty
 
+instance BindsNames (InModule (ParameterFun PName)) where
+  namingEnv (InModule ns ParameterFun { .. }) = BuildNamingEnv $
+    do let Located { .. } = pfName
+       ntName <- newTop ns thing pfFixity srcRange
+       return (singletonE thing ntName)
+
+instance BindsNames (InModule (ParameterType PName)) where
+  namingEnv (InModule ns ParameterType { .. }) = BuildNamingEnv $
+    -- XXX: we don't seem to have a fixity environment at the type level
+    do let Located { .. } = ptName
+       ntName <- newTop ns thing Nothing srcRange
+       return (singletonT thing ntName)
+
 -- NOTE: we use the same name at the type and expression level, as there's only
 -- ever one name introduced in the declaration. The names are only ever used in
 -- different namespaces, so there's no ambiguity.
 instance BindsNames (InModule (Newtype PName)) where
   namingEnv (InModule ns Newtype { .. }) = BuildNamingEnv $
     do let Located { .. } = nName
-       ntName <- liftSupply (mkDeclared ns (getIdent thing) Nothing srcRange)
+       ntName <- newTop ns thing Nothing srcRange
        return (singletonT thing ntName `mappend` singletonE thing ntName)
 
 -- | The naming environment for a single declaration.
@@ -316,17 +379,17 @@ instance BindsNames (InModule (Decl PName)) where
       do n <- mkName (bName b) (bFixity b)
          return (singletonE (thing (bName b)) n `mappend` fixity n b)
 
-    DSignature ns _sig    -> foldMap qualBind ns
-    DPragma ns _p         -> foldMap qualBind ns
-    DType (TySyn lqn _ _) -> qualType lqn
-    DLocated d' _         -> namingEnv (InModule pfx d')
-    DPatBind _pat _e      -> panic "ModuleSystem" ["Unexpected pattern binding"]
-    DFixity{}             -> panic "ModuleSystem" ["Unexpected fixity declaration"]
+    DSignature ns _sig      -> foldMap qualBind ns
+    DPragma ns _p           -> foldMap qualBind ns
+    DType (TySyn lqn _ _)   -> qualType lqn
+    DProp (PropSyn lqn _ _) -> qualType lqn
+    DLocated d' _           -> namingEnv (InModule pfx d')
+    DPatBind _pat _e        -> panic "ModuleSystem" ["Unexpected pattern binding"]
+    DFixity{}               -> panic "ModuleSystem" ["Unexpected fixity declaration"]
 
     where
 
-    mkName ln fx =
-      liftSupply (mkDeclared pfx (getIdent (thing ln)) fx (srcRange ln))
+    mkName ln fx = newTop pfx (thing ln) fx (srcRange ln)
 
     qualBind ln = BuildNamingEnv $
       do n <- mkName ln Nothing
