@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.ModuleSystem.Renamer
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -29,16 +29,21 @@ module Cryptol.ModuleSystem.Renamer (
 
 import Cryptol.ModuleSystem.Name
 import Cryptol.ModuleSystem.NamingEnv
+import Cryptol.ModuleSystem.Exports
 import Cryptol.Prims.Syntax
 import Cryptol.Parser.AST
 import Cryptol.Parser.Position
-import Cryptol.Utils.Ident (packIdent,packInfix)
+import Cryptol.TypeCheck.Type (TCon(..))
+import Cryptol.Utils.Ident (packInfix,packIdent)
 import Cryptol.Utils.Panic (panic)
 import Cryptol.Utils.PP
 
 import qualified Data.Foldable as F
+import           Data.Map.Strict ( Map )
 import qualified Data.Map.Strict as Map
 import qualified Data.Sequence as Seq
+import qualified Data.Semigroup as S
+import qualified Data.Set as Set
 import           Data.String (IsString(..))
 import           MonadLib hiding (mapM, mapM_)
 
@@ -109,7 +114,7 @@ instance PP RenamerError where
       hang (text "[error] at" <+> pp (srcRange lqn))
          4 (fsep [ text "Expected a value named", quotes (pp (thing lqn))
                  , text "but found a type instead"
-                 , text "Did you mean `(" <> pp (thing lqn) <> text")?" ])
+                 , text "Did you mean `(" <.> pp (thing lqn) <.> text")?" ])
 
     ExpectedType lqn disp -> fixNameDisp disp $
       hang (text "[error] at" <+> pp (srcRange lqn))
@@ -144,13 +149,15 @@ data RenamerWarning
     -- Warn when fixity is used to resolve parses, and the relative
     -- fixity is planned to change.  See https://github.com/GaloisInc/cryptol/issues/241
   | DangerousFixity (Located Name) (Located Name) NameDisp
+
+  | UnusedName Name NameDisp
     deriving (Show, Generic, NFData)
 
 instance PP RenamerWarning where
   ppPrec _ (SymbolShadowed new originals disp) = fixNameDisp disp $
     hang (text "[warning] at" <+> loc)
        4 $ fsep [ text "This binding for" <+> sym
-                , (text "shadows the existing binding" <> plural) <+> text "from" ]
+                , (text "shadows the existing binding" <.> plural) <+> text "from" ]
         $$ vcat (map ppLocName originals)
 
     where
@@ -162,10 +169,15 @@ instance PP RenamerWarning where
 
   ppPrec _ (DangerousFixity o1 o2 disp) = fixNameDisp disp $
     hang (text "[warning] at" <+> pp (srcRange o1))
-       4 $ fsep [ text "Using fixity to resolve the parsing of operators" <+> pp (thing o1) <+> text "and" <+> pp (thing o2) <> text ";"
-                , text "the relative fixity of these operators is planned to change in a future Cryptol release."
+       4 $ fsep [ text "Using fixity to resolve the parsing of operators" <+> pp (thing o1) <+> text "and" <+> pp (thing o2) <.> text ";"
+                , text "the relative fixity of these operators has recently changed!"
                 , text "Use parentheses to disambiguate this parse, or consider replacing (&&) with (/\\), or (||) with (\\/)."
+                , text "Ignore this message if you are confident this expression is parsing correctly; it will be removed"
+                , text "in a future release."
                 ]
+  ppPrec _ (UnusedName x disp) = fixNameDisp disp $
+    hang (text "[warning] at" <+> pp (nameLoc x))
+       4 (text "Unused name:" <+> pp x)
 
 -- Renaming Monad --------------------------------------------------------------
 
@@ -177,23 +189,30 @@ data RO = RO
   }
 
 data RW = RW
-  { rwWarnings :: !(Seq.Seq RenamerWarning)
-  , rwErrors   :: !(Seq.Seq RenamerError)
-  , rwSupply   :: !Supply
+  { rwWarnings      :: !(Seq.Seq RenamerWarning)
+  , rwErrors        :: !(Seq.Seq RenamerError)
+  , rwSupply        :: !Supply
+  , rwNameUseCount  :: !(Map Name Int)
+    -- ^ How many times did we refer to each name.
+    -- Used to generate warnings for unused definitions.
   }
 
 newtype RenameM a = RenameM
   { unRenameM :: ReaderT RO (StateT RW Lift) a }
 
-instance Monoid a => Monoid (RenameM a) where
+instance S.Semigroup a => S.Semigroup (RenameM a) where
+  {-# INLINE (<>) #-}
+  a <> b =
+    do x <- a
+       y <- b
+       return (x S.<> y)
+
+instance (S.Semigroup a, Monoid a) => Monoid (RenameM a) where
   {-# INLINE mempty #-}
   mempty = return mempty
 
   {-# INLINE mappend #-}
-  mappend a b =
-    do x <- a
-       y <- b
-       return (mappend x y)
+  mappend = (S.<>)
 
 instance Functor RenameM where
   {-# INLINE fmap #-}
@@ -221,43 +240,49 @@ instance FreshM RenameM where
 
 runRenamer :: Supply -> ModName -> NamingEnv -> RenameM a
            -> (Either [RenamerError] (a,Supply),[RenamerWarning])
-runRenamer s ns env m = (res,F.toList (rwWarnings rw))
+runRenamer s ns env m = (res, warnUnused ns env ro rw ++ F.toList (rwWarnings rw))
   where
-
-  (a,rw) = runM (unRenameM m) RO { roLoc = emptyRange
-                                 , roNames = env
-                                 , roMod = ns
-                                 , roDisp = neverQualifyMod ns
-                                            `mappend` toNameDisp env
-                                 }
+  (a,rw) = runM (unRenameM m) ro
                               RW { rwErrors   = Seq.empty
                                  , rwWarnings = Seq.empty
                                  , rwSupply   = s
+                                 , rwNameUseCount = Map.empty
                                  }
 
+  ro = RO { roLoc = emptyRange
+          , roNames = env
+          , roMod = ns
+          , roDisp = neverQualifyMod ns `mappend` toNameDisp env
+          }
+
   res | Seq.null (rwErrors rw) = Right (a,rwSupply rw)
       | otherwise              = Left (F.toList (rwErrors rw))
 
+-- | Record an error.  XXX: use a better name
 record :: (NameDisp -> RenamerError) -> RenameM ()
 record f = RenameM $
   do RO { .. } <- ask
      RW { .. } <- get
      set RW { rwErrors = rwErrors Seq.|> f roDisp, .. }
 
+-- | Report a warning.
 recordW :: (NameDisp -> RenamerWarning) -> RenameM ()
 recordW f = RenameM $
   do RO { .. } <- ask
      RW { .. } <- get
      set RW { rwWarnings = rwWarnings Seq.|> f roDisp, .. }
 
+-- | Get the source range for wahtever we are currently renaming.
 curLoc :: RenameM Range
 curLoc  = RenameM (roLoc `fmap` ask)
 
+-- | Annotate something with the current range.
 located :: a -> RenameM (Located a)
 located thing =
   do srcRange <- curLoc
      return Located { .. }
 
+-- | Do the given computation using the source code range from `loc` if any.
 withLoc :: HasLoc loc => loc -> RenameM a -> RenameM a
 withLoc loc m = RenameM $ case getLoc loc of
 
@@ -342,6 +367,25 @@ checkNamingEnv env = (F.toList out, [])
 
   check ns acc = containsOverlap disp ns Seq.>< acc
 
+recordUse :: Name -> RenameM ()
+recordUse x = RenameM $ sets_ $ \rw ->
+  rw { rwNameUseCount = Map.insertWith (+) x 1 (rwNameUseCount rw) }
+
+
+warnUnused :: ModName -> NamingEnv -> RO -> RW -> [RenamerWarning]
+warnUnused m0 env ro rw =
+  map warn
+  $ Map.keys
+  $ Map.filterWithKey keep
+  $ rwNameUseCount rw
+  where
+  warn x   = UnusedName x (roDisp ro)
+  keep k n = n == 1 && isLocal k
+  oldNames = fst (visibleNames env)
+  isLocal nm = case nameInfo nm of
+                 Declared m sys -> sys == UserName &&
+                                   m == m0 && nm `Set.notMember` oldNames
+                 Parameter  -> True
 
 -- Renaming --------------------------------------------------------------------
 
@@ -353,13 +397,36 @@ renameModule m =
   do env    <- liftSupply (namingEnv' m)
      -- NOTE: we explicitly hide shadowing errors here, by using shadowNames'
      decls' <-  shadowNames' CheckOverlap env (traverse rename (mDecls m))
-     return (env,m { mDecls = decls' })
+     let m1 = m { mDecls = decls' }
+         exports = modExports m1
+     mapM_ recordUse (eTypes exports)
+     return (env,m1)
 
 instance Rename TopDecl where
   rename td     = case td of
     Decl d      -> Decl      <$> traverse rename d
     TDNewtype n -> TDNewtype <$> traverse rename n
     Include n   -> return (Include n)
+    DParameterFun f  -> DParameterFun  <$> rename f
+    DParameterType f -> DParameterType <$> rename f
+
+    DParameterConstraint d -> DParameterConstraint <$> mapM renameLocated d
+
+renameLocated :: Rename f => Located (f PName) -> RenameM (Located (f Name))
+renameLocated x =
+  do y <- rename (thing x)
+     return x { thing = y }
+
+instance Rename ParameterType where
+  rename a =
+    do n' <- rnLocated renameType (ptName a)
+       return a { ptName = n' }
+
+instance Rename ParameterFun where
+  rename a =
+    do n'   <- rnLocated renameVar (pfName a)
+       sig' <- renameSchema (pfSchema a)
+       return a { pfName = n', pfSchema = snd sig' }
 
 rnLocated :: (a -> RenameM b) -> Located a -> RenameM (Located b)
 rnLocated f loc = withLoc loc $
@@ -379,6 +446,7 @@ instance Rename Decl where
                             shadowNames pe (DPatBind pat' <$> rename e)
 
     DType syn         -> DType         <$> rename syn
+    DProp syn         -> DProp         <$> rename syn
     DLocated d' r     -> withLoc r
                        $ DLocated      <$> rename d'  <*> pure r
     DFixity{}         -> panic "Renamer" ["Unexpected fixity declaration"
@@ -426,9 +494,10 @@ typeExists :: PName -> RenameM (Maybe Name)
 typeExists pn =
   do ro <- RenameM ask
      case Map.lookup pn (neTypes (roNames ro)) of
-       Just [n]  -> return (Just n)
+       Just [n]  -> recordUse n >> return (Just n)
        Just []   -> panic "Renamer" ["Invalid type renaming environment"]
        Just syms -> do n <- located pn
+                       mapM_ recordUse syms
                        record (MultipleSyms n syms)
                        return (Just (head syms))
        Nothing -> return Nothing
@@ -491,11 +560,17 @@ instance Rename TParam where
 
 instance Rename Prop where
   rename p      = case p of
-    CFin t        -> CFin     <$> rename t
-    CEqual l r    -> CEqual   <$> rename l  <*> rename r
-    CGeq l r      -> CGeq     <$> rename l  <*> rename r
-    CArith t      -> CArith   <$> rename t
-    CCmp t        -> CCmp     <$> rename t
+    CFin t        -> CFin       <$> rename t
+    CEqual l r    -> CEqual     <$> rename l <*> rename r
+    CNeq l r      -> CNeq       <$> rename l <*> rename r
+    CGeq l r      -> CGeq       <$> rename l <*> rename r
+    CZero t       -> CZero      <$> rename t
+    CLogic t      -> CLogic     <$> rename t
+    CArith t      -> CArith     <$> rename t
+    CCmp t        -> CCmp       <$> rename t
+    CSignedCmp t  -> CSignedCmp <$> rename t
+    CLiteral l r  -> CLiteral   <$> rename l <*> rename r
+    CUser qn ps   -> CUser      <$> renameType qn <*> traverse rename ps
     CLocated p' r -> withLoc r
                    $ CLocated <$> rename p' <*> pure r
 
@@ -510,13 +585,15 @@ translateProp ty = go ty
 
     TLocated t' r -> (`CLocated` r) <$> go t'
 
-    -- these are the only cases that will produce valid props.
+    TApp (PC x) [l,r]
+      | PEqual <- x -> CEqual <$> rename l <*> rename r
+      | PNeq   <- x -> CNeq   <$> rename l <*> rename r
+      | PGeq   <- x -> CGeq   <$> rename l <*> rename r
+
     TUser n [l,r]
-      | i == packIdent "==" -> CEqual <$> rename l <*> rename r
-      | i == packIdent ">=" -> CGeq   <$> rename l <*> rename r
-      | i == packIdent "<=" -> CGeq   <$> rename r <*> rename l
-      where
-      i = getIdent n
+      | isLeq n -> CGeq <$> rename r <*> rename l
+
+    TUser n ts -> CUser <$> renameType n <*> traverse rename ts
 
     -- record an error, but continue renaming to gather any other errors
     _ ->
@@ -526,10 +603,10 @@ translateProp ty = go ty
 
 -- | Check to see if this identifier is a reserved type/type-function.
 isReserved :: PName -> Bool
-isReserved pn = Map.member pn tfunNames || isReservedTyCon pn
+isReserved pn = case primTyFromPName pn of
+                  Just _ -> True
+                  _      -> False
 
-isReservedTyCon :: PName -> Bool
-isReservedTyCon pn = Map.member pn tconNames
 
 -- | Resolve fixity, then rename the resulting type.
 instance Rename Type where
@@ -541,19 +618,12 @@ instance Rename Type where
     go  TBit         = return TBit
     go (TNum c)      = return (TNum c)
     go (TChar c)     = return (TChar c)
-    go  TInf         = return TInf
 
     go (TUser pn ps)
 
-      -- all type functions
-      | Just (arity,fun) <- Map.lookup pn tfunNames =
-        do when (arity /= length ps) (record (MalformedBuiltin ty0 pn))
-           ps' <- traverse go ps
-           return (TApp fun ps')
-
-      -- built-in types like Bit and inf
-      | Just ty <- Map.lookup pn tconNames =
-        rename ty
+      | Just pt <- primTyFromPName pn =
+        do ps' <- traverse go ps
+           return (TApp (primTyCon pt) ps')
 
     go (TUser qn ps)   = TUser    <$> renameType qn <*> traverse go ps
     go (TApp f xs)     = TApp f   <$> traverse go xs
@@ -596,48 +666,36 @@ resolveTypeFixity  = go
     TBit         -> return t
     TNum _       -> return t
     TChar _      -> return t
-    TInf         -> return t
     TWild        -> return t
 
 
 type TOp = Type PName -> Type PName -> Type PName
 
-infixProps :: [PName]
-infixProps  = map (mkUnqual . packInfix) [ "==", ">=", "<=" ]
-
 mkTInfix :: Type PName -> (TOp,Fixity) -> Type PName -> RenameM (Type PName)
 
--- only if the function is one of props
-mkTInfix t@(TUser o1 [x,y]) op@(o2,f2) z
-  | o1 `elem` infixProps =
-    do let f1 = Fixity NonAssoc 0
-       case compareFixity f1 f2 of
-         FCLeft  -> return (o2 t z)
-         FCRight -> do r <- mkTInfix y op z
-                       return (TUser o1 [x,r])
+mkTInfix t op@(o2,f2) z =
+  case t of
+    TLocated t1 _ -> mkTInfix t1 op z
 
-         -- Just reconstruct with the TUser part being an application. If this was
-         -- a real error, it will be caught during renaming.
-         FCError -> return (o2 t z)
+    TUser op1 [x,y] | isLeq op1 -> doFixity (TUser op1) leqFixity x y
+    TApp tc [x,y]
+      | Just pt <- primTyFromTC tc
+      , Just f1 <- primTyFixity pt -> doFixity (TApp tc) f1 x y
 
--- In this case, we know the fixities of both sides.
-mkTInfix t@(TApp o1 [x,y]) op@(o2,f2) z
-  | Just (a1,p1) <- Map.lookup o1 tBinOpPrec =
-     case compareFixity (Fixity a1 p1) f2 of
-       FCLeft  -> return (o2 t z)
-       FCRight -> do r <- mkTInfix y op z
-                     return (TApp o1 [x,r])
+    _ -> return (o2 t z)
 
-       -- As the fixity table is known, and this is a case where the fixity came
-       -- from that table, it's a real error if the fixities didn't work out.
-       FCError -> panic "Renamer" [ "fixity problem for type operators"
-                                  , show (o2 t z) ]
+  where
+  doFixity mk f1 x y =
+    case compareFixity f1 f2 of
+      FCLeft  -> return (o2 t z)
+      FCRight -> do r <- mkTInfix y op z
+                    return (mk [x,r])
 
-mkTInfix (TLocated t _) op z =
-     mkTInfix t op z
+      -- As the fixity table is known, and this is a case where the fixity came
+      -- from that table, it's a real error if the fixities didn't work out.
+      FCError -> panic "Renamer" [ "fixity problem for type operators"
+                                 , show (o2 t z) ]
 
-mkTInfix t (op,_) z =
-     return (op t z)
 
 
 -- | When possible, rewrite the type operator to a known constructor, otherwise
@@ -645,7 +703,7 @@ mkTInfix t (op,_) z =
 lookupFixity :: Located PName -> (TOp,Fixity)
 lookupFixity op =
   case lkp of
-    Just (p,f) -> (\x y -> TApp p [x,y], f)
+    Just res -> res
 
     -- unknown type operator, just use default fixity
     -- NOTE: this works for the props defined above, as all other operators
@@ -654,9 +712,22 @@ lookupFixity op =
 
   where
   sym = thing op
-  lkp = do (_,n)           <- Map.lookup (thing op) tfunNames
-           (fAssoc,fLevel) <- Map.lookup n tBinOpPrec
-           return (n,Fixity { .. })
+  lkp = do pt <- primTyFromPName (thing op)
+           fi <- primTyFixity pt
+           return (\x y -> TApp (primTyCon pt) [x,y], fi)
+        `mplus`
+        do guard (isLeq sym)
+           return (\x y -> TUser sym [x,y], leqFixity)
+
+leqFixity :: Fixity
+leqFixity = Fixity NonAssoc 30
+
+leqIdent :: Ident
+leqIdent  = packInfix "<="
+
+isLeq :: PName -> Bool
+isLeq x = getIdent x == leqIdent
+
 
 -- | Rename a binding.
 instance Rename Bind where
@@ -693,9 +764,11 @@ instance Rename Pattern where
                      $ PLocated <$> rename p'    <*> pure loc
 
 instance Rename Expr where
-  rename e = case e of
+  rename expr = case expr of
     EVar n        -> EVar <$> renameVar n
     ELit l        -> return (ELit l)
+    ENeg e        -> rename (EApp (EVar (mkUnqual (packIdent "negate"))) e)
+    EComplement e -> rename (EApp (EVar (mkUnqual (packIdent "complement"))) e)
     ETuple es     -> ETuple  <$> traverse rename es
     ERecord fs    -> ERecord <$> traverse (rnNamed rename) fs
     ESel e' s     -> ESel    <$> rename e' <*> pure s
@@ -883,7 +956,6 @@ patternEnv  = go
   typeEnv TBit       = return mempty
   typeEnv TNum{}     = return mempty
   typeEnv TChar{}    = return mempty
-  typeEnv TInf       = return mempty
 
   typeEnv (TUser pn ps) =
     do mb <- typeExists pn
@@ -942,6 +1014,15 @@ instance Rename TySyn where
                               <*> traverse rename ps
                               <*> rename ty
 
+instance Rename PropSyn where
+  rename (PropSyn n ps cs) =
+    do when (isReserved (thing n))
+            (record (BoundReservedType (thing n) (getLoc n) (text "constraint synonym")))
+
+       shadowNames ps $ PropSyn <$> rnLocated renameType n
+                                <*> traverse rename ps
+                                <*> traverse rename cs
+
 
 -- Utilities -------------------------------------------------------------------
 
