@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Parser.AST
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -15,6 +15,7 @@
 {-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE OverloadedStrings #-}
 module Cryptol.Parser.AST
   ( -- * Names
     Ident, mkIdent, mkInfix, isInfixIdent, nullIdent, identText
@@ -28,7 +29,7 @@ module Cryptol.Parser.AST
   , Schema(..)
   , TParam(..)
   , Kind(..)
-  , Type(..), tconNames
+  , Type(..)
   , Prop(..)
 
     -- * Declarations
@@ -39,15 +40,16 @@ module Cryptol.Parser.AST
   , Fixity(..), defaultFixity
   , FixityCmp(..), compareFixity
   , TySyn(..)
+  , PropSyn(..)
   , Bind(..)
   , BindDef(..), LBindDef
   , Pragma(..)
   , ExportType(..)
-  , ExportSpec(..), exportBind, exportType
-  , isExportedBind, isExportedType
   , TopLevel(..)
   , Import(..), ImportSpec(..)
   , Newtype(..)
+  , ParameterType(..)
+  , ParameterFun(..)
 
     -- * Interactive
   , ReplInput(..)
@@ -69,18 +71,17 @@ module Cryptol.Parser.AST
   , cppKind, ppSelector
   ) where
 
+import Cryptol.Parser.Fixity
 import Cryptol.Parser.Name
 import Cryptol.Parser.Position
-import Cryptol.Prims.Syntax (TFun(..))
+import Cryptol.Parser.Selector
+import Cryptol.TypeCheck.Type (TCon(..))
 import Cryptol.Utils.Ident
 import Cryptol.Utils.PP
-import Cryptol.Utils.Panic (panic)
 
-import qualified Data.Set as Set
 import           Data.List(intersperse)
 import           Data.Bits(shiftR)
 import           Data.Maybe (catMaybes)
-import qualified Data.Map as Map
 import           Numeric(showIntAtBase)
 
 import GHC.Generics (Generic)
@@ -104,10 +105,14 @@ type LString  = Located String
 newtype Program name = Program [TopDecl name]
                        deriving (Show)
 
-data Module name = Module { mName    :: Located ModName
-                          , mImports :: [Located Import]
-                          , mDecls   :: [TopDecl name]
-                          } deriving (Show, Generic, NFData)
+-- | A parsed module.
+data Module name = Module
+  { mName     :: Located ModName            -- ^ Name of the module
+  , mInstance :: !(Maybe (Located ModName)) -- ^ Functor to instantiate
+                                            -- (if this is a functor instnaces)
+  , mImports  :: [Located Import]           -- ^ Imports for the module
+  , mDecls    :: [TopDecl name]             -- ^ Declartions for the module
+  } deriving (Show, Generic, NFData)
 
 
 modRange :: Module name -> Range
@@ -119,9 +124,14 @@ modRange m = rCombs $ catMaybes
     ]
 
 
-data TopDecl name = Decl (TopLevel (Decl name))
-                  | TDNewtype (TopLevel (Newtype name))
-                  | Include (Located FilePath)
+data TopDecl name =
+    Decl (TopLevel (Decl name))
+  | TDNewtype (TopLevel (Newtype name)) -- ^ @newtype T as = t
+  | Include (Located FilePath)          -- ^ @include File@
+  | DParameterType (ParameterType name) -- ^ @parameter type T : #@
+  | DParameterConstraint [Located (Prop name)]
+                                        -- ^ @parameter type constraint (fin T)@
+  | DParameterFun  (ParameterFun name)  -- ^ @parameter someVal : [256]@
                     deriving (Show, Generic, NFData)
 
 data Decl name = DSignature [Located name] (Schema name)
@@ -130,8 +140,28 @@ data Decl name = DSignature [Located name] (Schema name)
                | DBind (Bind name)
                | DPatBind (Pattern name) (Expr name)
                | DType (TySyn name)
+               | DProp (PropSyn name)
                | DLocated (Decl name) Range
-                 deriving (Eq, Show, Generic, NFData)
+                 deriving (Eq, Show, Generic, NFData, Functor)
+
+
+-- | A type parameter
+data ParameterType name = ParameterType
+  { ptName    :: Located name     -- ^ name of type parameter
+  , ptKind    :: Kind             -- ^ kind of parameter
+  , ptDoc     :: Maybe String     -- ^ optional documentation
+  , ptFixity  :: Maybe Fixity     -- ^ info for infix use
+  , ptNumber  :: !Int             -- ^ number of the parameter
+  } deriving (Eq,Show,Generic,NFData)
+
+-- | A value parameter
+data ParameterFun name = ParameterFun
+  { pfName   :: Located name      -- ^ name of value parameter
+  , pfSchema :: Schema name       -- ^ schema for parameter
+  , pfDoc    :: Maybe String      -- ^ optional documentation
+  , pfFixity :: Maybe Fixity      -- ^ info for infix use
+  } deriving (Eq,Show,Generic,NFData)
+
 
 -- | An import declaration.
 data Import = Import { iModule    :: !ModName
@@ -149,7 +179,10 @@ data ImportSpec = Hiding [Ident]
                   deriving (Eq, Show, Generic, NFData)
 
 data TySyn n  = TySyn (Located n) [TParam n] (Type n)
-                deriving (Eq, Show, Generic, NFData)
+                deriving (Eq, Show, Generic, NFData, Functor)
+
+data PropSyn n = PropSyn (Located n) [TParam n] [Prop n]
+                 deriving (Eq, Show, Generic, NFData, Functor)
 
 {- | Bindings.  Notes:
 
@@ -163,45 +196,23 @@ data TySyn n  = TySyn (Located n) [TParam n] (Type n)
       by the type checker.  However, they are useful when de-sugaring
       patterns.
 -}
-data Bind name = Bind { bName      :: Located name -- ^ Defined thing
-                      , bParams    :: [Pattern name]-- ^ Parameters
-                      , bDef       :: Located (BindDef name) -- ^ Definition
-                      , bSignature :: Maybe (Schema name) -- ^ Optional type sig
-                      , bInfix     :: Bool         -- ^ Infix operator?
-                      , bFixity    :: Maybe Fixity -- ^ Optional fixity info
-                      , bPragmas   :: [Pragma]     -- ^ Optional pragmas
-                      , bMono      :: Bool         -- ^ Is this a monomorphic binding
-                      , bDoc       :: Maybe String -- ^ Optional doc string
-                      } deriving (Eq, Show, Generic, NFData)
+data Bind name = Bind
+  { bName      :: Located name            -- ^ Defined thing
+  , bParams    :: [Pattern name]          -- ^ Parameters
+  , bDef       :: Located (BindDef name)  -- ^ Definition
+  , bSignature :: Maybe (Schema name)     -- ^ Optional type sig
+  , bInfix     :: Bool                    -- ^ Infix operator?
+  , bFixity    :: Maybe Fixity            -- ^ Optional fixity info
+  , bPragmas   :: [Pragma]                -- ^ Optional pragmas
+  , bMono      :: Bool                    -- ^ Is this a monomorphic binding
+  , bDoc       :: Maybe String            -- ^ Optional doc string
+  } deriving (Eq, Generic, NFData, Functor, Show)
 
 type LBindDef = Located (BindDef PName)
 
 data BindDef name = DPrim
                   | DExpr (Expr name)
-                    deriving (Eq, Show, Generic, NFData)
-
-data Fixity   = Fixity { fAssoc :: !Assoc
-                       , fLevel :: !Int
-                       } deriving (Eq, Show, Generic, NFData)
-
-data FixityCmp = FCError
-               | FCLeft
-               | FCRight
-                 deriving (Show,Eq)
-
-compareFixity :: Fixity -> Fixity -> FixityCmp
-compareFixity (Fixity a1 p1) (Fixity a2 p2) =
-  case compare p1 p2 of
-    GT -> FCLeft
-    LT -> FCRight
-    EQ -> case (a1,a2) of
-            (LeftAssoc,LeftAssoc)   -> FCLeft
-            (RightAssoc,RightAssoc) -> FCRight
-            _                       -> FCError
-
--- | The fixity used when none is provided.
-defaultFixity :: Fixity
-defaultFixity  = Fixity LeftAssoc 100
+                    deriving (Eq, Show, Generic, NFData, Functor)
 
 data Pragma   = PragmaNote String
               | PragmaProperty
@@ -223,41 +234,13 @@ data ExportType = Public
                 | Private
                   deriving (Eq, Show, Ord, Generic, NFData)
 
+-- | A top-level module declaration.
 data TopLevel a = TopLevel { tlExport :: ExportType
                            , tlDoc    :: Maybe (Located String)
                            , tlValue  :: a
                            }
   deriving (Show, Generic, NFData, Functor, Foldable, Traversable)
 
-data ExportSpec name = ExportSpec { eTypes  :: Set.Set name
-                                  , eBinds  :: Set.Set name
-                                  } deriving (Show, Generic, NFData)
-
-instance Ord name => Monoid (ExportSpec name) where
-  mempty      = ExportSpec { eTypes = mempty, eBinds = mempty }
-  mappend l r = ExportSpec { eTypes = mappend (eTypes l) (eTypes r)
-                           , eBinds  = mappend (eBinds  l) (eBinds  r)
-                           }
-
--- | Add a binding name to the export list, if it should be exported.
-exportBind :: Ord name => TopLevel name -> ExportSpec name
-exportBind n
-  | tlExport n == Public = mempty { eBinds = Set.singleton (tlValue n) }
-  | otherwise            = mempty
-
--- | Check to see if a binding is exported.
-isExportedBind :: Ord name => name -> ExportSpec name -> Bool
-isExportedBind n = Set.member n . eBinds
-
--- | Add a type synonym name to the export list, if it should be exported.
-exportType :: Ord name => TopLevel name -> ExportSpec name
-exportType n
-  | tlExport n == Public = mempty { eTypes = Set.singleton (tlValue n) }
-  | otherwise            = mempty
-
--- | Check to see if a type synonym is exported.
-isExportedType :: Ord name => name -> ExportSpec name -> Bool
-isExportedType n = Set.member n . eTypes
 
 -- | Infromation about the representation of a numeric constant.
 data NumInfo  = BinLit Int                      -- ^ n-digit binary literal
@@ -275,6 +258,8 @@ data Literal  = ECNum Integer NumInfo           -- ^ @0x10@  (HexLit 2)
 
 data Expr n   = EVar n                          -- ^ @ x @
               | ELit Literal                    -- ^ @ 0x10 @
+              | ENeg (Expr n)                   -- ^ @ -1 @
+              | EComplement (Expr n)            -- ^ @ ~1 @
               | ETuple [Expr n]                 -- ^ @ (1,2,3) @
               | ERecord [Named (Expr n)]        -- ^ @ { x = 1, y = 2 } @
               | ESel (Expr n) Selector          -- ^ @ e.l @
@@ -293,34 +278,15 @@ data Expr n   = EVar n                          -- ^ @ x @
 
               | EParens (Expr n)                -- ^ @ (e)   @ (Removed by Fixity)
               | EInfix (Expr n) (Located n) Fixity (Expr n)-- ^ @ a + b @ (Removed by Fixity)
-                deriving (Eq, Show, Generic, NFData)
+                deriving (Eq, Show, Generic, NFData, Functor)
 
 data TypeInst name = NamedInst (Named (Type name))
                    | PosInst (Type name)
-                     deriving (Eq, Show, Generic, NFData)
-
-{- | Selectors are used for projecting from various components.
-Each selector has an option spec to specify the shape of the thing
-that is being selected.  Currently, there is no surface syntax for
-list selectors, but they are used during the desugaring of patterns.
--}
-
-data Selector = TupleSel Int   (Maybe Int)
-                -- ^ Zero-based tuple selection.
-                -- Optionally specifies the shape of the tuple (one-based).
-
-              | RecordSel Ident (Maybe [Ident])
-                -- ^ Record selection.
-                -- Optionally specifies the shape of the record.
-
-              | ListSel Int    (Maybe Int)
-                -- ^ List selection.
-                -- Optionally specifies the length of the list.
-                deriving (Eq, Show, Ord, Generic, NFData)
+                     deriving (Eq, Show, Generic, NFData, Functor)
 
 data Match name = Match (Pattern name) (Expr name)              -- ^ p <- e
                 | MatchLet (Bind name)
-                  deriving (Eq, Show, Generic, NFData)
+                  deriving (Eq, Show, Generic, NFData, Functor)
 
 data Pattern n = PVar (Located n)              -- ^ @ x @
                | PWild                         -- ^ @ _ @
@@ -330,53 +296,63 @@ data Pattern n = PVar (Located n)              -- ^ @ x @
                | PTyped (Pattern n) (Type n)   -- ^ @ x : [8] @
                | PSplit (Pattern n) (Pattern n)-- ^ @ (x # y) @
                | PLocated (Pattern n) Range    -- ^ Location information
-                 deriving (Eq, Show, Generic, NFData)
+                 deriving (Eq, Show, Generic, NFData, Functor)
 
 data Named a = Named { name :: Located Ident, value :: a }
   deriving (Eq, Show, Foldable, Traversable, Generic, NFData, Functor)
 
 data Schema n = Forall [TParam n] [Prop n] (Type n) (Maybe Range)
-  deriving (Eq, Show, Generic, NFData)
+  deriving (Eq, Show, Generic, NFData, Functor)
 
-data Kind = KNum | KType
+data Kind = KNum | KType | KFun Kind Kind
   deriving (Eq, Show, Generic, NFData)
 
 data TParam n = TParam { tpName  :: n
                        , tpKind  :: Maybe Kind
                        , tpRange :: Maybe Range
                        }
-  deriving (Eq, Show, Generic, NFData)
+  deriving (Eq, Show, Generic, NFData, Functor)
 
 data Type n = TFun (Type n) (Type n)  -- ^ @[8] -> [8]@
             | TSeq (Type n) (Type n)  -- ^ @[8] a@
             | TBit                    -- ^ @Bit@
             | TNum Integer            -- ^ @10@
             | TChar Char              -- ^ @'a'@
-            | TInf                    -- ^ @inf@
             | TUser n [Type n]        -- ^ A type variable or synonym
-            | TApp TFun [Type n]      -- ^ @2 + x@
+
+            | TApp TCon [Type n]
+              -- ^ @2 + x@
+              -- Note that the parser never produces these; instead it
+              -- produces a "TUser" value.  The "TApp" is introduced by
+              -- the renamer when it spots built-in functions.
+              -- XXX: We should just add primitive declarations for the
+              -- built-in type functions, and simplify all this.
+
+
             | TRecord [Named (Type n)]-- ^ @{ x : [8], y : [32] }@
             | TTuple [Type n]         -- ^ @([8], [32])@
             | TWild                   -- ^ @_@, just some type.
             | TLocated (Type n) Range -- ^ Location information
             | TParens (Type n)        -- ^ @ (ty) @
             | TInfix (Type n) (Located n) Fixity (Type n) -- ^ @ ty + ty @
-              deriving (Eq, Show, Generic, NFData)
+              deriving (Eq, Show, Generic, NFData, Functor)
 
-tconNames :: Map.Map PName (Type PName)
-tconNames  = Map.fromList
-  [ (mkUnqual (packIdent "Bit"), TBit)
-  , (mkUnqual (packIdent "inf"), TInf)
-  ]
 
 data Prop n   = CFin (Type n)             -- ^ @ fin x   @
               | CEqual (Type n) (Type n)  -- ^ @ x == 10 @
+              | CNeq (Type n) (Type n)    -- ^ @ x != 10 @
               | CGeq (Type n) (Type n)    -- ^ @ x >= 10 @
+              | CZero (Type n)            -- ^ @ Zero a  @
+              | CLogic (Type n)           -- ^ @ Logic a @
               | CArith (Type n)           -- ^ @ Arith a @
               | CCmp (Type n)             -- ^ @ Cmp a @
+              | CSignedCmp (Type n)       -- ^ @ SignedCmp a @
+              | CLiteral (Type n) (Type n)-- ^ @ Literal val a @
+              | CUser n [Type n]          -- ^ Constraint synonym
               | CLocated (Prop n) Range   -- ^ Location information
               | CType (Type n)            -- ^ After parsing
-                deriving (Eq, Show, Generic, NFData)
+                deriving (Eq, Show, Generic, NFData, Functor)
+
 
 --------------------------------------------------------------------------------
 -- Note: When an explicit location is missing, we could use the sub-components
@@ -467,6 +443,15 @@ instance HasLoc (TopDecl name) where
     Decl tld    -> getLoc tld
     TDNewtype n -> getLoc n
     Include lfp -> getLoc lfp
+    DParameterType d -> getLoc d
+    DParameterFun d  -> getLoc d
+    DParameterConstraint d -> getLoc d
+
+instance HasLoc (ParameterType name) where
+  getLoc a = getLoc (ptName a)
+
+instance HasLoc (ParameterFun name) where
+  getLoc a = getLoc (pfName a)
 
 instance HasLoc (Module name) where
   getLoc m
@@ -517,6 +502,23 @@ instance (Show name, PPName name) => PP (TopDecl name) where
       Decl    d   -> pp d
       TDNewtype n -> pp n
       Include l   -> text "include" <+> text (show (thing l))
+      DParameterFun d -> pp d
+      DParameterType d -> pp d
+      DParameterConstraint d ->
+        "parameter" <+> "type" <+> "constraint" <+> prop
+        where prop = case map pp d of
+                       [x] -> x
+                       []  -> "()"
+                       xs  -> parens (hsep (punctuate comma xs))
+
+instance (Show name, PPName name) => PP (ParameterType name) where
+  ppPrec _ a = text "parameter" <+> text "type" <+>
+               ppPrefixName (ptName a) <+> text ":" <+> pp (ptKind a)
+
+instance (Show name, PPName name) => PP (ParameterFun name) where
+  ppPrec _ a = text "parameter" <+> ppPrefixName (pfName a) <+> text ":"
+                  <+> pp (pfSchema a)
+
 
 instance (Show name, PPName name) => PP (Decl name) where
   ppPrec n decl =
@@ -527,6 +529,7 @@ instance (Show name, PPName name) => PP (Decl name) where
       DFixity f ns    -> ppFixity f ns
       DPragma xs p    -> ppPragma xs p
       DType ts        -> ppPrec n ts
+      DProp ps        -> ppPrec n ps
       DLocated d _    -> ppPrec n d
 
 ppFixity :: PPName name => Fixity -> [Located name] -> Doc
@@ -579,7 +582,8 @@ instance (Show name, PPName name) => PP (Bind name) where
 
           lhsOp = case bParams b of
                     [x,y] -> pp x <+> ppL f <+> pp y
-                    _     -> panic "AST" [ "Malformed infix operator", show b ]
+                    xs -> parens (parens (ppL f) <+> fsep (map (ppPrec 0) xs))
+                    -- _     -> panic "AST" [ "Malformed infix operator", show b ]
 
 
 instance (Show name, PPName name) => PP (BindDef name) where
@@ -590,6 +594,12 @@ instance (Show name, PPName name) => PP (BindDef name) where
 instance PPName name => PP (TySyn name) where
   ppPrec _ (TySyn x xs t) = text "type" <+> ppL x <+> fsep (map (ppPrec 1) xs)
                                         <+> text "=" <+> pp t
+
+instance PPName name => PP (PropSyn name) where
+  ppPrec _ (PropSyn x xs ps) =
+    text "constraint" <+> ppL x <+> fsep (map (ppPrec 1) xs)
+                      <+> text "=" <+> parens (commaSep (map pp ps))
+
 instance PP Literal where
   ppPrec _ lit =
     case lit of
@@ -609,7 +619,7 @@ ppNumLit n info =
   where
   pad base pref w =
     let txt = showIntAtBase base ("0123456789abcdef" !!) n ""
-    in text pref <> text (replicate (w - length txt) '0') <> text txt
+    in text pref <.> text (replicate (w - length txt) '0') <.> text txt
 
   poly w = let (res,deg) = bits Nothing [] 0 n
                z | w == 0 = []
@@ -619,10 +629,10 @@ ppNumLit n info =
 
   polyTerm 0 = text "1"
   polyTerm 1 = text "x"
-  polyTerm p = text "x" <> text "^^" <> int p
+  polyTerm p = text "x" <.> text "^^" <.> int p
 
   polyTerm0 0 = text "0"
-  polyTerm0 p = text "0" <> text "*" <> polyTerm p
+  polyTerm0 p = text "0" <.> text "*" <.> polyTerm p
 
   bits d res p num
     | num == 0  = (res,d)
@@ -660,22 +670,26 @@ instance (Show name, PPName name) => PP (Expr name) where
       -- atoms
       EVar x        -> ppPrefixName x
       ELit x        -> pp x
+
+      ENeg x        -> wrap n 3 (text "-" <.> ppPrec 4 x)
+      EComplement x -> wrap n 3 (text "~" <.> ppPrec 4 x)
+
       ETuple es     -> parens (commaSep (map pp es))
       ERecord fs    -> braces (commaSep (map (ppNamed "=") fs))
       EList es      -> brackets (commaSep (map pp es))
-      EFromTo e1 e2 e3 -> brackets (pp e1 <> step <+> text ".." <+> end)
+      EFromTo e1 e2 e3 -> brackets (pp e1 <.> step <+> text ".." <+> end)
         where step = maybe empty (\e -> comma <+> pp e) e2
               end  = maybe empty pp e3
-      EInfFrom e1 e2 -> brackets (pp e1 <> step <+> text "...")
+      EInfFrom e1 e2 -> brackets (pp e1 <.> step <+> text "...")
         where step = maybe empty (\e -> comma <+> pp e) e2
       EComp e mss   -> brackets (pp e <+> vcat (map arm mss))
         where arm ms = text "|" <+> commaSep (map pp ms)
-      ETypeVal t    -> text "`" <> ppPrec 5 t     -- XXX
-      EAppT e ts    -> ppPrec 4 e <> text "`" <> braces (commaSep (map pp ts))
-      ESel    e l   -> ppPrec 4 e <> text "." <> pp l
+      ETypeVal t    -> text "`" <.> ppPrec 5 t     -- XXX
+      EAppT e ts    -> ppPrec 4 e <.> text "`" <.> braces (commaSep (map pp ts))
+      ESel    e l   -> ppPrec 4 e <.> text "." <.> pp l
 
       -- low prec
-      EFun xs e     -> wrap n 0 ((text "\\" <> hsep (map (ppPrec 3) xs)) <+>
+      EFun xs e     -> wrap n 0 ((text "\\" <.> hsep (map (ppPrec 3) xs)) <+>
                                  text "->" <+> pp e)
 
       EIf e1 e2 e3  -> wrap n 0 $ sep [ text "if"   <+> pp e1
@@ -708,30 +722,6 @@ instance (Show name, PPName name) => PP (Expr name) where
      return Infix { .. }
    isInfix _ = Nothing
 
-instance PP Selector where
-  ppPrec _ sel =
-    case sel of
-      TupleSel x sig    -> int x <+> ppSig tupleSig sig
-      RecordSel x sig  -> pp x  <+> ppSig recordSig sig
-      ListSel x sig    -> int x <+> ppSig listSig sig
-
-    where
-    tupleSig n   = int n
-    recordSig xs = braces $ fsep $ punctuate comma $ map pp xs
-    listSig n    = int n
-
-    ppSig f = maybe empty (\x -> text "/* of" <+> f x <+> text "*/")
-
-
--- | Display the thing selected by the selector, nicely.
-ppSelector :: Selector -> Doc
-ppSelector sel =
-  case sel of
-    TupleSel x _  -> ordinal x <+> text "field"
-    RecordSel x _ -> text "field" <+> pp x
-    ListSel x _   -> ordinal x <+> text "element"
-
-
 
 instance PPName name => PP (Pattern name) where
   ppPrec n pat =
@@ -762,11 +752,13 @@ instance PPName name => PP (Schema name) where
 instance PP Kind where
   ppPrec _ KType  = text "*"
   ppPrec _ KNum   = text "#"
+  ppPrec n (KFun k1 k2) = wrap n 1 (ppPrec 1 k1 <+> "->" <+> ppPrec 0 k2)
 
 -- | "Conversational" printing of kinds (e.g., to use in error messages)
 cppKind :: Kind -> Doc
-cppKind KType = text "a value type"
-cppKind KNum  = text "a numeric type"
+cppKind KType     = text "a value type"
+cppKind KNum      = text "a numeric type"
+cppKind (KFun {}) = text "a type-constructor type"
 
 instance PPName name => PP (TParam name) where
   ppPrec n (TParam p Nothing _)   = ppPrec n p
@@ -783,12 +775,11 @@ instance PPName name => PP (Type name) where
       TTuple ts      -> parens $ commaSep $ map pp ts
       TRecord fs     -> braces $ commaSep $ map (ppNamed ":") fs
       TBit           -> text "Bit"
-      TInf           -> text "inf"
       TNum x         -> integer x
       TChar x        -> text (show x)
       TSeq t1 TBit   -> brackets (pp t1)
       TSeq t1 t2     -> optParens (n > 3)
-                      $ brackets (pp t1) <> ppPrec 3 t2
+                      $ brackets (pp t1) <.> ppPrec 3 t2
 
       _ | Just tinf <- isInfix ty ->
               optParens (n > 2)
@@ -821,14 +812,20 @@ instance PPName name => PP (Type name) where
 instance PPName name => PP (Prop name) where
   ppPrec n prop =
     case prop of
-      CFin t       -> text "fin"   <+> ppPrec 4 t
-      CArith t     -> text "Arith" <+> ppPrec 4 t
-      CCmp t       -> text "Cmp"   <+> ppPrec 4 t
-      CEqual t1 t2 -> ppPrec 2 t1 <+> text "==" <+> ppPrec 2 t2
-      CGeq t1 t2   -> ppPrec 2 t1 <+> text ">=" <+> ppPrec 2 t2
-      CLocated c _ -> ppPrec n c
-
-      CType t      -> ppPrec n t
+      CFin t         -> text "fin"   <+> ppPrec 4 t
+      CZero t        -> text "Zero"  <+> ppPrec 4 t
+      CLogic t       -> text "Logic" <+> ppPrec 4 t
+      CArith t       -> text "Arith" <+> ppPrec 4 t
+      CCmp t         -> text "Cmp"   <+> ppPrec 4 t
+      CSignedCmp t   -> text "SignedCmp" <+> ppPrec 4 t
+      CLiteral t1 t2 -> text "Literal" <+> ppPrec 4 t1 <+> ppPrec 4 t2
+      CEqual t1 t2   -> ppPrec 2 t1 <+> text "==" <+> ppPrec 2 t2
+      CNeq t1 t2     -> ppPrec 2 t1 <+> text "!=" <+> ppPrec 2 t2
+      CGeq t1 t2     -> ppPrec 2 t1 <+> text ">=" <+> ppPrec 2 t2
+      CUser f ts     -> optParens (n > 2)
+                      $ ppPrefixName f <+> fsep (map (ppPrec 4) ts)
+      CLocated c _   -> ppPrec n c
+      CType t        -> ppPrec n t
 
 
 --------------------------------------------------------------------------------
@@ -852,9 +849,10 @@ instance NoPos (Program name) where
   noPos (Program x) = Program (noPos x)
 
 instance NoPos (Module name) where
-  noPos m = Module { mName    = mName m
-                   , mImports = noPos (mImports m)
-                   , mDecls   = noPos (mDecls m)
+  noPos m = Module { mName      = mName m
+                   , mInstance  = mInstance m
+                   , mImports   = noPos (mImports m)
+                   , mDecls     = noPos (mDecls m)
                    }
 
 instance NoPos (TopDecl name) where
@@ -863,6 +861,15 @@ instance NoPos (TopDecl name) where
       Decl    x   -> Decl     (noPos x)
       TDNewtype n -> TDNewtype(noPos n)
       Include x   -> Include  (noPos x)
+      DParameterFun d  -> DParameterFun (noPos d)
+      DParameterType d -> DParameterType (noPos d)
+      DParameterConstraint d -> DParameterConstraint (noPos d)
+
+instance NoPos (ParameterType name) where
+  noPos a = a
+
+instance NoPos (ParameterFun x) where
+  noPos x = x { pfSchema = noPos (pfSchema x) }
 
 instance NoPos a => NoPos (TopLevel a) where
   noPos tl = tl { tlValue = noPos (tlValue tl) }
@@ -876,6 +883,7 @@ instance NoPos (Decl name) where
       DFixity f ns     -> DFixity f (noPos ns)
       DBind      x     -> DBind      (noPos x)
       DType      x     -> DType      (noPos x)
+      DProp      x     -> DProp      (noPos x)
       DLocated   x _   -> noPos x
 
 instance NoPos (Newtype name) where
@@ -902,15 +910,19 @@ instance NoPos Pragma where
 
 
 
-
 instance NoPos (TySyn name) where
   noPos (TySyn x y z) = TySyn (noPos x) (noPos y) (noPos z)
 
+instance NoPos (PropSyn name) where
+  noPos (PropSyn x y z) = PropSyn (noPos x) (noPos y) (noPos z)
+
 instance NoPos (Expr name) where
   noPos expr =
     case expr of
       EVar x        -> EVar     x
       ELit x        -> ELit     x
+      ENeg x        -> ENeg     (noPos x)
+      EComplement x -> EComplement (noPos x)
       ETuple x      -> ETuple   (noPos x)
       ERecord x     -> ERecord  (noPos x)
       ESel x y      -> ESel     (noPos x) y
@@ -967,7 +979,6 @@ instance NoPos (Type name) where
       TFun x y      -> TFun     (noPos x) (noPos y)
       TSeq x y      -> TSeq     (noPos x) (noPos y)
       TBit          -> TBit
-      TInf          -> TInf
       TNum n        -> TNum n
       TChar n       -> TChar n
       TLocated x _  -> noPos x
@@ -978,11 +989,15 @@ instance NoPos (Prop name) where
   noPos prop =
     case prop of
       CEqual  x y   -> CEqual  (noPos x) (noPos y)
+      CNeq x y      -> CNeq (noPos x) (noPos y)
       CGeq x y      -> CGeq (noPos x) (noPos y)
       CFin x        -> CFin (noPos x)
+      CZero x       -> CZero  (noPos x)
+      CLogic x      -> CLogic (noPos x)
       CArith x      -> CArith (noPos x)
       CCmp x        -> CCmp   (noPos x)
+      CSignedCmp x  -> CSignedCmp (noPos x)
+      CLiteral x y  -> CLiteral (noPos x) (noPos y)
+      CUser x y     -> CUser x (noPos y)
       CLocated c _  -> noPos c
       CType t       -> CType (noPos t)
-
-
