@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Parser.LexerUtils
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -19,8 +19,8 @@ import Cryptol.Utils.Panic
 
 import           Data.Char(toLower,generalCategory,isAscii,ord,isSpace)
 import qualified Data.Char as Char
-import           Data.Text.Lazy (Text)
-import qualified Data.Text.Lazy as T
+import           Data.Text(Text)
+import qualified Data.Text as T
 import           Data.Word(Word8)
 
 import GHC.Generics (Generic)
@@ -214,8 +214,9 @@ splitQual t =
 
 --------------------------------------------------------------------------------
 numToken :: Integer -> Text -> TokenT
-numToken rad ds = Num (toVal ds) (fromInteger rad) (fromIntegral (T.length ds))
+numToken rad ds = Num (toVal ds') (fromInteger rad) (fromIntegral (T.length ds'))
   where
+  ds' = T.filter (/= '_') ds
   toVal = T.foldl' (\x c -> rad * x + fromDigit c) 0
 
 fromDigit :: Char -> Integer
@@ -260,9 +261,10 @@ isExplicit Explicit{} = True
 isExplicit Virtual{}  = False
 
 startsLayout :: TokenT -> Bool
-startsLayout (KW KW_where)   = True
-startsLayout (KW KW_private) = True
-startsLayout _               = False
+startsLayout (KW KW_where)    = True
+startsLayout (KW KW_private)  = True
+startsLayout (KW KW_parameter) = True
+startsLayout _                = False
 
 -- Add separators computed from layout
 layout :: Config -> [Located Token] -> [Located Token]
@@ -358,11 +360,7 @@ data TokenV   = VCurlyL| VCurlyR | VSemi
 data TokenW   = BlockComment | LineComment | Space | DocStr
                 deriving (Eq, Show, Generic, NFData)
 
-data TokenKW  = KW_Arith
-              | KW_Bit
-              | KW_Cmp
-              | KW_else
-              | KW_Eq
+data TokenKW  = KW_else
               | KW_extern
               | KW_fin
               | KW_if
@@ -390,6 +388,8 @@ data TokenKW  = KW_Arith
               | KW_infixr
               | KW_infix
               | KW_primitive
+              | KW_parameter
+              | KW_constraint
                 deriving (Eq, Show, Generic, NFData)
 
 -- | The named operators are a special case for parsing types, and 'Other' is
