@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Parser.Names
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -15,19 +15,8 @@ import Cryptol.Parser.AST
 
 import           Data.Set (Set)
 import qualified Data.Set as Set
-import           Data.Foldable (fold)
 
 
-modExports :: Ord name => Module name -> ExportSpec name
-modExports m = fold (concat [ exportedNames d | d <- mDecls m ])
-  where
-  names by td = [ td { tlValue = thing n } | n <- fst (by (tlValue td)) ]
-
-  exportedNames (Decl td) = map exportBind  (names  namesD td)
-                         ++ map exportType (names tnamesD td)
-  exportedNames (TDNewtype nt) = map exportType (names tnamesNT nt)
-  exportedNames (Include {})  = []
-
 -- | The names defined by a newtype.
 tnamesNT :: Newtype name -> ([Located name], ())
 tnamesNT x = ([ nName x ], ())
@@ -49,6 +38,7 @@ namesD decl =
     DFixity{}     -> ([],Set.empty)
     DPragma {}    -> ([],Set.empty)
     DType {}      -> ([],Set.empty)
+    DProp {}      -> ([],Set.empty)
     DLocated d _  -> namesD d
 
 -- | The names defined and used by a single declarations in such a way
@@ -64,11 +54,15 @@ allNamesD decl =
     DFixity _ ns    -> ns
     DPragma ns _    -> ns
     DType ts        -> [tsName ts]
+    DProp ps        -> [psName ps]
     DLocated d _    -> allNamesD d
 
 tsName :: TySyn name -> Located name
 tsName (TySyn lqn _ _) = lqn
 
+psName :: PropSyn name -> Located name
+psName (PropSyn lqn _ _) = lqn
+
 -- | The names defined and used by a single binding.
 namesB :: Ord name => Bind name -> ([Located name], Set name)
 namesB b = ([bName b], boundNames (namesPs (bParams b)) (namesDef (thing (bDef b))))
@@ -85,6 +79,8 @@ namesE expr =
   case expr of
     EVar x        -> Set.singleton x
     ELit _        -> Set.empty
+    ENeg e        -> namesE e
+    EComplement e -> namesE e
     ETuple es     -> Set.unions (map namesE es)
     ERecord fs    -> Set.unions (map (namesE . value) fs)
     ESel e _      -> namesE e
@@ -155,7 +151,6 @@ namesT vs = go
       TBit          -> Set.empty
       TNum _        -> Set.empty
       TChar _       -> Set.empty
-      TInf          -> Set.empty
       TApp _ ts     -> Set.unions (map go ts)
       TTuple ts     -> Set.unions (map go ts)
       TRecord fs    -> Set.unions (map (go . value) fs)
@@ -167,6 +162,24 @@ namesT vs = go
       TParens t     -> namesT vs t
       TInfix a _ _ b-> Set.union (namesT vs a) (namesT vs b)
 
+-- | Given the set of type variables that are in scope,
+-- compute the type/constraint synonyms used by a prop.
+namesC :: Ord name => Set name -> Prop name -> Set name
+namesC vs prop =
+  case prop of
+    CFin t        -> namesT vs t
+    CEqual t1 t2  -> Set.union (namesT vs t1) (namesT vs t2)
+    CNeq   t1 t2  -> Set.union (namesT vs t1) (namesT vs t2)
+    CGeq t1 t2    -> Set.union (namesT vs t1) (namesT vs t2)
+    CZero t       -> namesT vs t
+    CLogic t      -> namesT vs t
+    CArith t      -> namesT vs t
+    CCmp t        -> namesT vs t
+    CSignedCmp t  -> namesT vs t
+    CLiteral t1 t2-> Set.union (namesT vs t1) (namesT vs t2)
+    CUser x ts    -> Set.insert x (Set.unions (map (namesT vs) ts))
+    CLocated p _  -> namesC vs p
+    CType t       -> namesT vs t
 
 -- | The type names defined and used by a group of mutually recursive declarations.
 tnamesDs :: Ord name => [Decl name] -> ([Located name], Set name)
@@ -186,6 +199,9 @@ tnamesD decl =
     DPatBind _ e         -> ([], tnamesE e)
     DLocated d _         -> tnamesD d
     DType (TySyn n ps t) -> ([n], Set.difference (tnamesT t) (Set.fromList (map tpName ps)))
+    DProp (PropSyn n ps cs)
+                         -> ([n], Set.difference (Set.unions (map tnamesC cs))
+                                  (Set.fromList (map tpName ps)))
 
 -- | The type names used by a single binding.
 tnamesB :: Ord name => Bind name -> Set name
@@ -205,6 +221,8 @@ tnamesE expr =
   case expr of
     EVar _        -> Set.empty
     ELit _        -> Set.empty
+    ENeg e        -> tnamesE e
+    EComplement e -> tnamesE e
     ETuple es     -> Set.unions (map tnamesE es)
     ERecord fs    -> Set.unions (map (tnamesE . value) fs)
     ESel e _      -> tnamesE e
@@ -258,13 +276,19 @@ tnamesS (Forall params props ty _) =
 tnamesC :: Ord name => Prop name -> Set name
 tnamesC prop =
   case prop of
-    CFin t       -> tnamesT t
-    CEqual t1 t2 -> Set.union (tnamesT t1) (tnamesT t2)
-    CGeq t1 t2   -> Set.union (tnamesT t1) (tnamesT t2)
-    CArith t     -> tnamesT t
-    CCmp t       -> tnamesT t
-    CLocated p _ -> tnamesC p
-    CType t      -> tnamesT t
+    CFin t         -> tnamesT t
+    CEqual t1 t2   -> Set.union (tnamesT t1) (tnamesT t2)
+    CNeq   t1 t2   -> Set.union (tnamesT t1) (tnamesT t2)
+    CGeq t1 t2     -> Set.union (tnamesT t1) (tnamesT t2)
+    CZero t        -> tnamesT t
+    CLogic t       -> tnamesT t
+    CArith t       -> tnamesT t
+    CCmp t         -> tnamesT t
+    CSignedCmp t   -> tnamesT t
+    CLiteral t1 t2 -> Set.union (tnamesT t1) (tnamesT t2)
+    CUser x ts     -> Set.insert x (Set.unions (map tnamesT ts))
+    CLocated p _   -> tnamesC p
+    CType t        -> tnamesT t
 
 -- | Compute the type synonyms/type variables used by a type.
 tnamesT :: Ord name => Type name -> Set name
@@ -276,7 +300,6 @@ tnamesT ty =
     TBit          -> Set.empty
     TNum _        -> Set.empty
     TChar __      -> Set.empty
-    TInf          -> Set.empty
     TApp _ ts     -> Set.unions (map tnamesT ts)
     TTuple ts     -> Set.unions (map tnamesT ts)
     TRecord fs    -> Set.unions (map (tnamesT . value) fs)
