@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Parser.NoPat
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -27,7 +27,6 @@ import Cryptol.Utils.Panic(panic)
 
 import           MonadLib hiding (mapM)
 import           Data.Maybe(maybeToList)
-import           Data.Either(partitionEithers)
 import qualified Data.Map as Map
 
 import GHC.Generics (Generic)
@@ -67,7 +66,7 @@ sel p x s = let (a,ts) = splitSimpleP p
 -- | Given a pattern, transform it into a simple pattern and a set of bindings.
 -- Simple patterns may only contain variables and type annotations.
 
--- XXX: We can replace the types in the selcetors with annotations on the bindings.
+-- XXX: We can replace the types in the selectors with annotations on the bindings.
 noPat :: Pattern PName -> NoPatM (Pattern PName, [Bind PName])
 noPat pat =
   case pat of
@@ -148,6 +147,8 @@ noPatE expr =
   case expr of
     EVar {}       -> return expr
     ELit {}       -> return expr
+    ENeg e        -> ENeg    <$> noPatE e
+    EComplement e -> EComplement <$> noPatE e
     ETuple es     -> ETuple  <$> mapM noPatE es
     ERecord es    -> ERecord <$> mapM noPatF es
     ESel e s      -> ESel    <$> noPatE e <*> return s
@@ -228,6 +229,7 @@ noMatchD decl =
                                               , bDoc = Nothing
                                               } : map DBind bs
     DType {}        -> return [decl]
+    DProp {}        -> return [decl]
 
     DLocated d r1   -> do bs <- inRange r1 $ noMatchD d
                           return $ map (`DLocated` r1) bs
@@ -254,19 +256,20 @@ noPatDs ds =
 
      return ds2
 
-noPatTopDs :: [TopLevel (Decl PName)] -> NoPatM [TopLevel (Decl PName)]
-noPatTopDs tds =
-  do noPatGroups <- mapM (noMatchD . tlValue) tds
 
-     let allDecls  = concat noPatGroups
+
+noPatTopDs :: [TopDecl PName] -> NoPatM [TopDecl PName]
+noPatTopDs tds =
+  do desugared <- concat <$> mapM desugar tds
+
+     let allDecls  = map tlValue (decls desugared)
          pragmaMap = Map.fromListWith (++) $ concatMap toPragma allDecls
          sigMap    = Map.fromListWith (++) $ concatMap toSig    allDecls
          fixMap    = Map.fromListWith (++) $ concatMap toFixity allDecls
-         docMap    = Map.fromListWith (++) $ concatMap toDocs   tds
+         docMap    = Map.fromListWith (++) $ concatMap toDocs   (decls tds)
 
-     let exportGroups = zipWith (\ td ds -> td { tlValue = ds }) tds noPatGroups
      (tds', (pMap,sMap,fMap,_)) <- runStateT (pragmaMap,sigMap,fixMap,docMap)
-                                             (annotTopDs exportGroups)
+                                             (annotTopDs desugared)
 
      forM_ (Map.toList pMap) $ \(n,ps) ->
        forM_ ps $ \p -> recordError $ PragmaNoBind (p { thing = n }) (thing p)
@@ -281,27 +284,23 @@ noPatTopDs tds =
 
      return tds'
 
+  where
+  decls xs = [ d | Decl d <- xs ]
+
+  desugar d =
+    case d of
+      Decl tl -> do ds <- noMatchD (tlValue tl)
+                    return [ Decl tl { tlValue = d1 } | d1 <- ds ]
+      x      -> return [x]
+
 
 noPatProg :: Program PName -> NoPatM (Program PName)
-noPatProg (Program topDs) =
-  do let (ds, others) = partitionEithers (map isDecl topDs)
-     ds1 <- noPatTopDs ds
-     return $ Program $ others ++ map Decl ds1
-
-  where
-  isDecl (Decl d) = Left d
-  isDecl d        = Right d
+noPatProg (Program topDs) = Program <$> noPatTopDs topDs
 
 noPatModule :: Module PName -> NoPatM (Module PName)
 noPatModule m =
-  do let (ds, others) = partitionEithers (map isDecl (mDecls m))
-     ds1 <- noPatTopDs ds
-     return m { mDecls = others ++ map Decl ds1 }
-
-  where
-  isDecl (Decl d) = Left d
-  isDecl d        = Right d
-
+  do ds1 <- noPatTopDs (mDecls m)
+     return m { mDecls = ds1 }
 
 --------------------------------------------------------------------------------
 
@@ -317,17 +316,36 @@ type AnnotMap = ( Map.Map PName [Located  Pragma       ]
 -- export specifications, this will favor the specification of the binding.
 -- This is most likely the intended behavior, so it's probably fine, but it does
 -- smell a bit.
-annotTopDs :: [TopLevel [Decl PName]]
-           -> StateT AnnotMap NoPatM [TopLevel (Decl PName)]
+annotTopDs :: [TopDecl PName] -> StateT AnnotMap NoPatM [TopDecl PName]
 annotTopDs tds =
   case tds of
 
-    (ds:dss) ->
-      do ds'  <- annotDs (tlValue ds)
-         rest <- annotTopDs dss
-         if null ds'
-            then return rest
-            else return ([ ds { tlValue = d } | d <- ds' ] ++ rest)
+    d : ds ->
+      case d of
+        Decl d1 ->
+          do ignore <- runExceptionT (annotD (tlValue d1))
+             case ignore of
+               Left _   -> annotTopDs ds
+               Right d2 -> (Decl (d1 { tlValue = d2 }) :) <$> annotTopDs ds
+
+        -- XXX: Add fixity once we support type-level fixities
+        DParameterType {} -> (d :) <$> annotTopDs ds
+        DParameterConstraint {} -> (d :) <$> annotTopDs ds
+        DParameterFun p ->
+          do (ps,ss,fs,ds') <- get
+             let rm _ _ = Nothing
+                 name = thing (pfName p)
+             case Map.updateLookupWithKey rm name fs of
+               (Nothing,_)  -> (d :) <$> annotTopDs ds
+               (Just f,fs1) ->
+                 do mbF <- lift (checkFixs name f)
+                    set (ps,ss,fs1,ds')
+                    let p1 = p { pfFixity = mbF }
+                    (DParameterFun p1 :) <$> annotTopDs ds
+
+        -- XXX: we may want to add pragmas to newtypes?
+        TDNewtype {} -> (d :) <$> annotTopDs ds
+        Include {}   -> (d :) <$> annotTopDs ds
 
     [] -> return []
 
@@ -352,6 +370,7 @@ annotD decl =
     DPragma {}    -> raise ()
     DPatBind {}   -> raise ()
     DType {}      -> return decl
+    DProp {}      -> return decl
     DLocated d r  -> (`DLocated` r) <$> annotD d
 
 -- | Add pragma/signature annotations to a binding.
@@ -433,6 +452,7 @@ toDocs TopLevel { .. }
       -- XXX revisit these
       DPragma _ _     -> []
       DType _         -> []
+      DProp _         -> []
 
 
 --------------------------------------------------------------------------------
@@ -491,12 +511,12 @@ instance PP Error where
         $$ nest 2 (vcat (map pp ss))
 
       SignatureNoBind x s ->
-        text "At" <+> pp (srcRange x) <> colon <+>
+        text "At" <+> pp (srcRange x) <.> colon <+>
         text "Type signature without a matching binding:"
          $$ nest 2 (pp s)
 
       PragmaNoBind x s ->
-        text "At" <+> pp (srcRange x) <> colon <+>
+        text "At" <+> pp (srcRange x) <.> colon <+>
         text "Pragma without a matching binding:"
          $$ nest 2 (pp s)
 
@@ -505,7 +525,7 @@ instance PP Error where
         $$ nest 2 (vcat (map pp locs))
 
       FixityNoBind n ->
-        text "At" <+> pp (srcRange n) <> colon <+>
+        text "At" <+> pp (srcRange n) <.> colon <+>
         text "Fixity declaration without a matching binding for:" <+>
          pp (thing n)
 
