@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Parser.ParserUtils
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -12,21 +12,22 @@
 {-# LANGUAGE DeriveAnyClass #-}
 {-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE OverloadedStrings #-}
 module Cryptol.Parser.ParserUtils where
 
 import Cryptol.Parser.AST
 import Cryptol.Parser.Lexer
 import Cryptol.Parser.Position
 import Cryptol.Parser.Utils (translateExprToNumT,widthIdent)
+import Cryptol.Utils.Ident(packModName)
 import Cryptol.Utils.PP
 import Cryptol.Utils.Panic
 
-import Data.Maybe(listToMaybe,fromMaybe)
+import Data.Maybe(fromMaybe)
 import Data.Bits(testBit,setBit)
 import Control.Monad(liftM,ap,unless)
-import qualified Data.Text as S
-import           Data.Text.Lazy (Text)
-import qualified Data.Text.Lazy as T
+import           Data.Text(Text)
+import qualified Data.Text as T
 
 
 import GHC.Generics (Generic)
@@ -39,62 +40,93 @@ parseString :: Config -> ParseM a -> String -> Either ParseError a
 parseString cfg p cs = parse cfg p (T.pack cs)
 
 parse :: Config -> ParseM a -> Text -> Either ParseError a
-parse cfg p cs    = case unP p cfg eofPos (S toks) of
+parse cfg p cs    = case unP p cfg eofPos S { sPrevTok = Nothing
+                                            , sTokens = toks
+                                            , sNextTyParamNum = 0
+                                            } of
                       Left err    -> Left err
                       Right (a,_) -> Right a
   where (toks,eofPos) = lexer cfg cs
 
 
 {- The parser is parameterized by the pozition of the final token. -}
-data ParseM a   = P { unP :: Config -> Position -> S -> Either ParseError (a,S) }
+newtype ParseM a =
+  P { unP :: Config -> Position -> S -> Either ParseError (a,S) }
 
 
 lexerP :: (Located Token -> ParseM a) -> ParseM a
-lexerP k = P $ \cfg p (S ts) ->
-  case ts of
+lexerP k = P $ \cfg p s ->
+  case sTokens s of
     t : _ | Err e <- tokenType it ->
       Left $ HappyErrorMsg (srcRange t) $
          case e of
            UnterminatedComment -> "unterminated comment"
            UnterminatedString  -> "unterminated string"
            UnterminatedChar    -> "unterminated character"
-           InvalidString       -> "invalid string literal: " ++ T.unpack (tokenText it)
-           InvalidChar         -> "invalid character literal: " ++ T.unpack (tokenText it)
-           LexicalError        -> "unrecognized character: " ++ T.unpack (tokenText it)
+           InvalidString       -> "invalid string literal:" ++
+                                    T.unpack (tokenText it)
+           InvalidChar         -> "invalid character literal:" ++
+                                    T.unpack (tokenText it)
+           LexicalError        -> "unrecognized character:" ++
+                                    T.unpack (tokenText it)
       where it = thing t
 
-    t : more -> unP (k t) cfg p (S more)
-    []       -> Left (HappyError (cfgSource cfg) p Nothing)
+    t : more -> unP (k t) cfg p s { sPrevTok = Just t, sTokens = more }
+    [] -> Left (HappyOutOfTokens (cfgSource cfg) p)
 
-data ParseError = HappyError FilePath Position (Maybe Token)
+data ParseError = HappyError FilePath         {- Name of source file -}
+                             (Located Token)  {- Offending token -}
                 | HappyErrorMsg Range String
+                | HappyUnexpected FilePath (Maybe (Located Token)) String
+                | HappyOutOfTokens FilePath Position
                   deriving (Show, Generic, NFData)
 
-newtype S = S [Located Token]
-
-instance PP ParseError where
-  ppPrec _ (HappyError _ _ tok) = case tok of
-                                  Nothing -> text "end of input"
-                                  Just t  -> pp t
-  ppPrec _ (HappyErrorMsg _ x) = text x
+data S = S { sPrevTok :: Maybe (Located Token)
+           , sTokens :: [Located Token]
+           , sNextTyParamNum :: !Int
+             -- ^ Keep track of the type parameters as they appear in the input
+           }
 
 ppError :: ParseError -> Doc
-ppError (HappyError path pos (Just tok))
-  | Err _ <- tokenType tok = text "Parse error at" <+>
-                              text path <> char ':' <> pp pos <> comma <+>
-                              pp tok
-ppError e@(HappyError path pos _) =
-                               text "Parse error at" <+>
-                               text path <> char ':' <> pp pos <> comma <+>
-                               text "unexpected" <+> pp e
+
+ppError (HappyError path ltok)
+  | Err _ <- tokenType tok =
+    text "Parse error at" <+>
+    text path <.> char ':' <.> pp pos <.> comma <+>
+    pp tok
+
+  | White DocStr <- tokenType tok =
+    "Unexpected documentation (/**) comment at" <+>
+    text path <.> char ':' <.> pp pos <.> colon $$
+    nest 2
+      "Documentation comments need to be followed by something to document."
+
+  | otherwise =
+    text "Parse error at" <+>
+    text path <.> char ':' <.> pp pos <.> comma $$
+    nest 2 (text "unexpected:" <+> pp tok)
+  where
+  pos = from (srcRange ltok)
+  tok = thing ltok
+
+ppError (HappyOutOfTokens path pos) =
+  text "Unexpected end of file at:" <+>
+    text path <.> char ':' <.> pp pos
+
 ppError (HappyErrorMsg p x)  = text "Parse error at" <+> pp p $$ nest 2 (text x)
 
-instance Monad ParseM where
-  return a  = P (\_ _ s -> Right (a,s))
-  fail s    = panic "[Parser] fail" [s]
-  m >>= k   = P (\cfg p s1 -> case unP m cfg p s1 of
-                            Left e       -> Left e
-                            Right (a,s2) -> unP (k a) cfg p s2)
+ppError (HappyUnexpected path ltok e) =
+  text "Parse error at" <+>
+   text path <.> char ':' <.> pp pos <.> comma $$
+   nest 2 unexp $$
+   nest 2 ("expected:" <+> text e)
+  where
+  (unexp,pos) =
+    case ltok of
+      Nothing -> (empty,start)
+      Just t  -> ( "unexpected:" <+> text (T.unpack (tokenText (thing t)))
+                 , from (srcRange t)
+                 )
 
 instance Functor ParseM where
   fmap = liftM
@@ -103,11 +135,19 @@ instance Applicative ParseM where
   pure  = return
   (<*>) = ap
 
+instance Monad ParseM where
+  return a  = P (\_ _ s -> Right (a,s))
+  fail s    = panic "[Parser] fail" [s]
+  m >>= k   = P (\cfg p s1 -> case unP m cfg p s1 of
+                            Left e       -> Left e
+                            Right (a,s2) -> unP (k a) cfg p s2)
+
 happyError :: ParseM a
-happyError = P $ \cfg p (S ls) ->
-  Left $ case listToMaybe ls of
-           Nothing -> HappyError (cfgSource cfg) p Nothing
-           Just l  -> HappyError (cfgSource cfg) (from (srcRange l)) (Just (thing l))
+happyError = P $ \cfg _ s ->
+  case sPrevTok s of
+    Just t  -> Left (HappyError (cfgSource cfg) t)
+    Nothing ->
+      Left (HappyErrorMsg emptyRange "Parse error at the beginning of the file")
 
 errorMessage :: Range -> String -> ParseM a
 errorMessage r x = P $ \_ _ _ -> Left (HappyErrorMsg r x)
@@ -115,8 +155,20 @@ errorMessage r x = P $ \_ _ _ -> Left (HappyErrorMsg r x)
 customError :: String -> Located Token -> ParseM a
 customError x t = P $ \_ _ _ -> Left (HappyErrorMsg (srcRange t) x)
 
-mkModName :: [T.Text] -> ModName
-mkModName strs = T.toStrict (T.intercalate (T.pack "::") strs)
+expected :: String -> ParseM a
+expected x = P $ \cfg _ s ->
+                    Left (HappyUnexpected (cfgSource cfg) (sPrevTok s) x)
+
+
+
+
+
+
+
+
+
+mkModName :: [Text] -> ModName
+mkModName = packModName
 
 -- Note that type variables are not resolved at this point: they are tcons.
 mkSchema :: [TParam PName] -> [Prop PName] -> Type PName -> Schema PName
@@ -124,7 +176,7 @@ mkSchema xs ps t = Forall xs ps t Nothing
 
 getName :: Located Token -> PName
 getName l = case thing l of
-              Token (Ident [] x) _ -> mkUnqual (mkIdent (T.toStrict x))
+              Token (Ident [] x) _ -> mkUnqual (mkIdent x)
               _ -> panic "[Parser] getName" ["not an Ident:", show l]
 
 getNum :: Located Token -> Integer
@@ -185,7 +237,6 @@ validDemotedType rng ty =
     TBit         -> bad "Type bit"
     TNum {}      -> ok
     TChar {}     -> ok
-    TInf         -> bad "Infinity type"
     TWild        -> bad "Wildcard types"
     TUser {}     -> ok
     TApp {}      -> ok
@@ -257,7 +308,7 @@ exprToNumT r expr =
 -- It is used to represent anonymous type applications.
 anonRecord :: Maybe Range -> [Type PName] -> Type PName
 anonRecord ~(Just r) ts = TRecord (map toField ts)
-  where noName    = Located { srcRange = r, thing = mkIdent (S.pack "") }
+  where noName    = Located { srcRange = r, thing = mkIdent (T.pack "") }
         toField t = Named { name = noName, value = t }
 
 exportDecl :: Maybe (Located String) -> ExportType -> Decl PName -> TopDecl PName
@@ -265,10 +316,36 @@ exportDecl mbDoc e d = Decl TopLevel { tlExport = e
                                      , tlDoc    = mbDoc
                                      , tlValue  = d }
 
-exportNewtype :: ExportType -> Newtype PName -> TopDecl PName
-exportNewtype e n = TDNewtype TopLevel { tlExport = e
-                                       , tlDoc    = Nothing
-                                       , tlValue  = n }
+exportNewtype :: ExportType -> Maybe (Located String) -> Newtype PName ->
+                                                            TopDecl PName
+exportNewtype e d n = TDNewtype TopLevel { tlExport = e
+                                         , tlDoc    = d
+                                         , tlValue  = n }
+
+mkParFun :: Maybe (Located String) ->
+            Located PName ->
+            Schema PName ->
+            TopDecl PName
+mkParFun mbDoc n s = DParameterFun ParameterFun { pfName = n
+                                                , pfSchema = s
+                                                , pfDoc = thing <$> mbDoc
+                                                , pfFixity = Nothing
+                                                }
+
+mkParType :: Maybe (Located String) ->
+             Located PName ->
+             Located Kind ->
+             ParseM (TopDecl PName)
+mkParType mbDoc n k =
+  do num <- P $ \_ _ s -> let nu = sNextTyParamNum s
+                          in Right (nu, s { sNextTyParamNum = nu + 1 })
+     return (DParameterType
+             ParameterType { ptName    = n
+                           , ptKind    = thing k
+                           , ptDoc     = thing <$> mbDoc
+                           , ptFixity  = Nothing
+                           , ptNumber  = num
+                           })
 
 changeExport :: ExportType -> [TopDecl PName] -> [TopDecl PName]
 changeExport e = map change
@@ -276,6 +353,10 @@ changeExport e = map change
   change (Decl d)      = Decl      d { tlExport = e }
   change (TDNewtype n) = TDNewtype n { tlExport = e }
   change td@Include{}  = td
+  change (DParameterType {}) = panic "changeExport" ["private type parameter?"]
+  change (DParameterFun {})  = panic "changeExport" ["private value parameter?"]
+  change (DParameterConstraint {}) =
+    panic "changeExport" ["private type constraint parameter?"]
 
 mkTypeInst :: Named (Type PName) -> TypeInst PName
 mkTypeInst x | nullIdent (thing (name x)) = PosInst (value x)
@@ -295,6 +376,14 @@ mkTySyn ln ps b
   | otherwise =
     return $ DType $ TySyn ln ps b
 
+mkPropSyn :: Located PName -> [TParam PName] -> Type PName -> ParseM (Decl PName)
+mkPropSyn ln ps b
+  | getIdent (thing ln) == widthIdent =
+    errorMessage (srcRange ln) "`width` is not a valid constraint synonym name."
+
+  | otherwise =
+    DProp . PropSyn ln ps . thing <$> mkProp b
+
 polyTerm :: Range -> Integer -> Integer -> ParseM (Bool, Integer)
 polyTerm rng k p
   | k == 0          = return (False, p)
@@ -321,7 +410,7 @@ mkPoly rng terms = mk 0 (map fromInteger bits)
 mkProperty :: LPName -> [Pattern PName] -> Expr PName -> Decl PName
 mkProperty f ps e = DBind Bind { bName       = f
                                , bParams     = reverse ps
-                               , bDef        = at e (Located emptyRange (DExpr (ETyped e TBit)))
+                               , bDef        = at e (Located emptyRange (DExpr e))
                                , bSignature  = Nothing
                                , bPragmas    = [PragmaProperty]
                                , bMono       = False
@@ -369,19 +458,25 @@ mkDoc ltxt = ltxt { thing = docStr }
          $ dropPrefix
          $ trimFront
          $ T.lines
-         $ T.dropWhileEnd (`elem` "/* \r\n\t")
+         $ T.dropWhileEnd commentChar
          $ thing ltxt
 
+  commentChar :: Char -> Bool
+  commentChar x = x `elem` ("/* \r\n\t" :: String)
+
+  prefixDroppable x = x `elem` ("* \r\n\t" :: String)
+
   trimFront []                     = []
   trimFront (l:ls)
-    | T.all (`elem` "/* \r\n\t") l = ls
-    | otherwise                    = T.dropWhile (`elem` "/* ") l : ls
+    | T.all commentChar l = ls
+    | otherwise           = T.dropWhile commentChar l : ls
 
   dropPrefix []        = []
-  dropPrefix [t]       = [T.dropWhile (`elem` "/* ") t]
+  dropPrefix [t]       = [T.dropWhile commentChar t]
   dropPrefix ts@(l:ls) =
     case T.uncons l of
-      Just (c,_) | all (commonPrefix c) ls -> dropPrefix (map (T.drop 1) ts)
+      Just (c,_) | prefixDroppable c &&
+                   all (commonPrefix c) ls -> dropPrefix (map (T.drop 1) ts)
       _                                    -> ts
 
     where
@@ -391,6 +486,11 @@ mkDoc ltxt = ltxt { thing = docStr }
         Nothing     -> False
 
 
+distrLoc :: Located [a] -> [Located a]
+distrLoc x = [ Located { srcRange = r, thing = a } | a <- thing x ]
+  where r = srcRange x
+
+
 mkProp :: Type PName -> ParseM (Located [Prop PName])
 mkProp ty =
   case ty of
@@ -415,7 +515,6 @@ mkProp ty =
       TBit{}    -> err
       TNum{}    -> err
       TChar{}   -> err
-      TInf{}    -> err
       TWild     -> err
       TRecord{} -> err
 
@@ -428,14 +527,52 @@ mkProp ty =
 
   -- these can be translated right away
   prefixProp r f xs
-    | i == arithIdent, [x] <- xs = return [CLocated (CArith x) r]
-    | i == finIdent,   [x] <- xs = return [CLocated (CFin x) r]
-    | i == cmpIdent,   [x] <- xs = return [CLocated (CCmp x) r]
-    | otherwise                  = errorMessage r "Invalid constraint"
+    | i == zeroIdent,      [x] <- xs = return [CLocated (CZero x) r]
+    | i == logicIdent,     [x] <- xs = return [CLocated (CLogic x) r]
+    | i == arithIdent,     [x] <- xs = return [CLocated (CArith x) r]
+    | i == finIdent,       [x] <- xs = return [CLocated (CFin x) r]
+    | i == cmpIdent,       [x] <- xs = return [CLocated (CCmp x) r]
+    | i == signedCmpIdent, [x] <- xs = return [CLocated (CSignedCmp x) r]
+    | i == literalIdent, [x,y] <- xs = return [CLocated (CLiteral x y) r]
+    | otherwise                      = return [CLocated (CType (TUser f xs)) r]
     where
     i = getIdent f
 
-arithIdent, finIdent, cmpIdent :: Ident
-arithIdent = mkIdent (S.pack "Arith")
-finIdent   = mkIdent (S.pack "fin")
-cmpIdent   = mkIdent (S.pack "Cmp")
+zeroIdent, logicIdent, arithIdent, finIdent, cmpIdent, signedCmpIdent, literalIdent :: Ident
+zeroIdent      = mkIdent "Zero"
+logicIdent     = mkIdent "Logic"
+arithIdent     = mkIdent "Arith"
+finIdent       = mkIdent "fin"
+cmpIdent       = mkIdent "Cmp"
+signedCmpIdent = mkIdent "SignedCmp"
+literalIdent   = mkIdent "Literal"
+
+-- | Make an ordinary module
+mkModule :: Located ModName ->
+            ([Located Import], [TopDecl PName]) ->
+            Module PName
+mkModule nm (is,ds) = Module { mName = nm
+                             , mInstance = Nothing
+                             , mImports = is
+                             , mDecls = ds
+                             }
+
+-- | Make an unnamed module---gets the name @Main@.
+mkAnonymousModule :: ([Located Import], [TopDecl PName]) ->
+                     Module PName
+mkAnonymousModule = mkModule Located { srcRange = emptyRange
+                                     , thing    = mkModName [T.pack "Main"]
+                                     }
+
+-- | Make a module which defines a functor instance.
+mkModuleInstance :: Located ModName ->
+                    Located ModName ->
+                    ([Located Import], [TopDecl PName]) ->
+                    Module PName
+mkModuleInstance nm fun (is,ds) =
+  Module { mName     = nm
+         , mInstance = Just fun
+         , mImports  = is
+         , mDecls    = ds
+         }
+
