@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Parser.Unlit
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -13,8 +13,8 @@ module Cryptol.Parser.Unlit
   ( unLit, PreProc(..), guessPreProc, knownExts
   ) where
 
-import           Data.Text.Lazy (Text)
-import qualified Data.Text.Lazy as Text
+import           Data.Text(Text)
+import qualified Data.Text as Text
 import           Data.Char(isSpace)
 import           System.FilePath(takeExtension)
 
