@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.REPL.Command
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -55,6 +55,7 @@ import qualified Cryptol.ModuleSystem.Name as M
 import qualified Cryptol.ModuleSystem.NamingEnv as M
 import qualified Cryptol.ModuleSystem.Renamer as M (RenamerWarning(SymbolShadowed))
 import qualified Cryptol.Utils.Ident as M
+import qualified Cryptol.ModuleSystem.Env as M
 
 import qualified Cryptol.Eval.Monad as E
 import qualified Cryptol.Eval.Value as E
@@ -65,11 +66,12 @@ import Cryptol.Parser
     (parseExprWith,parseReplWith,ParseError(),Config(..),defaultConfig
     ,parseModName,parseHelpName)
 import qualified Cryptol.TypeCheck.AST as T
+import qualified Cryptol.TypeCheck.Error as T
+import qualified Cryptol.TypeCheck.Parseable as T
 import qualified Cryptol.TypeCheck.Subst as T
-import qualified Cryptol.TypeCheck.InferTypes as T
 import           Cryptol.TypeCheck.Solve(defaultReplExpr)
-import qualified Cryptol.TypeCheck.Solver.CrySAT as CrySAT
-import Cryptol.TypeCheck.PP (dump,ppWithNames)
+import qualified Cryptol.TypeCheck.Solver.SMT as SMT
+import Cryptol.TypeCheck.PP (dump)
 import Cryptol.Utils.PP
 import Cryptol.Utils.Panic(panic)
 import qualified Cryptol.Parser.AST as P
@@ -78,12 +80,12 @@ import Cryptol.Symbolic (ProverCommand(..), QueryType(..), SatNum(..),ProverStat
 import qualified Cryptol.Symbolic as Symbolic
 
 import qualified Control.Exception as X
-import Control.Monad hiding (mapM, mapM_)
+import Control.Monad hiding (mapM, mapM)
 import qualified Data.ByteString as BS
 import Data.Bits ((.&.))
 import Data.Char (isSpace,isPunctuation,isSymbol)
 import Data.Function (on)
-import Data.List (intercalate,nub,sortBy,partition)
+import Data.List (intercalate, nub, sortBy, partition, isPrefixOf)
 import Data.Maybe (fromMaybe,mapMaybe)
 import System.Environment (lookupEnv)
 import System.Exit (ExitCode(ExitSuccess))
@@ -93,14 +95,14 @@ import System.FilePath((</>), isPathSeparator)
 import System.Directory(getHomeDirectory,setCurrentDirectory,doesDirectoryExist)
 import qualified Data.Map as Map
 import qualified Data.Set as Set
-import qualified Data.IntMap as IntMap
 import System.IO(hFlush,stdout)
 import System.Random.TF(newTFGen)
 import Numeric (showFFloat)
-import qualified Data.Text as ST
-import qualified Data.Text.Lazy as T
+import qualified Data.Text as T
 import Data.IORef(newIORef,readIORef)
 
+import GHC.Float (log1p, expm1)
+
 import Prelude ()
 import Prelude.Compat
 
@@ -137,9 +139,11 @@ data CommandBody
   | FileExprArg (FilePath -> String -> REPL ())
   | DeclsArg    (String   -> REPL ())
   | ExprTypeArg (String   -> REPL ())
+  | ModNameArg  (String   -> REPL ())
   | FilenameArg (FilePath -> REPL ())
   | OptionArg   (String   -> REPL ())
   | ShellArg    (String   -> REPL ())
+  | HelpArg     (String   -> REPL ())
   | NoArg       (REPL ())
 
 
@@ -165,54 +169,59 @@ nbCommands  = foldl insert emptyTrie nbCommandList
 nbCommandList :: [CommandDescr]
 nbCommandList  =
   [ CommandDescr [ ":t", ":type" ] (ExprArg typeOfCmd)
-    "check the type of an expression"
-  , CommandDescr [ ":b", ":browse" ] (ExprTypeArg browseCmd)
-    "display the current environment"
-  , CommandDescr [ ":?", ":help" ] (ExprArg helpCmd)
-    "display a brief description about a function"
+    "Check the type of an expression."
+  , CommandDescr [ ":b", ":browse" ] (ModNameArg browseCmd)
+    "Display environment for all loaded modules, or for a specific module."
+  , CommandDescr [ ":?", ":help" ] (HelpArg helpCmd)
+    "Display a brief description of a function, type, or command."
   , CommandDescr [ ":s", ":set" ] (OptionArg setOptionCmd)
-    "set an environmental option (:set on its own displays current values)"
+    "Set an environmental option (:set on its own displays current values)."
   , CommandDescr [ ":check" ] (ExprArg (void . qcCmd QCRandom))
-    "use random testing to check that the argument always returns true (if no argument, check all properties)"
+    "Use random testing to check that the argument always returns true.\n(If no argument, check all properties.)"
   , CommandDescr [ ":exhaust" ] (ExprArg (void . qcCmd QCExhaust))
-    "use exhaustive testing to prove that the argument always returns true (if no argument, check all properties)"
+    "Use exhaustive testing to prove that the argument always returns\ntrue. (If no argument, check all properties.)"
   , CommandDescr [ ":prove" ] (ExprArg proveCmd)
-    "use an external solver to prove that the argument always returns true (if no argument, check all properties)"
+    "Use an external solver to prove that the argument always returns\ntrue. (If no argument, check all properties.)"
   , CommandDescr [ ":sat" ] (ExprArg satCmd)
-    "use a solver to find a satisfying assignment for which the argument returns true (if no argument, find an assignment for all properties)"
+    "Use a solver to find a satisfying assignment for which the argument\nreturns true. (If no argument, find an assignment for all properties.)"
   , CommandDescr [ ":debug_specialize" ] (ExprArg specializeCmd)
-    "do type specialization on a closed expression"
+    "Do type specialization on a closed expression."
   , CommandDescr [ ":eval" ] (ExprArg refEvalCmd)
-    "evaluate an expression with the reference evaluator"
+    "Evaluate an expression with the reference evaluator."
+  , CommandDescr [ ":ast" ] (ExprArg astOfCmd)
+    "Print out the pre-typechecked AST of a given term."
+  , CommandDescr [ ":extract-coq" ] (NoArg allTerms)
+    "Print out the post-typechecked AST of all currently defined terms,\nin a Coq-parseable format."
   ]
 
 commandList :: [CommandDescr]
 commandList  =
   nbCommandList ++
   [ CommandDescr [ ":q", ":quit" ] (NoArg quitCmd)
-    "exit the REPL"
+    "Exit the REPL."
   , CommandDescr [ ":l", ":load" ] (FilenameArg loadCmd)
-    "load a module"
+    "Load a module by filename."
   , CommandDescr [ ":r", ":reload" ] (NoArg reloadCmd)
-    "reload the currently loaded module"
+    "Reload the currently loaded module."
   , CommandDescr [ ":e", ":edit" ] (FilenameArg editCmd)
-    "edit the currently loaded module"
+    "Edit the currently loaded module."
   , CommandDescr [ ":!" ] (ShellArg runShellCmd)
-    "execute a command in the shell"
+    "Execute a command in the shell."
   , CommandDescr [ ":cd" ] (FilenameArg cdCmd)
-    "set the current working directory"
+    "Set the current working directory."
   , CommandDescr [ ":m", ":module" ] (FilenameArg moduleCmd)
-    "load a module"
+    "Load a module by its name."
   , CommandDescr [ ":w", ":writeByteArray" ] (FileExprArg writeFileCmd)
-    "write data of type `fin n => [n][8]` to a file"
+    "Write data of type 'fin n => [n][8]' to a file."
   , CommandDescr [ ":readByteArray" ] (FilenameArg readFileCmd)
-    "read data from a file as type `fin n => [n][8]`, binding the value to variable `it`"
+    "Read data from a file as type 'fin n => [n][8]', binding\nthe value to variable 'it'."
   ]
 
 genHelp :: [CommandDescr] -> [String]
 genHelp cs = map cmdHelp cs
   where
-  cmdHelp cmd  = concat [ "  ", cmdNames cmd, pad (cmdNames cmd), cHelp cmd ]
+  cmdHelp cmd  = concat $ [ "  ", cmdNames cmd, pad (cmdNames cmd),
+                            intercalate ("\n  " ++ pad []) (lines (cHelp cmd)) ]
   cmdNames cmd = intercalate ", " (cNames cmd)
   padding      = 2 + maximum (map (length . cmdNames) cs)
   pad n        = replicate (max 0 (padding - length n)) ' '
@@ -248,6 +257,12 @@ getPPValOpts =
                      , E.useInfLength = infLength
                      }
 
+getEvalOpts :: REPL E.EvalOpts
+getEvalOpts =
+  do ppOpts <- getPPValOpts
+     l      <- getLogger
+     return E.EvalOpts { E.evalPPOpts = ppOpts, E.evalLogger = l }
+
 evalCmd :: String -> REPL ()
 evalCmd str = do
   letEnabled <- getLetEnabled
@@ -258,7 +273,7 @@ evalCmd str = do
     P.ExprInput expr -> do
       (val,_ty) <- replEvalExpr expr
       ppOpts <- getPPValOpts
-      valDoc <- io $ rethrowEvalError $ E.runEval $ E.ppValue ppOpts val
+      valDoc <- rEvalRethrow (E.ppValue ppOpts val)
 
       -- This is the point where the value gets forced. We deepseq the
       -- pretty-printed representation of it, rather than the value
@@ -273,6 +288,14 @@ evalCmd str = do
       -- be generalized if mono-binds is enabled
       replEvalDecl decl
 
+printCounterexample :: Bool -> P.Expr P.PName -> [E.Value] -> REPL ()
+printCounterexample isSat pexpr vs =
+  do ppOpts <- getPPValOpts
+     docs <- mapM (rEval . E.ppValue ppOpts) vs
+     let doc = ppPrec 3 pexpr -- function application has precedence 3
+     rPrint $ hang doc 2 (sep docs) <+>
+       text (if isSat then "= True" else "= False")
+
 data QCMode = QCRandom | QCExhaust deriving (Eq, Show)
 
 -- | Randomly test a property, or exhaustively check it if the number
@@ -292,67 +315,77 @@ qcCmd qcMode "" =
 qcCmd qcMode str =
   do expr <- replParseExpr str
      (val,ty) <- replEvalExpr expr
-     EnvNum testNum  <- getUser "tests"
+     EnvNum testNum <- getUser "tests"
      case testableType ty of
-       Just (sz,vss) | qcMode == QCExhaust || sz <= toInteger testNum -> do
+       Just (Just sz,tys,vss) | qcMode == QCExhaust || sz <= toInteger testNum -> do
             rPutStrLn "Using exhaustive testing."
             let f _ [] = panic "Cryptol.REPL.Command"
                                     ["Exhaustive testing ran out of test cases"]
                 f _ (vs : vss1) = do
-                  result <- io $ runOneTest val vs
+                  evo <- getEvalOpts
+                  result <- io $ runOneTest evo val vs
                   return (result, vss1)
                 testSpec = TestSpec {
                     testFn = f
                   , testProp = str
                   , testTotal = sz
-                  , testPossible = sz
+                  , testPossible = Just sz
                   , testRptProgress = ppProgress
                   , testClrProgress = delProgress
-                  , testRptFailure = ppFailure
+                  , testRptFailure = ppFailure tys expr
                   , testRptSuccess = do
                       delTesting
-                      prtLn $ "passed " ++ show sz ++ " tests."
+                      prtLn $ "Passed " ++ show sz ++ " tests."
                       rPutStrLn "Q.E.D."
                   }
             prt testingMsg
             report <- runTests testSpec vss
             return [report]
 
-       Just (sz,_) -> case TestR.testableType ty of
+       Just (sz,tys,_) | qcMode == QCRandom ->
+         case TestR.testableType ty of
               Nothing   -> raise (TypeNotTestable ty)
               Just gens -> do
                 rPutStrLn "Using random testing."
+                evo <- getEvalOpts
                 let testSpec = TestSpec {
-                        testFn = \sz' g -> io $ TestR.runOneTest val gens sz' g
+                        testFn = \sz' g ->
+                                      io $ TestR.runOneTest evo val gens sz' g
                       , testProp = str
                       , testTotal = toInteger testNum
                       , testPossible = sz
                       , testRptProgress = ppProgress
                       , testClrProgress = delProgress
-                      , testRptFailure = ppFailure
+                      , testRptFailure = ppFailure tys expr
                       , testRptSuccess = do
                           delTesting
-                          prtLn $ "passed " ++ show testNum ++ " tests."
+                          prtLn $ "Passed " ++ show testNum ++ " tests."
                       }
                 prt testingMsg
                 g <- io newTFGen
                 report <- runTests testSpec g
-                when (isPass (reportResult report)) $ do
-                  let szD = fromIntegral sz :: Double
-                      percent = fromIntegral (testNum * 100) / szD
+                when (isPass (reportResult report)) $
+                  case sz of
+                    Nothing -> return ()
+                    Just n -> rPutStrLn $ coverageString testNum n
+                return [report]
+       _ -> raise (TypeNotTestable ty)
+
+  where
+  testingMsg = "Testing... "
+
+  coverageString testNum sz =
+                  let (percent, expectedUnique) = expectedCoverage testNum sz
                       showValNum
                         | sz > 2 ^ (20::Integer) =
                           "2^^" ++ show (lg2 sz)
                         | otherwise = show sz
-                  rPutStrLn $ "Coverage: "
+                  in "Expected test coverage: "
                     ++ showFFloat (Just 2) percent "% ("
-                    ++ show testNum ++ " of "
-                    ++ showValNum ++ " values)"
-                return [report]
-       Nothing -> return []
+                    ++ showFFloat (Just 0) expectedUnique " of "
+                    ++ showValNum
+                    ++ " values)"
 
-  where
-  testingMsg = "testing..."
 
   totProgressWidth = 4    -- 100%
 
@@ -371,28 +404,77 @@ qcCmd qcMode str =
         pad     = replicate (totProgressWidth - width) ' '
     in prt (pad ++ percent)
 
-  del n       = unlessBatch $ prt (replicate n '\BS')
+  del n       = unlessBatch
+              $ prt (replicate n '\BS' ++ replicate n ' ' ++ replicate n '\BS')
   delTesting  = del (length testingMsg)
   delProgress = del totProgressWidth
 
-  ppFailure failure = do
+  ppFailure tys pexpr failure = do
     delTesting
     opts <- getPPValOpts
     case failure of
-      FailFalse [] -> do
-        prtLn "FAILED"
       FailFalse vs -> do
-        prtLn "FAILED for the following inputs:"
-        mapM_ (\v -> rPrint =<< (io $ E.runEval $ E.ppValue opts v)) vs
+        let isSat = False
+        printCounterexample isSat pexpr vs
+        case (tys,vs) of
+          ([t],[v]) -> bindItVariableVal t v
+          _ -> let fs = [ M.packIdent ("arg" ++ show (i::Int)) | i <- [ 1 .. ] ]
+                   t = T.TRec (zip fs tys)
+                   v = E.VRecord (zip fs (map return vs))
+               in bindItVariableVal t v
+
       FailError err [] -> do
         prtLn "ERROR"
         rPrint (pp err)
       FailError err vs -> do
         prtLn "ERROR for the following inputs:"
-        mapM_ (\v -> rPrint =<< (io $ E.runEval $ E.ppValue opts v)) vs
+        mapM_ (\v -> rPrint =<< (rEval $ E.ppValue opts v)) vs
         rPrint (pp err)
       Pass -> panic "Cryptol.REPL.Command" ["unexpected Test.Pass"]
 
+
+-- | This function computes the expected coverage percentage and
+-- expected number of unique test vectors when using random testing.
+--
+-- The expected test coverage proportion is:
+--  @1 - ((n-1)/n)^k@
+--
+-- This formula takes into account the fact that test vectors are chosen
+-- uniformly at random _with replacement_, and thus the same vectors
+-- may be generated multiple times.  If the test vectors were chosen
+-- randomly without replacement, the proportion would instead be @k/n@.
+--
+-- We compute raising to the @k@ power in the log domain to improve 
+-- numerical precision. The equivalant comptutation is:
+--   @-expm1( k * log1p (-1/n) )@
+--
+-- Where @expm1(x) = exp(x) - 1@ and @log1p(x) = log(1 + x)@.
+--
+-- However, if @sz@ is large enough, even carefully preserving
+-- precision may not be enough to get sensible results.  In such
+-- situations, we expect the naive approximation @k/n@ to be very
+-- close to accurate and the expected number of unique values is
+-- essentially equal to the number of tests.
+expectedCoverage :: Int -> Integer -> (Double, Double)
+expectedCoverage testNum sz =
+    -- If the Double computation has enough precision, use the
+    --  "with replacement" formula.
+    if testNum > 0 && proportion > 0 then
+       (100.0 * proportion, szD * proportion)
+    else
+       (100.0 * naiveProportion, numD)
+
+  where
+   szD :: Double
+   szD = fromInteger sz
+
+   numD :: Double
+   numD = fromIntegral testNum
+
+   naiveProportion = numD / szD
+
+   proportion = negate (expm1 (numD * log1p (negate (recip szD))))
+
 satCmd, proveCmd :: String -> REPL ()
 satCmd = cmdProveSat True
 proveCmd = cmdProveSat False
@@ -404,6 +486,12 @@ showProverStats mprover stat = rPutStrLn msg
   msg = "(Total Elapsed Time: " ++ SBV.showTDiff stat ++
         maybe "" (\p -> ", using " ++ show p) mprover ++ ")"
 
+rethrowErrorCall :: REPL a -> REPL a
+rethrowErrorCall m = REPL (\r -> unREPL m r `X.catch` handler)
+  where
+    handler (X.ErrorCallWithLocation s _) = X.throwIO (SBVError s)
+
+
 -- | Console-specific version of 'proveSat'. Prints output to the
 -- console, and binds the @it@ variable to a record whose form depends
 -- on the expression given. See ticket #66 for a discussion of this
@@ -444,8 +532,7 @@ cmdProveSat isSat str = do
             Just path -> io $ writeFile path smtlib
             Nothing -> rPutStr smtlib
     _ -> do
-      (firstProver,result,stats) <- onlineProveSat isSat str mfile
-      ppOpts <- getPPValOpts
+      (firstProver,result,stats) <- rethrowErrorCall (onlineProveSat isSat str mfile)
       case result of
         Symbolic.EmptyResult         ->
           panic "REPL.Command" [ "got EmptyResult for online prover query" ]
@@ -457,13 +544,6 @@ cmdProveSat isSat str = do
         Symbolic.AllSatResult tevss -> do
           let tess = map (map $ \(t,e,_) -> (t,e)) tevss
               vss  = map (map $ \(_,_,v) -> v)     tevss
-              ppvs vs = do
-                parseExpr <- replParseExpr str
-                docs <- mapM (io . E.runEval . E.ppValue ppOpts) vs
-                let -- function application has precedence 3
-                    doc = ppPrec 3 parseExpr
-                rPrint $ hang doc 2 (sep docs) <+>
-                  text (if isSat then "= True" else "= False")
           resultRecs <- mapM (mkSolverResult cexStr isSat . Right) tess
           let collectTes tes = (t, es)
                 where
@@ -475,10 +555,11 @@ cmdProveSat isSat str = do
               (ty, exprs) =
                 case resultRecs of
                   [] -> panic "REPL.Command.onlineProveSat"
-                          [ "no satisfying assignments after mkSovlerResult" ]
+                          [ "no satisfying assignments after mkSolverResult" ]
                   [(t, e)] -> (t, [e])
                   _        -> collectTes resultRecs
-          forM_ vss ppvs
+          pexpr <- replParseExpr str
+          forM_ vss (printCounterexample isSat pexpr)
           case (ty, exprs) of
             (t, [e]) -> bindItVariable t e
             (t, es ) -> bindItVariables t es
@@ -578,8 +659,21 @@ refEvalCmd :: String -> REPL ()
 refEvalCmd str = do
   parseExpr <- replParseExpr str
   (_, expr, _schema) <- replCheckExpr parseExpr
+  validEvalContext expr
   val <- liftModuleCmd (rethrowEvalError . R.evaluate expr)
-  rPrint $ R.ppValue val
+  opts <- getPPValOpts
+  rPrint $ R.ppValue opts val
+
+astOfCmd :: String -> REPL ()
+astOfCmd str = do
+ expr <- replParseExpr str
+ (re,_,_) <- replCheckExpr (P.noPos expr)
+ rPrint (fmap M.nameUnique re)
+
+allTerms :: REPL ()
+allTerms = do
+  me <- getModuleEnv
+  rPrint $ T.showParseable $ concatMap T.mDecls $ M.loadedModules me
 
 typeOfCmd :: String -> REPL ()
 typeOfCmd str = do
@@ -588,9 +682,8 @@ typeOfCmd str = do
   (_re,def,sig) <- replCheckExpr expr
 
   -- XXX need more warnings from the module system
-  --io (mapM_ printWarning ws)
   whenDebug (rPutStrLn (dump def))
-  (_,_,names) <- getFocusedEnv
+  (_,_,_,names) <- getFocusedEnv
   -- type annotation ':' has precedence 2
   rPrint $ runDoc names $ ppPrec 2 expr <+> text ":" <+> pp sig
 
@@ -622,18 +715,31 @@ writeFileCmd file str = do
                        (\(n,b) -> T.tIsBit b && T.tIsNum n == Just 8)
                        (T.tIsSeq x)
   serializeValue (E.VSeq n vs) = do
-    ws <- io $ E.runEval (mapM (>>=E.fromVWord "serializeValue") $ E.enumerateSeqMap n vs)
+    ws <- rEval
+            (mapM (>>=E.fromVWord "serializeValue") $ E.enumerateSeqMap n vs)
     return $ BS.pack $ map serializeByte ws
   serializeValue _             =
     panic "Cryptol.REPL.Command.writeFileCmd"
       ["Impossible: Non-VSeq value of type [n][8]."]
   serializeByte (E.BV _ v) = fromIntegral (v .&. 0xFF)
 
+
+rEval :: E.Eval a -> REPL a
+rEval m = do ev <- getEvalOpts
+             io (E.runEval ev m)
+
+rEvalRethrow :: E.Eval a -> REPL a
+rEvalRethrow m = do ev <- getEvalOpts
+                    io $ rethrowEvalError $ E.runEval ev m
+
 reloadCmd :: REPL ()
 reloadCmd  = do
   mb <- getLoadedMod
   case mb of
-    Just m  -> loadCmd (lPath m)
+    Just lm  ->
+      case lName lm of
+        Just m | M.isParamInstModName m -> loadHelper (M.loadModuleByName m)
+        _ -> loadCmd (lPath lm)
     Nothing -> return ()
 
 
@@ -646,7 +752,7 @@ editCmd path
         Just m -> do
           success <- replEdit (lPath m)
           if success
-             then loadCmd (lPath m)
+             then reloadCmd
              else return ()
 
         Nothing   -> do
@@ -655,17 +761,15 @@ editCmd path
 
   | otherwise = do
       _  <- replEdit path
-      mb <- getLoadedMod
-      case mb of
-        Nothing -> loadCmd path
-        Just _  -> return ()
+      setEditPath path
+      reloadCmd
 
 moduleCmd :: String -> REPL ()
 moduleCmd modString
   | null modString = return ()
   | otherwise      = do
       case parseModName modString of
-        Just m -> loadCmd =<< liftModuleCmd (M.findModule m)
+        Just m  -> loadHelper (M.loadModuleByName m)
         Nothing -> rPutStrLn "Invalid module name."
 
 loadPrelude :: REPL ()
@@ -674,37 +778,48 @@ loadPrelude  = moduleCmd $ show $ pp M.preludeName
 loadCmd :: FilePath -> REPL ()
 loadCmd path
   | null path = return ()
-  | otherwise = do
-      setLoadedMod LoadedModule
-        { lName = Nothing
-        , lPath = path
-        }
+  | otherwise = do setEditPath path
+                   loadHelper (M.loadModuleByPath path)
 
-      m <- liftModuleCmd (M.loadModuleByPath path)
-      whenDebug (rPutStrLn (dump m))
-      setLoadedMod LoadedModule
+loadHelper :: M.ModuleCmd (FilePath,T.Module) -> REPL ()
+loadHelper how =
+  do clearLoadedMod
+     (path,m) <- liftModuleCmd how
+     whenDebug (rPutStrLn (dump m))
+     setLoadedMod LoadedModule
         { lName = Just (T.mName m)
         , lPath = path
         }
-      setDynEnv mempty
+     setDynEnv mempty
 
 quitCmd :: REPL ()
 quitCmd  = stop
 
 
 browseCmd :: String -> REPL ()
-browseCmd pfx = do
-  (iface,names,disp) <- getFocusedEnv
+browseCmd input = do
+  (_,iface,names,disp) <- getFocusedEnv
+
+  let mnames = map (M.textToModName . T.pack) (words input)
+  validModNames <- (:) M.interactiveName <$> getModNames
+  let checkModName m =
+        unless (m `elem` validModNames) $
+        rPutStrLn ("error: " ++ show m ++ " is not a loaded module.")
+  mapM_ checkModName mnames
+
+  let f &&& g = \x -> f x && g x
+      isUser x = case M.nameInfo x of
+                   M.Declared _ M.SystemName -> False
+                   _ -> True
+      inSet s x = x `Set.member` s
+
   let (visibleTypes,visibleDecls) = M.visibleNames names
 
-      (visibleType,visibleDecl)
-        | null pfx  =
-          ((`Set.member` visibleTypes)
-          ,(`Set.member` visibleDecls))
+      restricted = if null mnames then const True else hasAnyModName mnames
+
+      visibleType = isUser &&& restricted &&& inSet visibleTypes
+      visibleDecl = isUser &&& restricted &&& inSet visibleDecls
 
-        | otherwise =
-          (\n -> n `Set.member` visibleTypes && pfx `isNamePrefix` n
-          ,\n -> n `Set.member` visibleDecls && pfx `isNamePrefix` n)
 
   browseTSyns    visibleType iface disp
   browseNewtypes visibleType iface disp
@@ -714,11 +829,14 @@ browseTSyns :: (M.Name -> Bool) -> M.IfaceDecls -> NameDisp -> REPL ()
 browseTSyns isVisible M.IfaceDecls { .. } names = do
   let tsyns = sortBy (M.cmpNameDisplay names `on` T.tsName)
               [ ts | ts <- Map.elems ifTySyns, isVisible (T.tsName ts) ]
-  unless (null tsyns) $ do
-    rPutStrLn "Type Synonyms"
-    rPutStrLn "============="
-    rPrint (runDoc names (nest 4 (vcat (map pp tsyns))))
-    rPutStrLn ""
+
+      (cts,tss) = partition isCtrait tsyns
+
+  ppBlock names pp "Type Synonyms" tss
+  ppBlock names pp "Constraint Synonyms" cts
+
+  where
+  isCtrait t = T.kindResult (T.kindOf (T.tsDef t)) == T.KProp
 
 browseNewtypes :: (M.Name -> Bool) -> M.IfaceDecls -> NameDisp -> REPL ()
 browseNewtypes isVisible M.IfaceDecls { .. } names = do
@@ -739,15 +857,19 @@ browseVars isVisible M.IfaceDecls { .. } names = do
   let isProp p     = T.PragmaProperty `elem` (M.ifDeclPragmas p)
       (props,syms) = partition isProp vars
 
-  ppBlock "Properties" props
-  ppBlock "Symbols"    syms
 
-  where
-  ppBlock name xs = unless (null xs) $
+  let ppVar M.IfaceDecl { .. } = hang (pp ifDeclName <+> char ':')
+                                   2 (pp ifDeclSig)
+
+  ppBlock names ppVar "Properties" props
+  ppBlock names ppVar "Symbols"    syms
+
+
+ppBlock :: NameDisp -> (a -> Doc) -> String -> [a] -> REPL ()
+ppBlock names ppFun name xs = unless (null xs) $
     do rPutStrLn name
        rPutStrLn (replicate (length name) '=')
-       let ppVar M.IfaceDecl { .. } = pp ifDeclName <+> char ':' <+> pp ifDeclSig
-       rPrint (runDoc names (nest 4 (vcat (map ppVar xs))))
+       rPrint (runDoc names (nest 4 (vcat (map ppFun xs))))
        rPutStrLn ""
 
 
@@ -767,62 +889,180 @@ setOptionCmd str
   describe k = do
     ev <- tryGetUser k
     case ev of
-      Just (EnvString s)   -> rPutStrLn (k ++ " = " ++ s)
-      Just (EnvProg p as)  -> rPutStrLn (k ++ " = " ++ intercalate " " (p:as))
-      Just (EnvNum n)      -> rPutStrLn (k ++ " = " ++ show n)
-      Just (EnvBool True)  -> rPutStrLn (k ++ " = on")
-      Just (EnvBool False) -> rPutStrLn (k ++ " = off")
-      Nothing              -> do rPutStrLn ("Unknown user option: `" ++ k ++ "`")
-                                 when (any isSpace k) $ do
-                                   let (k1, k2) = break isSpace k
-                                   rPutStrLn ("Did you mean: `:set " ++ k1 ++ " =" ++ k2 ++ "`?")
+      Just v  -> rPutStrLn (k ++ " = " ++ showEnvVal v)
+      Nothing -> do rPutStrLn ("Unknown user option: `" ++ k ++ "`")
+                    when (any isSpace k) $ do
+                      let (k1, k2) = break isSpace k
+                      rPutStrLn ("Did you mean: `:set " ++ k1 ++ " =" ++ k2 ++ "`?")
 
+showEnvVal :: EnvVal -> String
+showEnvVal ev =
+  case ev of
+    EnvString s   -> s
+    EnvProg p as  -> intercalate " " (p:as)
+    EnvNum n      -> show n
+    EnvBool True  -> "on"
+    EnvBool False -> "off"
 
 -- XXX at the moment, this can only look at declarations.
 helpCmd :: String -> REPL ()
 helpCmd cmd
   | null cmd  = mapM_ rPutStrLn (genHelp commandList)
+  | cmd0 : args <- words cmd, ":" `isPrefixOf` cmd0 =
+    case findCommandExact cmd0 of
+      []  -> void $ runCommand (Unknown cmd0)
+      [c] -> showCmdHelp c args
+      cs  -> void $ runCommand (Ambiguous cmd0 (concatMap cNames cs))
   | otherwise =
     case parseHelpName cmd of
       Just qname ->
-        do (env,rnEnv,nameEnv) <- getFocusedEnv
-           name <- liftModuleCmd (M.renameVar rnEnv qname)
-           case Map.lookup name (M.ifDecls env) of
-             Just M.IfaceDecl { .. } ->
-               do rPutStrLn ""
+        do (params,env,rnEnv,nameEnv) <- getFocusedEnv
+           let vNames = M.lookupValNames  qname rnEnv
+               tNames = M.lookupTypeNames qname rnEnv
+               pNames = T.primTyFromPName qname
 
-                  let property
-                        | P.PragmaProperty `elem` ifDeclPragmas = text "property"
-                        | otherwise                             = empty
-                  rPrint $ runDoc nameEnv
-                         $ nest 4
-                         $ property
-                           <+> pp qname
-                           <+> colon
-                           <+> pp (ifDeclSig)
-
-                  let mbFix = ifDeclFixity `mplus`
-                              (guard ifDeclInfix >> return P.defaultFixity)
-                  case mbFix of
-                    Just f  ->
-                      let msg = "Precedence " ++ show (P.fLevel f) ++ ", " ++
-                                 (case P.fAssoc f of
-                                    P.LeftAssoc   -> "associates to the left."
-                                    P.RightAssoc  -> "associates to the right."
-                                    P.NonAssoc    -> "does not associate.")
-
-                      in rPutStrLn ('\n' : msg)
-                    Nothing -> return ()
-
-                  case ifDeclDoc of
-                    Just str -> rPutStrLn ('\n' : str)
-                    Nothing  -> return ()
-
-             Nothing -> rPutStrLn "// No documentation is available."
+           mapM_ (showTypeHelp params env nameEnv) tNames
+           mapM_ (showValHelp params env nameEnv qname) vNames
+           mapM_ (showPrimTyHelp nameEnv) pNames
 
+           when (null (vNames ++ tNames) && pNames == Nothing) $
+             rPrint $ "Undefined name:" <+> pp qname
       Nothing ->
            rPutStrLn ("Unable to parse name: " ++ cmd)
 
+  where
+  noInfo nameEnv name =
+    case M.nameInfo name of
+      M.Declared m _ ->
+                      rPrint $runDoc nameEnv ("Name defined in module" <+> pp m)
+      M.Parameter  -> rPutStrLn "// No documentation is available."
+
+  showPrimTyHelp nameEnv pt =
+    do rPutStrLn ""
+       let i    = T.primTyIdent pt
+           nm   = pp (T.primTyIdent pt)
+           pnam = if P.isInfixIdent i then parens nm else nm
+           sig  = "primitive type" <+> pnam <+> ":" <+> pp (T.kindOf (T.primTyCon pt))
+       rPrint $ runDoc nameEnv $ nest 4 sig
+       doShowFix (T.primTyFixity pt)
+       rPutStrLn ""
+       rPutStrLn (T.primTyDoc pt)
+       rPutStrLn ""
+
+  showTypeHelp params env nameEnv name =
+    fromMaybe (noInfo nameEnv name) $
+    msum [ fromTySyn, fromNewtype, fromTyParam ]
+
+    where
+    fromTySyn =
+      do ts <- Map.lookup name (M.ifTySyns env)
+         return (doShowTyHelp nameEnv (pp ts) (T.tsDoc ts))
+
+    fromNewtype =
+      do nt <- Map.lookup name (M.ifNewtypes env)
+         let decl = pp nt $$ (pp name <+> text ":" <+> pp (T.newtypeConType nt))
+         return $ doShowTyHelp nameEnv decl (T.ntDoc nt)
+
+    fromTyParam =
+      do p <- Map.lookup name (M.ifParamTypes params)
+         let uses c = T.TVBound (T.mtpParam p) `Set.member` T.fvs c
+             ctrs = filter uses (map P.thing (M.ifParamConstraints params))
+             ctrDoc = case ctrs of
+                        [] -> empty
+                        [x] -> pp x
+                        xs  -> parens $ hsep $ punctuate comma $ map pp xs
+             decl = text "parameter" <+> pp name <+> text ":"
+                      <+> pp (T.mtpKind p)
+                   $$ ctrDoc
+         return $ doShowTyHelp nameEnv decl (T.mtpDoc p)
+
+  doShowTyHelp nameEnv decl doc =
+    do rPutStrLn ""
+       rPrint (runDoc nameEnv (nest 4 decl))
+       case doc of
+         Nothing -> return ()
+         Just d  -> rPutStrLn "" >> rPutStrLn d
+
+  doShowFix fx =
+    case fx of
+      Just f  ->
+        let msg = "Precedence " ++ show (P.fLevel f) ++ ", " ++
+                   (case P.fAssoc f of
+                      P.LeftAssoc   -> "associates to the left."
+                      P.RightAssoc  -> "associates to the right."
+                      P.NonAssoc    -> "does not associate.")
+
+        in rPutStrLn ('\n' : msg)
+
+      Nothing -> return ()
+
+  showValHelp params env nameEnv qname name =
+    fromMaybe (noInfo nameEnv name)
+              (msum [ fromDecl, fromNewtype, fromParameter ])
+    where
+    fromDecl =
+      do M.IfaceDecl { .. } <- Map.lookup name (M.ifDecls env)
+         return $
+           do rPutStrLn ""
+
+              let property
+                    | P.PragmaProperty `elem` ifDeclPragmas = text "property"
+                    | otherwise                             = empty
+              rPrint $ runDoc nameEnv
+                     $ nest 4
+                     $ property
+                       <+> pp qname
+                       <+> colon
+                       <+> pp (ifDeclSig)
+
+              doShowFix $ ifDeclFixity `mplus`
+                          (guard ifDeclInfix >> return P.defaultFixity)
+
+              case ifDeclDoc of
+                Just str -> rPutStrLn ('\n' : str)
+                Nothing  -> return ()
+
+    fromNewtype =
+      do _ <- Map.lookup name (M.ifNewtypes env)
+         return $ return ()
+
+    fromParameter =
+      do p <- Map.lookup name (M.ifParamFuns params)
+         return $
+           do rPutStrLn ""
+              rPrint $ runDoc nameEnv
+                     $ nest 4
+                     $ text "parameter" <+> pp qname
+                                        <+> colon
+                                        <+> pp (T.mvpType p)
+
+              doShowFix (T.mvpFixity p)
+
+              case T.mvpDoc p of
+                Just str -> rPutStrLn ('\n' : str)
+                Nothing  -> return ()
+
+  showCmdHelp c [arg] | ":set" `elem` cNames c = showOptionHelp arg
+  showCmdHelp c _args =
+    do rPutStrLn ("\n    " ++ intercalate ", " (cNames c))
+       rPutStrLn ""
+       rPutStrLn (cHelp c)
+       rPutStrLn ""
+
+  showOptionHelp arg =
+    case lookupTrieExact arg userOptions of
+      [opt] ->
+        do let k = optName opt
+           ev <- tryGetUser k
+           rPutStrLn $ "\n    " ++ k ++ " = " ++ maybe "???" showEnvVal ev
+           rPutStrLn ""
+           rPutStrLn ("Default value: " ++ showEnvVal (optDefault opt))
+           rPutStrLn ""
+           rPutStrLn (optHelp opt)
+           rPutStrLn ""
+      [] -> rPutStrLn ("Unknown setting name `" ++ arg ++ "`")
+      _  -> rPutStrLn ("Ambiguous setting name `" ++ arg ++ "`")
+
 
 runShellCmd :: String -> REPL ()
 runShellCmd cmd
@@ -848,22 +1088,13 @@ handleCtrlC a = do rPutStrLn "Ctrl-C"
 
 -- Utilities -------------------------------------------------------------------
 
-isNamePrefix :: String -> M.Name -> Bool
-isNamePrefix pfx =
-  let pfx' = ST.pack pfx
-   in \n -> case M.nameInfo n of
-              M.Declared _ -> pfx' `ST.isPrefixOf` M.identText (M.nameIdent n)
-              M.Parameter  -> False
+hasAnyModName :: [M.ModName] -> M.Name -> Bool
+hasAnyModName mnames n =
+  case M.nameInfo n of
+    M.Declared m _ -> m `elem` mnames
+    M.Parameter  -> False
 
 
-{-
-printWarning :: (Range,Warning) -> IO ()
-printWarning = print . ppWarning
-
-printError :: (Range,Error) -> IO ()
-printError = print . ppError
--}
-
 -- | Lift a parsing action into the REPL monad.
 replParse :: (String -> Either ParseError a) -> String -> REPL a
 replParse parse str = case parse str of
@@ -883,7 +1114,10 @@ getPrimMap :: REPL M.PrimMap
 getPrimMap  = liftModuleCmd M.getPrimMap
 
 liftModuleCmd :: M.ModuleCmd a -> REPL a
-liftModuleCmd cmd = moduleCmdResult =<< io . cmd =<< getModuleEnv
+liftModuleCmd cmd =
+  do evo <- getEvalOpts
+     env <- getModuleEnv
+     moduleCmdResult =<< io (cmd (evo,env))
 
 moduleCmdResult :: M.ModuleRes a -> REPL a
 moduleCmdResult (res,ws0) = do
@@ -911,11 +1145,17 @@ moduleCmdResult (res,ws0) = do
       filterShadowing w = Just w
 
   let ws = mapMaybe filterDefaults . mapMaybe filterShadowing $ ws0
-  (_,_,names) <- getFocusedEnv
+  (_,_,_,names) <- getFocusedEnv
   mapM_ (rPrint . runDoc names . pp) ws
   case res of
     Right (a,me') -> setModuleEnv me' >> return a
-    Left err      -> raise (ModuleSystemError names err)
+    Left err      ->
+      do e <- case err of
+                M.ErrorInFile file e -> do setEditPath file
+                                           return e
+                _ -> return err
+         raise (ModuleSystemError names e)
+
 
 replCheckExpr :: P.Expr P.PName -> REPL (P.Expr M.Name,T.Expr,T.Schema)
 replCheckExpr e = liftModuleCmd $ M.checkExpr e
@@ -944,17 +1184,16 @@ replSpecExpr e = liftModuleCmd $ S.specialize e
 replEvalExpr :: P.Expr P.PName -> REPL (E.Value, T.Type)
 replEvalExpr expr =
   do (_,def,sig) <- replCheckExpr expr
-
+     validEvalContext def
      me <- getModuleEnv
      let cfg = M.meSolverConfig me
-     mbDef <- io $ CrySAT.withSolver cfg (\s -> defaultReplExpr s def sig)
+     mbDef <- io $ SMT.withSolver cfg (\s -> defaultReplExpr s def sig)
 
      (def1,ty) <-
         case mbDef of
           Nothing -> raise (EvalPolyError sig)
           Just (tys,def1) ->
-            do let nms = T.addTNames (T.sVars sig) IntMap.empty
-               mapM_ (warnDefault nms) tys
+            do warnDefaults tys
                let su = T.listSubst [ (T.tpVar a, t) | (a,t) <- tys ]
                return (def1, T.apSubst su (T.sType sig))
 
@@ -964,8 +1203,16 @@ replEvalExpr expr =
      bindItVariable ty def1
      return (val,ty)
   where
-  warnDefault ns (x,t) =
-        rPrint $ text "Assuming" <+> ppWithNames ns x <+> text "=" <+> pp t
+  warnDefaults ts =
+    case ts of
+      [] -> return ()
+      _  ->
+        do rPutStrLn "Showing a specific instance of polymorphic result:"
+           mapM_ warnDefault ts
+
+  warnDefault (x,t) =
+    rPrint ("  *" <+> nest 2  ("Using" <+> quotes (pp t) <+> "for" <+>
+                                pp (T.tvarDesc (T.tpInfo x))))
 
 itIdent :: M.Ident
 itIdent  = M.packIdent "it"
@@ -984,7 +1231,7 @@ replReadFile fp handler =
 -- it to the current dynamic environment
 bindItVariable :: T.Type -> T.Expr -> REPL ()
 bindItVariable ty expr = do
-  freshIt <- freshName itIdent
+  freshIt <- freshName itIdent M.UserName
   let schema = T.Forall { T.sVars  = []
                         , T.sProps = []
                         , T.sType  = ty
@@ -1003,6 +1250,21 @@ bindItVariable ty expr = do
                            `M.shadowing` M.deNames denv
   setDynEnv $ denv { M.deNames = nenv' }
 
+
+-- | Extend the dynamic environment with a fresh binding for "it",
+-- as defined by the given value.  If we cannot determine the definition
+-- of the value, then we don't bind `it`.
+bindItVariableVal :: T.Type -> E.Value -> REPL ()
+bindItVariableVal ty val =
+  do prims   <- getPrimMap
+     mb      <- rEval (E.toExpr prims ty val)
+     case mb of
+       Nothing   -> return ()
+       Just expr -> bindItVariable ty expr
+
+
+
+
 -- | Creates a fresh binding of "it" to a finite sequence of
 -- expressions of the same type, and adds that sequence to the current
 -- dynamic environment
@@ -1016,6 +1278,7 @@ bindItVariables ty exprs = bindItVariable seqTy seqExpr
 replEvalDecl :: P.Decl P.PName -> REPL ()
 replEvalDecl decl = do
   dgs <- replCheckDecls [decl]
+  validEvalContext dgs
   whenDebug (mapM_ (\dg -> (rPutStrLn (dump dg))) dgs)
   liftModuleCmd (M.evalDecls dgs)
 
@@ -1089,9 +1352,11 @@ parseCommand findCmd line = do
       ExprArg     body -> Just (Command (body args'))
       DeclsArg    body -> Just (Command (body args'))
       ExprTypeArg body -> Just (Command (body args'))
+      ModNameArg  body -> Just (Command (body args'))
       FilenameArg body -> Just (Command (body =<< expandHome args'))
       OptionArg   body -> Just (Command (body args'))
       ShellArg    body -> Just (Command (body args'))
+      HelpArg     body -> Just (Command (body args'))
       NoArg       body -> Just (Command  body)
       FileExprArg body ->
         case extractFilePath args' of
