@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.REPL.Monad
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -39,7 +39,8 @@ module Cryptol.REPL.Monad (
   , getExprNames
   , getTypeNames
   , getPropertyNames
-  , LoadedModule(..), getLoadedMod, setLoadedMod
+  , getModNames
+  , LoadedModule(..), getLoadedMod, setLoadedMod, clearLoadedMod, setEditPath
   , setSearchPath, prependSearchPath
   , getPrompt
   , shouldContinue
@@ -48,6 +49,7 @@ module Cryptol.REPL.Monad (
   , disableLet
   , enableLet
   , getLetEnabled
+  , validEvalContext
   , updateREPLTitle
   , setUpdateREPLTitle
 
@@ -61,6 +63,7 @@ module Cryptol.REPL.Monad (
 
     -- ** Configurable Output
   , getPutStr
+  , getLogger
   , setPutStr
 
     -- ** Smoke Test
@@ -82,13 +85,15 @@ import Cryptol.Parser.NoPat (Error)
 import Cryptol.Parser.Position (emptyRange, Range(from))
 import qualified Cryptol.TypeCheck.AST as T
 import qualified Cryptol.TypeCheck as T
+import qualified Cryptol.IR.FreeVars as T
 import qualified Cryptol.Utils.Ident as I
 import Cryptol.Utils.PP
 import Cryptol.Utils.Panic (panic)
+import Cryptol.Utils.Logger(Logger, logPutStr, funLogger)
 import qualified Cryptol.Parser.AST as P
 import Cryptol.Symbolic (proverNames, lookupProver, SatNum(..))
 
-import Control.Monad (ap,unless,when)
+import Control.Monad (ap,unless,when,msum)
 import Control.Monad.Base
 import Control.Monad.IO.Class
 import Control.Monad.Trans.Control
@@ -96,12 +101,13 @@ import Data.Char (isSpace)
 import Data.IORef
     (IORef,newIORef,readIORef,modifyIORef,atomicModifyIORef)
 import Data.List (intercalate, isPrefixOf, unfoldr, sortBy)
-import Data.Maybe (catMaybes)
+import Data.Maybe (catMaybes,mapMaybe)
 import Data.Ord (comparing)
 import Data.Typeable (Typeable)
 import System.Directory (findExecutable)
 import qualified Control.Exception as X
 import qualified Data.Map as Map
+import qualified Data.Set as Set
 import Text.Read (readMaybe)
 
 import Data.SBV.Dynamic (sbvCheckSolverInstallation)
@@ -111,26 +117,44 @@ import Prelude.Compat
 
 -- REPL Environment ------------------------------------------------------------
 
+-- | This indicates what the user would like to work on.
 data LoadedModule = LoadedModule
-  { lName :: Maybe P.ModName -- ^ Focused module
-  , lPath :: FilePath        -- ^ Focused file
+  { lName :: Maybe P.ModName -- ^ Working on this module.
+  , lPath :: FilePath        -- ^ Working on this file.
   }
 
 -- | REPL RW Environment.
 data RW = RW
   { eLoadedMod   :: Maybe LoadedModule
+    -- ^ This is the name of the currently "focused" module.
+    -- This is what we edit (:e) or reload (:r)
+
   , eContinue    :: Bool
+    -- ^ Should we keep going when we encounter an error, or give up.
+
   , eIsBatch     :: Bool
+    -- ^ Are we in batch mode.
+
   , eModuleEnv   :: M.ModuleEnv
+    -- ^ The current environment of all things loaded.
+
   , eUserEnv     :: UserEnv
-  , ePutStr      :: String -> IO ()
+    -- ^ User settings
+
+  , eLogger      :: Logger
+    -- ^ Use this to send messages to the user
+
   , eLetEnabled  :: Bool
+    -- ^ Should we allow `let` on the command line
+
   , eUpdateTitle :: REPL ()
+    -- ^ Execute this every time we load a module.
+    -- This is used to change the title of terminal when loading a module.
   }
 
 -- | Initial, empty environment.
-defaultRW :: Bool -> IO RW
-defaultRW isBatch = do
+defaultRW :: Bool -> Logger -> IO RW
+defaultRW isBatch l = do
   env <- M.initialModuleEnv
   return RW
     { eLoadedMod   = Nothing
@@ -138,7 +162,7 @@ defaultRW isBatch = do
     , eIsBatch     = isBatch
     , eModuleEnv   = env
     , eUserEnv     = mkUserEnv userOptions
-    , ePutStr      = putStr
+    , eLogger      = l
     , eLetEnabled  = True
     , eUpdateTitle = return ()
     }
@@ -155,9 +179,9 @@ mkPrompt rw
 newtype REPL a = REPL { unREPL :: IORef RW -> IO a }
 
 -- | Run a REPL action with a fresh environment.
-runREPL :: Bool -> REPL a -> IO a
-runREPL isBatch m = do
-  ref <- newIORef =<< defaultRW isBatch
+runREPL :: Bool -> Logger -> REPL a -> IO a
+runREPL isBatch l m = do
+  ref <- newIORef =<< defaultRW isBatch l
   unREPL m ref
 
 instance Functor REPL where
@@ -209,6 +233,8 @@ data REPLException
   | ModuleSystemError NameDisp M.ModuleError
   | EvalPolyError T.Schema
   | TypeNotTestable T.Type
+  | EvalInParamModule P.ModName [M.Name]
+  | SBVError String
     deriving (Show,Typeable)
 
 instance X.Exception REPLException
@@ -232,6 +258,9 @@ instance PP REPLException where
                          $$ text "Type:" <+> pp s
     TypeNotTestable t    -> text "The expression is not of a testable type."
                          $$ text "Type:" <+> pp t
+    EvalInParamModule _ xs ->
+      text "Expression depends on a module parameter:" <+> hsep (map pp xs)
+    SBVError s           -> text "SBV error:" $$ text s
 
 -- | Raise an exception.
 raise :: REPLException -> REPL a
@@ -239,7 +268,8 @@ raise exn = io (X.throwIO exn)
 
 
 catch :: REPL a -> (REPLException -> REPL a) -> REPL a
-catch m k = REPL (\ ref -> unREPL m ref `X.catch` \ e -> unREPL (k e) ref)
+catch m k = REPL (\ ref -> rethrowEvalError (unREPL m ref) `X.catch` \ e -> unREPL (k e) ref)
+
 
 rethrowEvalError :: IO a -> IO a
 rethrowEvalError m = run `X.catch` rethrow
@@ -256,6 +286,7 @@ rethrowEvalError m = run `X.catch` rethrow
 
 -- Primitives ------------------------------------------------------------------
 
+
 io :: IO a -> REPL a
 io m = REPL (\ _ -> m)
 
@@ -272,6 +303,12 @@ modifyRW_ f = REPL (\ ref -> modifyIORef ref f)
 getPrompt :: REPL String
 getPrompt  = mkPrompt `fmap` getRW
 
+
+clearLoadedMod :: REPL ()
+clearLoadedMod = do modifyRW_ (\rw -> rw { eLoadedMod = upd <$> eLoadedMod rw })
+                    updateREPLTitle
+  where upd x = x { lName = Nothing }
+
 -- | Set the name of the currently focused file, edited by @:e@ and loaded via
 -- @:r@.
 setLoadedMod :: LoadedModule -> REPL ()
@@ -279,6 +316,9 @@ setLoadedMod n = do
   modifyRW_ (\ rw -> rw { eLoadedMod = Just n })
   updateREPLTitle
 
+setEditPath :: FilePath -> REPL ()
+setEditPath p = setLoadedMod LoadedModule { lName = Nothing, lPath = p }
+
 getLoadedMod :: REPL (Maybe LoadedModule)
 getLoadedMod  = eLoadedMod `fmap` getRW
 
@@ -323,6 +363,42 @@ enableLet  = modifyRW_ (\ rw -> rw { eLetEnabled = True })
 getLetEnabled :: REPL Bool
 getLetEnabled = fmap eLetEnabled getRW
 
+-- | Is evaluation enabled.  If the currently focused module is
+-- parameterized, then we cannot evalute.
+validEvalContext :: T.FreeVars a => a -> REPL ()
+validEvalContext a =
+  do me <- eModuleEnv <$> getRW
+     case M.meFocusedModule me of
+       Nothing -> return ()
+       Just fm ->
+         case M.lookupModule fm me of
+           Just lm
+             | Set.null tps && Set.null vps -> return ()
+             | Just xs <- check (T.freeVars a) ->
+                              raise $ EvalInParamModule fm xs
+             | otherwise -> return ()
+             where
+             m = M.lmModule lm
+
+             tps = Set.fromList $ map T.mtpParam $ Map.elems $ T.mParamTypes m
+             vps = Map.keysSet (T.mParamFuns m)
+
+             bad ds | not $ null badTP = Just $ mapMaybe T.tpName badTP
+                    | not $ null badVP = Just badVP
+                    | otherwise = Nothing
+               where
+               badTP = Set.toList $ Set.intersection tps (T.tyParams ds)
+               badVP = Set.toList $ Set.intersection vps (T.valDeps ds)
+
+             check ds   = msum (bad ds : map badSub (Set.toList (T.valDeps ds)))
+             badSub x   = maybe Nothing bad (Map.lookup x deps)
+             deps       = T.moduleDeps m
+
+           Nothing ->
+             panic "getEvalEnabled" ["The focused module is not loaded."
+                                    , show fm ]
+
+
 -- | Update the title
 updateREPLTitle :: REPL ()
 updateREPLTitle  = unlessBatch $ do
@@ -335,18 +411,23 @@ setUpdateREPLTitle m = modifyRW_ (\rw -> rw { eUpdateTitle = m })
 
 -- | Set the REPL's string-printer
 setPutStr :: (String -> IO ()) -> REPL ()
-setPutStr fn = modifyRW_ (\rw -> rw { ePutStr = fn })
+setPutStr fn = modifyRW_ $ \rw -> rw { eLogger = funLogger fn }
 
 -- | Get the REPL's string-printer
 getPutStr :: REPL (String -> IO ())
-getPutStr = fmap ePutStr getRW
+getPutStr =
+  do rw <- getRW
+     return (logPutStr (eLogger rw))
+
+getLogger :: REPL Logger
+getLogger = eLogger <$> getRW
 
 
 -- | Use the configured output action to print a string
 rPutStr :: String -> REPL ()
 rPutStr str = do
-  rw <- getRW
-  io $ ePutStr rw str
+  f <- getPutStr
+  io (f str)
 
 -- | Use the configured output action to print a string with a trailing newline
 rPutStrLn :: String -> REPL ()
@@ -356,14 +437,15 @@ rPutStrLn str = rPutStr $ str ++ "\n"
 rPrint :: Show a => a -> REPL ()
 rPrint x = rPutStrLn (show x)
 
-getFocusedEnv :: REPL (M.IfaceDecls,M.NamingEnv,NameDisp)
+getFocusedEnv :: REPL (M.IfaceParams,M.IfaceDecls,M.NamingEnv,NameDisp)
 getFocusedEnv  = do
   me <- getModuleEnv
   -- dyNames is a NameEnv that removes the #Uniq prefix from interactively-bound
   -- variables.
   let (dyDecls,dyNames,dyDisp) = M.dynamicEnv me
-  let (fDecls,fNames,fDisp) = M.focusedEnv me
-  return ( dyDecls `mappend` fDecls
+  let (fParams,fDecls,fNames,fDisp) = M.focusedEnv me
+  return ( fParams
+         , dyDecls `mappend` fDecls
          , dyNames `M.shadowing` fNames
          , dyDisp `mappend` fDisp)
 
@@ -390,41 +472,46 @@ getFocusedEnv  = do
 
 getVars :: REPL (Map.Map M.Name M.IfaceDecl)
 getVars  = do
-  (decls,_,_) <- getFocusedEnv
+  (_,decls,_,_) <- getFocusedEnv
   return (M.ifDecls decls)
 
 getTSyns :: REPL (Map.Map M.Name T.TySyn)
 getTSyns  = do
-  (decls,_,_) <- getFocusedEnv
+  (_,decls,_,_) <- getFocusedEnv
   return (M.ifTySyns decls)
 
 getNewtypes :: REPL (Map.Map M.Name T.Newtype)
 getNewtypes = do
-  (decls,_,_) <- getFocusedEnv
+  (_,decls,_,_) <- getFocusedEnv
   return (M.ifNewtypes decls)
 
 -- | Get visible variable names.
 getExprNames :: REPL [String]
 getExprNames =
-  do (_, fNames, _) <- getFocusedEnv
+  do (_,_, fNames, _) <- getFocusedEnv
      return (map (show . pp) (Map.keys (M.neExprs fNames)))
 
 -- | Get visible type signature names.
 getTypeNames :: REPL [String]
 getTypeNames  =
-  do (_, fNames, _) <- getFocusedEnv
+  do (_,_, fNames, _) <- getFocusedEnv
      return (map (show . pp) (Map.keys (M.neTypes fNames)))
 
 -- | Return a list of property names, sorted by position in the file.
 getPropertyNames :: REPL ([M.Name],NameDisp)
 getPropertyNames =
-  do (decls,_,names) <- getFocusedEnv
+  do (_,decls,_,names) <- getFocusedEnv
      let xs = M.ifDecls decls
          ps = sortBy (comparing (from . M.nameLoc))
             $ [ x | (x,d) <- Map.toList xs, T.PragmaProperty `elem` M.ifDeclPragmas d ]
 
      return (ps, names)
 
+getModNames :: REPL [I.ModName]
+getModNames =
+  do me <- getModuleEnv
+     return (map T.mName (M.loadedModules me))
+
 getModuleEnv :: REPL M.ModuleEnv
 getModuleEnv  = eModuleEnv `fmap` getRW
 
@@ -439,6 +526,7 @@ setDynEnv denv = do
   me <- getModuleEnv
   setModuleEnv (me { M.meDynEnv = denv })
 
+
 -- | Given an existing qualified name, prefix it with a
 -- relatively-unique string. We make it unique by prefixing with a
 -- character @#@ that is not lexically valid in a module name.
@@ -446,8 +534,8 @@ uniqify :: M.Name -> REPL M.Name
 
 uniqify name =
   case M.nameInfo name of
-    M.Declared ns ->
-      M.liftSupply (M.mkDeclared ns (M.nameIdent name) (M.nameFixity name) (M.nameLoc name))
+    M.Declared ns s ->
+      M.liftSupply (M.mkDeclared ns s (M.nameIdent name) (M.nameFixity name) (M.nameLoc name))
 
     M.Parameter ->
       panic "[REPL] uniqify" ["tried to uniqify a parameter: " ++ pretty name]
@@ -465,8 +553,9 @@ uniqify name =
 
 -- | Generate a fresh name using the given index. The name will reside within
 -- the "<interactive>" namespace.
-freshName :: I.Ident -> REPL M.Name
-freshName i = M.liftSupply (M.mkDeclared I.interactiveName i Nothing emptyRange)
+freshName :: I.Ident -> M.NameSource -> REPL M.Name
+freshName i sys =
+  M.liftSupply (M.mkDeclared I.interactiveName sys i Nothing emptyRange)
 
 
 -- User Environment Interaction ------------------------------------------------
@@ -492,44 +581,38 @@ setUser :: String -> String -> REPL ()
 setUser name val = case lookupTrieExact name userOptions of
 
   [opt] -> setUserOpt opt
-  []    -> io (putStrLn ("Unknown env value `" ++ name ++ "`"))
-  _     -> io (putStrLn ("Ambiguous env value `" ++ name ++ "`"))
+  []    -> rPutStrLn ("Unknown env value `" ++ name ++ "`")
+  _     -> rPutStrLn ("Ambiguous env value `" ++ name ++ "`")
 
   where
   setUserOpt opt = case optDefault opt of
-    EnvString _ -> do r <- io (optCheck opt (EnvString val))
-                      case r of
-                        Just err -> io (putStrLn err)
-                        Nothing  -> writeEnv (EnvString val)
+    EnvString _ -> doCheck (EnvString val)
 
     EnvProg _ _ ->
       case splitOptArgs val of
-        prog:args -> do r <- io (optCheck opt (EnvProg prog args))
-                        case r of
-                          Just err -> io (putStrLn err)
-                          Nothing  -> writeEnv (EnvProg prog args)
-        []        -> io (putStrLn ("Failed to parse command for field, `" ++ name ++ "`"))
+        prog:args -> doCheck (EnvProg prog args)
+        [] -> rPutStrLn ("Failed to parse command for field, `" ++ name ++ "`")
 
     EnvNum _ -> case reads val of
-      [(x,_)] -> do r <- io (optCheck opt (EnvNum x))
-                    case r of
-                      Just err -> io (putStrLn err)
-                      Nothing  -> writeEnv (EnvNum x)
-
-      _       -> io (putStrLn ("Failed to parse number for field, `" ++ name ++ "`"))
+      [(x,_)] -> doCheck (EnvNum x)
+      _ -> rPutStrLn ("Failed to parse number for field, `" ++ name ++ "`")
 
     EnvBool _
-      | any (`isPrefixOf` val) ["enable","on","yes"] ->
+      | any (`isPrefixOf` val) ["enable", "on", "yes", "true"] ->
         writeEnv (EnvBool True)
-      | any (`isPrefixOf` val) ["disable","off","no"] ->
+      | any (`isPrefixOf` val) ["disable", "off", "no", "false"] ->
         writeEnv (EnvBool False)
       | otherwise ->
-        io (putStrLn ("Failed to parse boolean for field, `" ++ name ++ "`"))
+        rPutStrLn ("Failed to parse boolean for field, `" ++ name ++ "`")
     where
-
+    doCheck v = do (r,ws) <- io (optCheck opt v)
+                   case r of
+                     Just err -> rPutStrLn err
+                     Nothing  -> do mapM_ rPutStrLn ws
+                                    writeEnv v
     writeEnv ev =
       do optEff opt ev
-         modifyRW_ (\rw -> rw { eUserEnv = Map.insert name ev (eUserEnv rw) })
+         modifyRW_ (\rw -> rw { eUserEnv = Map.insert (optName opt) ev (eUserEnv rw) })
 
 splitOptArgs :: String -> [String]
 splitOptArgs  = unfoldr (parse "")
@@ -581,50 +664,58 @@ mkOptionMap  = foldl insert emptyTrie
   where
   insert m d = insertTrie (optName d) d m
 
+-- | Returns maybe an error, and some warnings
+type Checker = EnvVal -> IO (Maybe String, [String])
+
+noCheck :: Checker
+noCheck _ = return (Nothing, [])
+
+noWarns :: Maybe String -> IO (Maybe String, [String])
+noWarns mb = return (mb, [])
+
 data OptionDescr = OptionDescr
   { optName    :: String
   , optDefault :: EnvVal
-  , optCheck   :: EnvVal -> IO (Maybe String)
+  , optCheck   :: Checker
   , optHelp    :: String
   , optEff     :: EnvVal -> REPL ()
   }
 
-simpleOpt :: String -> EnvVal -> (EnvVal -> IO (Maybe String)) -> String
-          -> OptionDescr
+simpleOpt :: String -> EnvVal -> Checker -> String -> OptionDescr
 simpleOpt optName optDefault optCheck optHelp =
   OptionDescr { optEff = \ _ -> return (), .. }
 
 userOptions :: OptionMap
 userOptions  = mkOptionMap
   [ simpleOpt "base" (EnvNum 16) checkBase
-    "the base to display words at"
-  , simpleOpt "debug" (EnvBool False) (const $ return Nothing)
-    "enable debugging output"
-  , simpleOpt "ascii" (EnvBool False) (const $ return Nothing)
-    "display 7- or 8-bit words using ASCII notation."
+    "The base to display words at (2, 8, 10, or 16)."
+  , simpleOpt "debug" (EnvBool False) noCheck
+    "Enable debugging output."
+  , simpleOpt "ascii" (EnvBool False) noCheck
+    "Whether to display 7- or 8-bit words using ASCII notation."
   , simpleOpt "infLength" (EnvNum 5) checkInfLength
     "The number of elements to display for infinite sequences."
-  , simpleOpt "tests" (EnvNum 100) (const $ return Nothing)
-    "The number of random tests to try."
+  , simpleOpt "tests" (EnvNum 100) noCheck
+    "The number of random tests to try with ':check'."
   , simpleOpt "satNum" (EnvString "1") checkSatNum
     "The maximum number of :sat solutions to display (\"all\" for no limit)."
   , simpleOpt "prover" (EnvString "z3") checkProver $
-    "The external SMT solver for :prove and :sat (" ++ proverListString ++ ")."
-  , simpleOpt "warnDefaulting" (EnvBool True) (const $ return Nothing)
-    "Choose if we should display warnings when defaulting."
-  , simpleOpt "warnShadowing" (EnvBool True) (const $ return Nothing)
-    "Choose if we should display warnings when shadowing symbols."
-  , simpleOpt "smtfile" (EnvString "-") (const $ return Nothing)
-    "The file to use for SMT-Lib scripts (for debugging or offline proving)"
-  , OptionDescr "mono-binds" (EnvBool True) (const $ return Nothing)
-    "Whether or not to generalize bindings in a where-clause" $
+    "The external SMT solver for ':prove' and ':sat'\n(" ++ proverListString ++ ")."
+  , simpleOpt "warnDefaulting" (EnvBool True) noCheck
+    "Choose whether to display warnings when defaulting."
+  , simpleOpt "warnShadowing" (EnvBool True) noCheck
+    "Choose whether to display warnings when shadowing symbols."
+  , simpleOpt "smtfile" (EnvString "-") noCheck
+    "The file to use for SMT-Lib scripts (for debugging or offline proving).\nUse \"-\" for stdout."
+  , OptionDescr "mono-binds" (EnvBool True) noCheck
+    "Whether or not to generalize bindings in a 'where' clause." $
     \case EnvBool b -> do me <- getModuleEnv
                           setModuleEnv me { M.meMonoBinds = b }
           _         -> return ()
 
   , OptionDescr "tc-solver" (EnvProg "z3" [ "-smt2", "-in" ])
-    (const (return Nothing)) -- TODO: check for the program in the path
-    "The solver that will be used by the type checker" $
+    noCheck  -- TODO: check for the program in the path
+    "The solver that will be used by the type checker." $
     \case EnvProg prog args -> do me <- getModuleEnv
                                   let cfg = M.meSolverConfig me
                                   setModuleEnv me { M.meSolverConfig =
@@ -633,65 +724,68 @@ userOptions  = mkOptionMap
           _                 -> return ()
 
   , OptionDescr "tc-debug" (EnvNum 0)
-    (const (return Nothing))
-    "Enable type-checker debugging output" $
+    noCheck
+    "Enable type-checker debugging output." $
     \case EnvNum n -> do me <- getModuleEnv
                          let cfg = M.meSolverConfig me
                          setModuleEnv me { M.meSolverConfig = cfg{ T.solverVerbose = fromIntegral n } }
           _        -> return ()
   , OptionDescr "core-lint" (EnvBool False)
-    (const (return Nothing))
-    "Enable sanity checking of type-checker" $
+    noCheck
+    "Enable sanity checking of type-checker." $
       let setIt x = do me <- getModuleEnv
                        setModuleEnv me { M.meCoreLint = x }
       in \case EnvBool True  -> setIt M.CoreLint
                EnvBool False -> setIt M.NoCoreLint
                _             -> return ()
 
-  , simpleOpt "prover-stats" (EnvBool True) (const (return Nothing))
+  , simpleOpt "prover-stats" (EnvBool True) noCheck
     "Enable prover timing statistics."
   ]
 
 
 -- | Check the value to the `base` option.
-checkBase :: EnvVal -> IO (Maybe String)
+checkBase :: Checker
 checkBase val = case val of
   EnvNum n
-    | n >= 2 && n <= 36 -> return Nothing
-    | otherwise         -> return $ Just "base must fall between 2 and 36"
-  _                     -> return $ Just "unable to parse a value for base"
+    | n `elem` [2,8,10,16] -> noWarns Nothing
+    | otherwise            -> noWarns $ Just "base must be 2, 8, 10, or 16"
+  _                     -> noWarns $ Just "unable to parse a value for base"
 
-checkInfLength :: EnvVal -> IO (Maybe String)
+checkInfLength :: Checker
 checkInfLength val = case val of
   EnvNum n
-    | n >= 0    -> return Nothing
-    | otherwise -> return $ Just "the number of elements should be positive"
-  _ -> return $ Just "unable to parse a value for infLength"
+    | n >= 0    -> noWarns Nothing
+    | otherwise -> noWarns $ Just "the number of elements should be positive"
+  _ -> noWarns $ Just "unable to parse a value for infLength"
 
-checkProver :: EnvVal -> IO (Maybe String)
+checkProver :: Checker
 checkProver val = case val of
   EnvString s
-    | s `notElem` proverNames     -> return $ Just $ "Prover must be " ++ proverListString
-    | s `elem` ["offline", "any"] -> return Nothing
-    | otherwise                   -> do let prover = lookupProver s
-                                        available <- sbvCheckSolverInstallation prover
-                                        unless available $
-                                          putStrLn $ "Warning: " ++ s ++ " installation not found"
-                                        return Nothing
+    | s `notElem` proverNames ->
+      noWarns $ Just $ "Prover must be " ++ proverListString
+    | s `elem` ["offline", "any"] -> noWarns Nothing
+    | otherwise ->
+      do let prover = lookupProver s
+         available <- sbvCheckSolverInstallation prover
+         let ws = if available
+                     then []
+                     else ["Warning: " ++ s ++ " installation not found"]
+         return (Nothing, ws)
 
-  _ -> return $ Just "unable to parse a value for prover"
+  _ -> noWarns $ Just "unable to parse a value for prover"
 
 proverListString :: String
 proverListString = concatMap (++ ", ") (init proverNames) ++ "or " ++ last proverNames
 
-checkSatNum :: EnvVal -> IO (Maybe String)
+checkSatNum :: Checker
 checkSatNum val = case val of
-  EnvString "all" -> return Nothing
+  EnvString "all" -> noWarns Nothing
   EnvString s ->
     case readMaybe s :: Maybe Int of
-      Just n | n >= 1 -> return Nothing
-      _               -> return $ Just "must be an integer > 0 or \"all\""
-  _ -> return $ Just "unable to parse a value for satNum"
+      Just n | n >= 1 -> noWarns Nothing
+      _               -> noWarns $ Just "must be an integer > 0 or \"all\""
+  _ -> noWarns $ Just "unable to parse a value for satNum"
 
 getUserSatNum :: REPL SatNum
 getUserSatNum = do
