@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Symbolic
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -18,6 +18,7 @@ module Cryptol.Symbolic where
 
 import Control.Monad.IO.Class
 import Control.Monad (replicateM, when, zipWithM, foldM)
+import Control.Monad.Writer (WriterT, runWriterT, tell, lift)
 import Data.List (intercalate, genericLength)
 import Data.IORef(IORef)
 import qualified Control.Exception as X
@@ -43,6 +44,7 @@ import Cryptol.TypeCheck.AST
 import Cryptol.Utils.Ident (Ident)
 import Cryptol.Utils.PP
 import Cryptol.Utils.Panic(panic)
+import Cryptol.Utils.Logger(logPutStrLn)
 
 import Prelude ()
 import Prelude.Compat
@@ -124,13 +126,14 @@ thmSMTResults :: SBV.ThmResult -> [SBV.SMTResult]
 thmSMTResults (SBV.ThmResult r) = [r]
 
 proverError :: String -> M.ModuleCmd (Maybe SBV.Solver, ProverResult)
-proverError msg modEnv =
+proverError msg (_,modEnv) =
   return (Right ((Nothing, ProverError msg), modEnv), [])
 
 satProve :: ProverCommand -> M.ModuleCmd (Maybe SBV.Solver, ProverResult)
 satProve ProverCommand {..} =
-  protectStack proverError $ \modEnv ->
-  M.runModuleM modEnv $ do
+  protectStack proverError $ \(evo,modEnv) ->
+
+  M.runModuleM (evo,modEnv) $ do
   let (isSat, mSatNum) = case pcQueryType of
         ProveQuery -> (False, Nothing)
         SatQuery sn -> case sn of
@@ -145,14 +148,19 @@ satProve ProverCommand {..} =
 
   let provers' = [ p { SBV.timing = SaveTiming pcProverStats, SBV.verbose = pcVerbose } | p <- provers ]
   let tyFn = if isSat then existsFinType else forallFinType
+  let lPutStrLn = M.withLogger logPutStrLn
+  let doEval :: MonadIO m => Eval.Eval a -> m a
+      doEval m  = liftIO $ Eval.runEval evo m
   let runProver fn tag e = do
         case provers of
           [prover] -> do
-            when pcVerbose $ M.io $
-              putStrLn $ "Trying proof with " ++ show (SBV.name (SBV.solver prover))
+            when pcVerbose $
+              lPutStrLn $ "Trying proof with " ++
+                                        show (SBV.name (SBV.solver prover))
             res <- M.io (fn prover e)
-            when pcVerbose $ M.io $
-              putStrLn $ "Got result from " ++ show (SBV.name (SBV.solver prover))
+            when pcVerbose $
+              lPutStrLn $ "Got result from " ++
+                                        show (SBV.name (SBV.solver prover))
             return (Just (SBV.name (SBV.solver prover)), tag res)
           _ ->
             return ( Nothing
@@ -162,30 +170,33 @@ satProve ProverCommand {..} =
                      | prover <- provers ]
                    )
       runProvers fn tag e = do
-        when pcVerbose $ M.io $
-          putStrLn $ "Trying proof with " ++
-                     intercalate ", " (map (show . SBV.name . SBV.solver) provers)
+        when pcVerbose $
+          lPutStrLn $ "Trying proof with " ++
+                  intercalate ", " (map (show . SBV.name . SBV.solver) provers)
         (firstProver, timeElapsed, res) <- M.io (fn provers' e)
-        when pcVerbose $ M.io $
-          putStrLn $ "Got result from " ++ show firstProver ++ ", time: " ++ showTDiff timeElapsed
+        when pcVerbose $
+          lPutStrLn $ "Got result from " ++ show firstProver ++
+                                            ", time: " ++ showTDiff timeElapsed
         return (Just firstProver, tag res)
   let runFn = case pcQueryType of
         ProveQuery -> runProvers SBV.proveWithAny thmSMTResults
         SatQuery sn -> case sn of
           SomeSat 1 -> runProvers SBV.satWithAny satSMTResults
           _         -> runProver SBV.allSatWith allSatSMTResults
+  let addAsm = case pcQueryType of
+        ProveQuery -> \x y -> SBV.svOr (SBV.svNot x) y
+        SatQuery _ -> \x y -> SBV.svAnd x y
   case predArgTypes pcSchema of
     Left msg -> return (Nothing, ProverError msg)
-    Right ts -> do when pcVerbose $ M.io $ putStrLn "Simulating..."
-                   v <- M.io $ Eval.runEval $ do
-                               env <- Eval.evalDecls extDgs mempty
-                               Eval.evalExpr env pcExpr
+    Right ts -> do when pcVerbose $ lPutStrLn "Simulating..."
+                   v <- doEval $ do env <- Eval.evalDecls extDgs mempty
+                                    Eval.evalExpr env pcExpr
                    prims <- M.getPrimMap
                    runRes <- runFn $ do
-                               args <- mapM tyFn ts
-                               b <- liftIO $ Eval.runEval
-                                      (fromVBit <$> foldM fromVFun v (map Eval.ready args))
-                               return b
+                               (args, asms) <- runWriterT (mapM tyFn ts)
+                               b <- doEval (fromVBit <$>
+                                      foldM fromVFun v (map Eval.ready args))
+                               return (foldr addAsm b asms)
                    let (firstProver, results') = runRes
                        results = maybe results' (\n -> take n results') mSatNum
                    esatexprs <- case results of
@@ -199,8 +210,8 @@ satProve ProverCommand {..} =
                            let Right (_, cws) = SBV.getModelAssignment result
                                (vs, _) = parseValues ts cws
                                sattys = unFinType <$> ts
-                           satexprs <- liftIO $ Eval.runEval
-                                           (zipWithM (Eval.toExpr prims) sattys vs)
+                           satexprs <-
+                             doEval (zipWithM (Eval.toExpr prims) sattys vs)
                            case zip3 sattys <$> (sequence satexprs) <*> pure vs of
                              Nothing ->
                                panic "Cryptol.Symbolic.sat"
@@ -221,23 +232,26 @@ satProve ProverCommand {..} =
 
 satProveOffline :: ProverCommand -> M.ModuleCmd (Either String String)
 satProveOffline ProverCommand {..} =
-  protectStack (\msg modEnv -> return (Right (Left msg, modEnv), [])) $ \modEnv -> do
+  protectStack (\msg (_,modEnv) -> return (Right (Left msg, modEnv), [])) $
+  \(evOpts,modEnv) -> do
     let isSat = case pcQueryType of
           ProveQuery -> False
           SatQuery _ -> True
     let extDgs = allDeclGroups modEnv ++ pcExtraDecls
     let tyFn = if isSat then existsFinType else forallFinType
+    let addAsm = if isSat then SBV.svAnd else \x y -> SBV.svOr (SBV.svNot x) y
     case predArgTypes pcSchema of
       Left msg -> return (Right (Left msg, modEnv), [])
       Right ts ->
-        do when pcVerbose $ putStrLn "Simulating..."
-           v <- liftIO $ Eval.runEval $
+        do when pcVerbose $ logPutStrLn (Eval.evalLogger evOpts) "Simulating..."
+           v <- liftIO $ Eval.runEval evOpts $
                    do env <- Eval.evalDecls extDgs mempty
                       Eval.evalExpr env pcExpr
            smtlib <- SBV.generateSMTBenchmark isSat $ do
-             args <- mapM tyFn ts
-             liftIO $ Eval.runEval
-                    (fromVBit <$> foldM fromVFun v (map Eval.ready args))
+             (args, asms) <- runWriterT (mapM tyFn ts)
+             b <- liftIO $ Eval.runEval evOpts
+                        (fromVBit <$> foldM fromVFun v (map Eval.ready args))
+             return (foldr addAsm b asms)
            return (Right (Right smtlib, modEnv), [])
 
 protectStack :: (String -> M.ModuleCmd a)
@@ -259,6 +273,11 @@ parseValues (t : ts) cws = (v : vs, cws'')
 parseValue :: FinType -> [SBV.CW] -> (Eval.Value, [SBV.CW])
 parseValue FTBit [] = panic "Cryptol.Symbolic.parseValue" [ "empty FTBit" ]
 parseValue FTBit (cw : cws) = (Eval.VBit (SBV.cwToBool cw), cws)
+parseValue FTInteger cws =
+  case SBV.genParse SBV.KUnbounded cws of
+    Just (x, cws') -> (Eval.VInteger x, cws')
+    Nothing        -> panic "Cryptol.Symbolic.parseValue" [ "no integer" ]
+parseValue (FTIntMod _) cws = parseValue FTInteger cws
 parseValue (FTSeq 0 FTBit) cws = (Eval.word 0 0, cws)
 parseValue (FTSeq n FTBit) cws =
   case SBV.genParse (SBV.KBounded False n) cws of
@@ -282,6 +301,8 @@ allDeclGroups = concatMap mDecls . M.loadedModules
 
 data FinType
     = FTBit
+    | FTInteger
+    | FTIntMod Integer
     | FTSeq Int FinType
     | FTTuple [FinType]
     | FTRecord [(Ident, FinType)]
@@ -295,6 +316,8 @@ finType :: TValue -> Maybe FinType
 finType ty =
   case ty of
     Eval.TVBit            -> Just FTBit
+    Eval.TVInteger        -> Just FTInteger
+    Eval.TVIntMod n       -> Just (FTIntMod n)
     Eval.TVSeq n t        -> FTSeq <$> numType n <*> finType t
     Eval.TVTuple ts       -> FTTuple <$> traverse finType ts
     Eval.TVRec fields     -> FTRecord <$> traverse (traverseSnd finType) fields
@@ -304,6 +327,8 @@ unFinType :: FinType -> Type
 unFinType fty =
   case fty of
     FTBit        -> tBit
+    FTInteger    -> tInteger
+    FTIntMod n   -> tIntMod (tNum n)
     FTSeq l ety  -> tSeq (tNum l) (unFinType ety)
     FTTuple ftys -> tTuple (unFinType <$> ftys)
     FTRecord fs  -> tRec (zip fns tys)
@@ -316,7 +341,7 @@ predArgTypes schema@(Forall ts ps ty)
   | null ts && null ps =
       case go <$> (Eval.evalType mempty ty) of
         Right (Just fts) -> Right fts
-        _                -> Left $ "Not a valid predicate type:\n" ++ show (pp schema)
+        _ -> Left $ "Not a valid predicate type:\n" ++ show (pp schema)
   | otherwise = Left $ "Not a monomorphic type:\n" ++ show (pp schema)
   where
     go :: TValue -> Maybe [FinType]
@@ -324,23 +349,35 @@ predArgTypes schema@(Forall ts ps ty)
     go (Eval.TVFun ty1 ty2)   = (:) <$> finType ty1 <*> go ty2
     go _                      = Nothing
 
-forallFinType :: FinType -> SBV.Symbolic Value
+inBoundsIntMod :: Integer -> SInteger -> SBool
+inBoundsIntMod n x =
+  SBV.svAnd (SBV.svLessEq (Eval.integerLit 0) x) (SBV.svLessThan x (Eval.integerLit n))
+
+forallFinType :: FinType -> WriterT [SBool] SBV.Symbolic Value
 forallFinType ty =
   case ty of
-    FTBit         -> VBit <$> forallSBool_
+    FTBit         -> VBit <$> lift forallSBool_
+    FTInteger     -> VInteger <$> lift forallSInteger_
+    FTIntMod n    -> do x <- lift forallSInteger_
+                        tell [inBoundsIntMod n x]
+                        return (VInteger x)
     FTSeq 0 FTBit -> return $ Eval.word 0 0
-    FTSeq n FTBit -> VWord (toInteger n) . return . Eval.WordVal <$> (forallBV_ n)
+    FTSeq n FTBit -> VWord (toInteger n) . return . Eval.WordVal <$> lift (forallBV_ n)
     FTSeq n t     -> do vs <- replicateM n (forallFinType t)
                         return $ VSeq (toInteger n) $ Eval.finiteSeqMap (map Eval.ready vs)
     FTTuple ts    -> VTuple <$> mapM (fmap Eval.ready . forallFinType) ts
     FTRecord fs   -> VRecord <$> mapM (traverseSnd (fmap Eval.ready . forallFinType)) fs
 
-existsFinType :: FinType -> SBV.Symbolic Value
+existsFinType :: FinType -> WriterT [SBool] SBV.Symbolic Value
 existsFinType ty =
   case ty of
-    FTBit         -> VBit <$> existsSBool_
+    FTBit         -> VBit <$> lift existsSBool_
+    FTInteger     -> VInteger <$> lift existsSInteger_
+    FTIntMod n    -> do x <- lift existsSInteger_
+                        tell [inBoundsIntMod n x]
+                        return (VInteger x)
     FTSeq 0 FTBit -> return $ Eval.word 0 0
-    FTSeq n FTBit -> VWord (toInteger n) . return . Eval.WordVal <$> (existsBV_ n)
+    FTSeq n FTBit -> VWord (toInteger n) . return . Eval.WordVal <$> lift (existsBV_ n)
     FTSeq n t     -> do vs <- replicateM n (existsFinType t)
                         return $ VSeq (toInteger n) $ Eval.finiteSeqMap (map Eval.ready vs)
     FTTuple ts    -> VTuple <$> mapM (fmap Eval.ready . existsFinType) ts
