@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Symbolic.Prims
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -20,16 +20,15 @@ module Cryptol.Symbolic.Prims where
 
 import Control.Monad (unless)
 import Data.Bits
-import Data.List (genericTake, sortBy)
-import Data.Ord (comparing)
 import qualified Data.Sequence as Seq
 import qualified Data.Foldable as Fold
 
-import Cryptol.Eval.Monad (Eval(..), ready, invalidIndex)
+import Cryptol.Eval.Monad (Eval(..), ready, invalidIndex, cryUserError)
 import Cryptol.Eval.Type  (finNat', TValue(..))
 import Cryptol.Eval.Value (BitWord(..), EvalPrims(..), enumerateSeqMap, SeqMap(..),
                           reverseSeqMap, wlam, nlam, WordValue(..),
-                          asWordVal, asBitsVal, fromWordVal,
+                          asWordVal, fromWordVal, enumerateWordValue,
+                          updateWordValue,
                           updateSeqMap, lookupSeqMap, memoMap )
 import Cryptol.Prims.Eval (binary, unary, arithUnary,
                            arithBinary, Binary, BinArith,
@@ -37,8 +36,10 @@ import Cryptol.Prims.Eval (binary, unary, arithUnary,
                            ccatV, splitAtV, joinV, ecSplitV,
                            reverseV, infFromV, infFromThenV,
                            fromThenV, fromToV, fromThenToV,
-                           transposeV, indexPrimOne, indexPrimMany,
-                           ecDemoteV, updatePrim)
+                           transposeV, indexPrim,
+                           ecToIntegerV, ecFromIntegerV,
+                           ecNumberV, updatePrim, randomV, liftWord,
+                           cmpValue, lg2)
 import Cryptol.Symbolic.Value
 import Cryptol.TypeCheck.AST (Decl(..))
 import Cryptol.TypeCheck.Solver.InfNat(Nat'(..))
@@ -60,43 +61,68 @@ traverseSnd f (x, y) = (,) x <$> f y
 
 -- Primitives ------------------------------------------------------------------
 
-instance EvalPrims SBool SWord where
+instance EvalPrims SBool SWord SInteger where
   evalPrim Decl { dName = n, .. }
     | Just prim <- asPrim n, Just val <- Map.lookup prim primTable = val
 
   evalPrim Decl { .. } =
       panic "Eval" [ "Unimplemented primitive", show dName ]
 
-  iteValue b x y
-    | Just b' <- SBV.svAsBool b = if b' then x else y
-    | otherwise = iteSValue b <$> x <*> y
+  iteValue b x1 x2
+    | Just b' <- SBV.svAsBool b = if b' then x1 else x2
+    | otherwise = do v1 <- x1
+                     v2 <- x2
+                     iteSValue b v1 v2
 
 -- See also Cryptol.Prims.Eval.primTable
 primTable :: Map.Map Ident Value
 primTable  = Map.fromList $ map (\(n, v) -> (mkIdent (T.pack n), v))
   [ ("True"        , VBit SBV.svTrue)
   , ("False"       , VBit SBV.svFalse)
-  , ("demote"      , ecDemoteV) -- Converts a numeric type into its corresponding value.
-                                -- { val, bits } (fin val, fin bits, bits >= width val) => [bits]
-  , ("+"           , binary (arithBinary (liftBinArith SBV.svPlus))) -- {a} (Arith a) => a -> a -> a
-  , ("-"           , binary (arithBinary (liftBinArith SBV.svMinus))) -- {a} (Arith a) => a -> a -> a
-  , ("*"           , binary (arithBinary (liftBinArith SBV.svTimes))) -- {a} (Arith a) => a -> a -> a
-  , ("/"           , binary (arithBinary (liftBinArith SBV.svQuot))) -- {a} (Arith a) => a -> a -> a
-  , ("%"           , binary (arithBinary (liftBinArith SBV.svRem))) -- {a} (Arith a) => a -> a -> a
-  , ("^^"          , binary (arithBinary sExp)) -- {a} (Arith a) => a -> a -> a
-  , ("lg2"         , unary (arithUnary sLg2)) -- {a} (Arith a) => a -> a
-  , ("negate"      , unary (arithUnary (\_ -> ready . SBV.svUNeg)))
-  , ("<"           , binary (cmpBinary cmpLt cmpLt SBV.svFalse))
-  , (">"           , binary (cmpBinary cmpGt cmpGt SBV.svFalse))
-  , ("<="          , binary (cmpBinary cmpLtEq cmpLtEq SBV.svTrue))
-  , (">="          , binary (cmpBinary cmpGtEq cmpGtEq SBV.svTrue))
-  , ("=="          , binary (cmpBinary cmpEq cmpEq SBV.svTrue))
-  , ("!="          , binary (cmpBinary cmpNotEq cmpNotEq SBV.svFalse))
+  , ("number"      , ecNumberV) -- Converts a numeric type into its corresponding value.
+                                -- { val, rep } (Literal val rep) => rep
+  , ("+"           , binary (arithBinary (liftBinArith SBV.svPlus) (liftBin SBV.svPlus)
+                             (const (liftBin SBV.svPlus)))) -- {a} (Arith a) => a -> a -> a
+  , ("-"           , binary (arithBinary (liftBinArith SBV.svMinus) (liftBin SBV.svMinus)
+                             (const (liftBin SBV.svMinus)))) -- {a} (Arith a) => a -> a -> a
+  , ("*"           , binary (arithBinary (liftBinArith SBV.svTimes) (liftBin SBV.svTimes)
+                             (const (liftBin SBV.svTimes)))) -- {a} (Arith a) => a -> a -> a
+  , ("/"           , binary (arithBinary (liftBinArith SBV.svQuot) (liftBin SBV.svQuot)
+                             (liftModBin SBV.svQuot))) -- {a} (Arith a) => a -> a -> a
+  , ("%"           , binary (arithBinary (liftBinArith SBV.svRem) (liftBin SBV.svRem)
+                             (liftModBin SBV.svRem))) -- {a} (Arith a) => a -> a -> a
+  , ("^^"          , binary (arithBinary sExp (liftBin SBV.svExp)
+                             sModExp)) -- {a} (Arith a) => a -> a -> a
+  , ("lg2"         , unary (arithUnary sLg2 svLg2 svModLg2)) -- {a} (Arith a) => a -> a
+  , ("negate"      , unary (arithUnary (\_ -> ready . SBV.svUNeg) (ready . SBV.svUNeg)
+                            (const (ready . SBV.svUNeg))))
+  , ("<"           , binary (cmpBinary cmpLt cmpLt cmpLt (cmpMod cmpLt) SBV.svFalse))
+  , (">"           , binary (cmpBinary cmpGt cmpGt cmpGt (cmpMod cmpGt) SBV.svFalse))
+  , ("<="          , binary (cmpBinary cmpLtEq cmpLtEq cmpLtEq (cmpMod cmpLtEq) SBV.svTrue))
+  , (">="          , binary (cmpBinary cmpGtEq cmpGtEq cmpGtEq (cmpMod cmpGtEq) SBV.svTrue))
+  , ("=="          , binary (cmpBinary cmpEq cmpEq cmpEq cmpModEq SBV.svTrue))
+  , ("!="          , binary (cmpBinary cmpNotEq cmpNotEq cmpNotEq cmpModNotEq SBV.svFalse))
+  , ("<$"          , let boolFail = evalPanic "<$" ["Attempted signed comparison on bare Bit values"]
+                         intFail = evalPanic "<$" ["Attempted signed comparison on Integer values"]
+                      in binary (cmpBinary boolFail cmpSignedLt intFail (const intFail) SBV.svFalse))
+  , ("/$"          , binary (arithBinary (liftBinArith signedQuot) (liftBin SBV.svQuot)
+                             (liftModBin SBV.svQuot))) -- {a} (Arith a) => a -> a -> a
+  , ("%$"          , binary (arithBinary (liftBinArith signedRem) (liftBin SBV.svRem)
+                             (liftModBin SBV.svRem)))
+  , (">>$"         , sshrV)
   , ("&&"          , binary (logicBinary SBV.svAnd SBV.svAnd))
   , ("||"          , binary (logicBinary SBV.svOr SBV.svOr))
   , ("^"           , binary (logicBinary SBV.svXOr SBV.svXOr))
   , ("complement"  , unary (logicUnary SBV.svNot SBV.svNot))
   , ("zero"        , tlam zeroV)
+  , ("toInteger"   , ecToIntegerV)
+  , ("fromInteger" , ecFromIntegerV (const id))
+  , ("fromZ"      , nlam $ \ modulus ->
+                    lam  $ \ x -> do
+                      val <- x
+                      case (modulus, val) of
+                        (Nat n, VInteger i) -> return $ VInteger (SBV.svRem i (integerLit n))
+                        _                   -> evalPanic "fromZ" ["Invalid arguments"])
   , ("<<"          , logicShift "<<"
                        SBV.svShiftLeft
                        (\sz i shft ->
@@ -122,6 +148,9 @@ primTable  = Map.fromList $ map (\(n, v) -> (mkIdent (T.pack n), v))
                             Inf -> evalPanic "cannot rotate infinite sequence" []
                             Nat n -> Just ((i+n-shft) `mod` n)))
 
+  , ("carry"      , liftWord carry)
+  , ("scarry"     , liftWord scarry)
+
   , ("#"          , -- {a,b,d} (fin a) => [a] d -> [b] d -> [a + b] d
      nlam $ \ front ->
      nlam $ \ back  ->
@@ -160,46 +189,11 @@ primTable  = Map.fromList $ map (\(n, v) -> (mkIdent (T.pack n), v))
   , ("infFrom"     , infFromV)
   , ("infFromThen" , infFromThenV)
 
-  , ("@"           , indexPrimOne  indexFront_bits indexFront)
-  , ("@@"          , indexPrimMany indexFront_bits indexFront)
-  , ("!"           , indexPrimOne  indexBack_bits indexBack)
-  , ("!!"          , indexPrimMany indexBack_bits indexBack)
+  , ("@"           , indexPrim indexFront_bits indexFront)
+  , ("!"           , indexPrim indexBack_bits indexBack)
 
-  , ("update"      , updatePrim updateFrontSym_bits updateFrontSym)
-  , ("updateEnd"   , updatePrim updateBackSym_bits updateBackSym)
-
-  , ("pmult"       , -- {a,b} (fin a, fin b) => [1 + a] -> [1 + b] -> [1 + a + b]
-      nlam $ \(finNat' -> i) ->
-      nlam $ \(finNat' -> j) ->
-      VFun $ \v1 -> return $
-      VFun $ \v2 -> do
-        let k = 1 + i + j
-            mul _  []     ps = ps
-            mul as (b:bs) ps = mul (SBV.svFalse : as) bs (ites b (as `addPoly` ps) ps)
-        xs <- sequence . Fold.toList . asBitsVal =<< fromWordVal "pmult 1" =<< v1
-        ys <- sequence . Fold.toList . asBitsVal =<< fromWordVal "pmult 2" =<< v2
-        let zs = genericTake k (mul xs ys [] ++ repeat SBV.svFalse)
-        return $ VWord k $ return $ BitsVal $ Seq.fromList $ map ready zs)
-
-  , ("pdiv"        , -- {a,b} (fin a, fin b) => [a] -> [b] -> [a]
-      nlam $ \(finNat' -> i) ->
-      nlam $ \(finNat' -> _j) ->
-      VFun $ \v1 -> return $
-      VFun $ \v2 -> do
-        xs <- sequence . Fold.toList . Seq.reverse . asBitsVal =<< fromWordVal "pdiv 1" =<< v1
-        ys <- sequence . Fold.toList . Seq.reverse . asBitsVal =<< fromWordVal "pdiv 2" =<< v2
-        let zs = genericTake i (fst (mdp xs ys) ++ repeat SBV.svFalse)
-        return $ VWord i $ return $ BitsVal $ Seq.reverse $ Seq.fromList $ map ready zs)
-
-  , ("pmod"        , -- {a,b} (fin a, fin b) => [a] -> [b+1] -> [b]
-      nlam $ \(finNat' -> _i) ->
-      nlam $ \(finNat' -> j) ->
-      VFun $ \v1 -> return $
-      VFun $ \v2 -> do
-        xs <- sequence . Fold.toList . Seq.reverse . asBitsVal =<< fromWordVal "pmod 1" =<< v1
-        ys <- sequence . Fold.toList . Seq.reverse . asBitsVal =<< fromWordVal "pmod 2" =<< v2
-        let zs = genericTake j (snd (mdp xs ys) ++ repeat SBV.svFalse)
-        return $ VWord j $ return $ BitsVal $ Seq.reverse $ Seq.fromList $ map ready zs)
+  , ("update"      , updatePrim updateFrontSym_word updateFrontSym)
+  , ("updateEnd"   , updatePrim updateBackSym_word updateBackSym)
 
     -- {at,len} (fin len) => [len][8] -> at
   , ("error"       ,
@@ -209,11 +203,11 @@ primTable  = Map.fromList $ map (\(n, v) -> (mkIdent (T.pack n), v))
         return $ zeroV at) -- error/undefined, is arbitrarily translated to 0
 
   , ("random"      ,
-      tlam $ \_a ->
-      wlam $ \_x ->
-         Thunk $ return $ panic
-            "Cryptol.Symbolic.Prims.evalECon"
-            [ "can't symbolically evaluate ECRandom" ])
+      tlam $ \a ->
+      wlam $ \x ->
+         case SBV.svAsInteger x of
+           Just i  -> return $ randomV a i
+           Nothing -> cryUserError "cannot evaluate 'random' with symbolic inputs")
 
      -- The trace function simply forces its first two
      -- values before returing the third in the symbolic
@@ -231,13 +225,6 @@ primTable  = Map.fromList $ map (\(n, v) -> (mkIdent (T.pack n), v))
   ]
 
 
-iteWord :: SBool
-        -> Eval (WordValue SBool SWord)
-        -> Eval (WordValue SBool SWord)
-        -> Eval (WordValue SBool SWord)
-iteWord c x y = mergeWord True c <$> x <*> y
-
-
 -- | Barrel-shifter algorithm. Takes a list of bits in big-endian order.
 shifter :: Monad m => (SBool -> a -> a -> a) -> (a -> Integer -> m a) -> a -> [SBool] -> m a
 shifter mux op = go
@@ -264,16 +251,23 @@ logicShift nm wop reindex =
              return $ VWord w $ do
                x >>= \case
                  WordVal x' -> WordVal . wop x' <$> asWordVal idx
-                 wv ->
-                   do idx_bits <- sequence $ Fold.toList $ asBitsVal idx
+                 BitsVal bs0 ->
+                   do idx_bits <- enumerateWordValue idx
                       let op bs shft = return $ Seq.fromFunction (Seq.length bs) $ \i ->
                                              case reindex (Nat w) (toInteger i) shft of
                                                Nothing -> return $ bitLit False
                                                Just i' -> Seq.index bs (fromInteger i')
-                      BitsVal <$> shifter (mergeBits True) op (asBitsVal wv) idx_bits
+                      BitsVal <$> shifter (mergeBits True) op bs0 idx_bits
+                 LargeBitsVal n bs0 ->
+                   do idx_bits <- enumerateWordValue idx
+                      let op bs shft = memoMap $ IndexSeqMap $ \i ->
+                                         case reindex (Nat w) i shft of
+                                           Nothing -> return $ VBit $ bitLit False
+                                           Just i' -> lookupSeqMap bs i'
+                      LargeBitsVal n <$> shifter (mergeSeqMap True) op bs0 idx_bits
 
-          VSeq w vs0  ->
-             do idx_bits <- sequence $ Fold.toList $ asBitsVal idx
+          VSeq w vs0 ->
+             do idx_bits <- enumerateWordValue idx
                 let op vs shft = memoMap $ IndexSeqMap $ \i ->
                                    case reindex (Nat w) i shft of
                                      Nothing -> return $ zeroV a
@@ -281,7 +275,7 @@ logicShift nm wop reindex =
                 VSeq w <$> shifter (mergeSeqMap True) op vs0 idx_bits
 
           VStream vs0 ->
-             do idx_bits <- sequence $ Fold.toList $ asBitsVal idx
+             do idx_bits <- enumerateWordValue idx
                 let op vs shft = memoMap $ IndexSeqMap $ \i ->
                                    case reindex Inf i shft of
                                      Nothing -> return $ zeroV a
@@ -292,15 +286,16 @@ logicShift nm wop reindex =
 
 
 selectV :: forall a
-         . (SBool -> Eval a -> Eval a -> Eval a)
-        -> WordValue SBool SWord
-        -> (Integer -> Eval a)
-        -> Eval a
+         . (SBool -> Eval a -> Eval a -> Eval a) -- ^ Mux function on @a@s
+        -> WordValue SBool SWord SInteger        -- ^ Symbolic index value
+        -> (Integer -> Eval a)                   -- ^ Function from concrete indices to answers
+        -> Eval a                                -- ^ overall answer
 selectV mux val f =
    case val of
      WordVal x | Just idx <- SBV.svAsInteger x -> f idx
                | otherwise -> sel 0 (unpackWord x)
      BitsVal bs -> sel 0 =<< sequence (Fold.toList bs)
+     LargeBitsVal n xs -> sel 0 . map fromVBit =<< sequence (enumerateSeqMap n xs)
 
  where
     sel offset []       = f offset
@@ -311,7 +306,7 @@ selectV mux val f =
 
 indexFront :: Maybe Integer
            -> TValue
-           -> SeqMap SBool SWord
+           -> SeqMap SBool SWord SInteger
            -> SWord
            -> Eval Value
 indexFront mblen a xs idx
@@ -323,22 +318,25 @@ indexFront mblen a xs idx
   = do wvs <- traverse (fromWordVal "indexFront" =<<) (enumerateSeqMap n xs)
        case asWordList wvs of
          Just ws ->
-           return $ VWord n $ ready $ WordVal $ SBV.svSelect ws (wordLit wlen 0) idx
-         Nothing -> foldr f def [0 .. 2^w  - 1]
+           return $ VWord wlen $ ready $ WordVal $ SBV.svSelect ws (wordLit wlen 0) idx
+         Nothing -> foldr f def idxs
 
   | otherwise
-  = foldr f def [0 .. 2^w  - 1]
+  = foldr f def idxs
 
  where
     k = SBV.kindOf idx
     w = SBV.intSizeOf idx
     def = ready $ zeroV a
     f n y = iteValue (SBV.svEqual idx (SBV.svInteger k n)) (lookupSeqMap xs n) y
+    idxs = case mblen of
+      Just n | n < 2^w -> [0 .. n-1]
+      _ -> [0 .. 2^w - 1]
 
 
 indexBack :: Maybe Integer
           -> TValue
-          -> SeqMap SBool SWord
+          -> SeqMap SBool SWord SInteger
           -> SWord
           -> Eval Value
 indexBack (Just n) a xs idx = indexFront (Just n) a (reverseSeqMap n xs) idx
@@ -346,7 +344,7 @@ indexBack Nothing _ _ _ = evalPanic "Expected finite sequence" ["indexBack"]
 
 indexFront_bits :: Maybe Integer
                 -> TValue
-                -> SeqMap SBool SWord
+                -> SeqMap SBool SWord SInteger
                 -> Seq.Seq SBool
                 -> Eval Value
 indexFront_bits mblen a xs bits0 = go 0 (length bits0) (Fold.toList bits0)
@@ -372,7 +370,7 @@ indexFront_bits mblen a xs bits0 = go 0 (length bits0) (Fold.toList bits0)
 
 indexBack_bits :: Maybe Integer
                -> TValue
-               -> SeqMap SBool SWord
+               -> SeqMap SBool SWord SInteger
                -> Seq.Seq SBool
                -> Eval Value
 indexBack_bits (Just n) a xs idx = indexFront_bits (Just n) a (reverseSeqMap n xs) idx
@@ -382,10 +380,10 @@ indexBack_bits Nothing _ _ _ = evalPanic "Expected finite sequence" ["indexBack_
 updateFrontSym
   :: Nat'
   -> TValue
-  -> SeqMap SBool SWord
-  -> WordValue SBool SWord
-  -> Eval (GenValue SBool SWord)
-  -> Eval (SeqMap SBool SWord)
+  -> SeqMap SBool SWord SInteger
+  -> WordValue SBool SWord SInteger
+  -> Eval (GenValue SBool SWord SInteger)
+  -> Eval (SeqMap SBool SWord SInteger)
 updateFrontSym len _eltTy vs w val = do
   -- TODO? alternate handling if w is a list of bits...
   case w of
@@ -399,33 +397,24 @@ updateFrontSym len _eltTy vs w val = do
           selectV iteValue w $ \j ->
             if i == j then val else lookupSeqMap vs i
 
-updateFrontSym_bits
+updateFrontSym_word
   :: Nat'
   -> TValue
-  -> Seq.Seq (Eval SBool)
-  -> WordValue SBool SWord
-  -> Eval (GenValue SBool SWord)
-  -> Eval (Seq.Seq (Eval SBool))
-updateFrontSym_bits Inf _ _ _ _ = evalPanic "Expected finite sequence" ["updateFrontSym_bits"]
-updateFrontSym_bits (Nat n) _eltTy bs w val = do
-  -- TODO? alternate handling if w is a list of bits...
-  case w of
-    WordVal wv | Just j <- SBV.svAsInteger wv -> do
-      unless (j < n) (invalidIndex j)
-      return $! Seq.update (fromInteger j) (fromVBit <$> val) bs
-    _ -> do
-      let mergeBit' c x y = mergeBit True c <$> x <*> y
-      return $ Seq.fromFunction (fromInteger n) $ \i ->
-        selectV mergeBit' w $ \j ->
-          if toInteger i == j then (fromVBit <$> val) else Seq.index bs i
+  -> WordValue SBool SWord SInteger
+  -> WordValue SBool SWord SInteger
+  -> Eval (GenValue SBool SWord SInteger)
+  -> Eval (WordValue SBool SWord SInteger)
+updateFrontSym_word Inf _ _ _ _ = evalPanic "Expected finite sequence" ["updateFrontSym_bits"]
+updateFrontSym_word (Nat _) _eltTy bs w val =
+  selectV (mergeWord' True) w $ \j -> updateWordValue bs j (fromVBit <$> val)
 
 updateBackSym
   :: Nat'
   -> TValue
-  -> SeqMap SBool SWord
-  -> WordValue SBool SWord
-  -> Eval (GenValue SBool SWord)
-  -> Eval (SeqMap SBool SWord)
+  -> SeqMap SBool SWord SInteger
+  -> WordValue SBool SWord SInteger
+  -> Eval (GenValue SBool SWord SInteger)
+  -> Eval (SeqMap SBool SWord SInteger)
 updateBackSym Inf _ _ _ _ = evalPanic "Expected finite sequence" ["updateBackSym"]
 updateBackSym (Nat n) _eltTy vs w val = do
   -- TODO? alternate handling if w is a list of bits...
@@ -438,26 +427,16 @@ updateBackSym (Nat n) _eltTy vs w val = do
           selectV iteValue w $ \j ->
             if i == (n - j - 1) then val else lookupSeqMap vs i
 
-updateBackSym_bits
+updateBackSym_word
   :: Nat'
   -> TValue
-  -> Seq.Seq (Eval SBool)
-  -> WordValue SBool SWord
-  -> Eval (GenValue SBool SWord)
-  -> Eval (Seq.Seq (Eval SBool))
-updateBackSym_bits Inf _ _ _ _ = evalPanic "Expected finite sequence" ["updateBackSym_bits"]
-updateBackSym_bits (Nat n) _eltTy bs w val = do
-  -- TODO? alternate handling if w is a list of bits...
-  case w of
-    WordVal wv | Just j <- SBV.svAsInteger wv -> do
-      unless (j < n) (invalidIndex j)
-      return $! Seq.update (fromInteger (n - j - 1)) (fromVBit <$> val) bs
-    _ -> do
-      let mergeBit' c x y = mergeBit True c <$> x <*> y
-      return $ Seq.fromFunction (fromInteger n) $ \i ->
-        selectV mergeBit' w $ \j ->
-          if toInteger i == (n - j - 1) then (fromVBit <$> val) else Seq.index bs i
-
+  -> WordValue SBool SWord SInteger
+  -> WordValue SBool SWord SInteger
+  -> Eval (GenValue SBool SWord SInteger)
+  -> Eval (WordValue SBool SWord SInteger)
+updateBackSym_word Inf _ _ _ _ = evalPanic "Expected finite sequence" ["updateBackSym_bits"]
+updateBackSym_word (Nat n) _eltTy bs w val = do
+  selectV (mergeWord' True) w $ \j -> updateWordValue bs (n - j - 1) (fromVBit <$> val)
 
 asBitList :: [Eval SBool] -> Maybe [SBool]
 asBitList = go id
@@ -467,20 +446,27 @@ asBitList = go id
        go _ _ = Nothing
 
 
-asWordList :: [WordValue SBool SWord] -> Maybe [SWord]
+asWordList :: [WordValue SBool SWord SInteger] -> Maybe [SWord]
 asWordList = go id
- where go :: ([SWord] -> [SWord]) -> [WordValue SBool SWord] -> Maybe [SWord]
+ where go :: ([SWord] -> [SWord]) -> [WordValue SBool SWord SInteger] -> Maybe [SWord]
        go f [] = Just (f [])
        go f (WordVal x :vs) = go (f . (x:)) vs
        go f (BitsVal bs:vs) =
               case asBitList (Fold.toList bs) of
                   Just xs -> go (f . (packWord xs:)) vs
                   Nothing -> Nothing
-
+       go _f (LargeBitsVal _ _ : _) = Nothing
 
 liftBinArith :: (SWord -> SWord -> SWord) -> BinArith SWord
 liftBinArith op _ x y = ready $ op x y
 
+liftBin :: (a -> b -> c) -> a -> b -> Eval c
+liftBin op x y = ready $ op x y
+
+liftModBin :: (SInteger -> SInteger -> a) -> Integer -> SInteger -> SInteger -> Eval a
+liftModBin op modulus x y = ready $ op (SBV.svRem x m) (SBV.svRem y m)
+  where m = integerLit modulus
+
 sExp :: Integer -> SWord -> SWord -> Eval SWord
 sExp _w x y = ready $ go (reverse (unpackWord y)) -- bits in little-endian order
   where go []       = literalSWord (SBV.intSizeOf x) 1
@@ -488,6 +474,10 @@ sExp _w x y = ready $ go (reverse (unpackWord y)) -- bits in little-endian order
             where a = go bs
                   s = SBV.svTimes a a
 
+sModExp :: Integer -> SInteger -> SInteger -> Eval SInteger
+sModExp modulus x y = ready $ SBV.svExp x (SBV.svRem y m)
+  where m = integerLit modulus
+
 -- | Ceiling (log_2 x)
 sLg2 :: Integer -> SWord -> Eval SWord
 sLg2 _w x = ready $ go 0
@@ -496,45 +486,30 @@ sLg2 _w x = ready $ go 0
     go i | i < SBV.intSizeOf x = SBV.svIte (SBV.svLessEq x (lit (2^i))) (lit (toInteger i)) (go (i + 1))
          | otherwise           = lit (toInteger i)
 
+-- | Ceiling (log_2 x)
+svLg2 :: SInteger -> Eval SInteger
+svLg2 x =
+  case SBV.svAsInteger x of
+    Just n -> ready $ SBV.svInteger SBV.KUnbounded (lg2 n)
+    Nothing -> evalPanic "cannot compute lg2 of symbolic unbounded integer" []
+
+svModLg2 :: Integer -> SInteger -> Eval SInteger
+svModLg2 modulus x = svLg2 (SBV.svRem x m)
+  where m = integerLit modulus
+
 -- Cmp -------------------------------------------------------------------------
 
-cmpValue :: (SBool -> SBool -> Eval a -> Eval a)
-         -> (SWord -> SWord -> Eval a -> Eval a)
-         -> (Value -> Value -> Eval a -> Eval a)
-cmpValue fb fw = cmp
-  where
-    cmp v1 v2 k =
-      case (v1, v2) of
-        (VRecord fs1, VRecord fs2) -> let vals = map snd . sortBy (comparing fst)
-                                      in  cmpValues (vals fs1) (vals fs2) k
-        (VTuple vs1 , VTuple vs2 ) -> cmpValues vs1 vs2 k
-        (VBit b1    , VBit b2    ) -> fb b1 b2 k
-        (VWord _ w1 , VWord _ w2 ) -> join (fw <$> (asWordVal =<< w1)
-                                               <*> (asWordVal =<< w2)
-                                               <*> return k)
-        (VSeq n vs1 , VSeq _ vs2 ) -> cmpValues (enumerateSeqMap n vs1)
-                                                (enumerateSeqMap n vs2) k
-        (VStream {} , VStream {} ) -> panic "Cryptol.Symbolic.Prims.cmpValue"
-                                        [ "Infinite streams are not comparable" ]
-        (VFun {}    , VFun {}    ) -> panic "Cryptol.Symbolic.Prims.cmpValue"
-                                        [ "Functions are not comparable" ]
-        (VPoly {}   , VPoly {}   ) -> panic "Cryptol.Symbolic.Prims.cmpValue"
-                                        [ "Polymorphic values are not comparable" ]
-        (_          , _          ) -> panic "Cryptol.Symbolic.Prims.cmpValue"
-                                        [ "type mismatch" ]
-
-    cmpValues (x1 : xs1) (x2 : xs2) k = do
-          x1' <- x1
-          x2' <- x2
-          cmp x1' x2' (cmpValues xs1 xs2 k)
-    cmpValues _ _ k = k
-
 cmpEq :: SWord -> SWord -> Eval SBool -> Eval SBool
 cmpEq x y k = SBV.svAnd (SBV.svEqual x y) <$> k
 
 cmpNotEq :: SWord -> SWord -> Eval SBool -> Eval SBool
 cmpNotEq x y k = SBV.svOr (SBV.svNotEqual x y) <$> k
 
+cmpSignedLt :: SWord -> SWord -> Eval SBool -> Eval SBool
+cmpSignedLt x y k = SBV.svOr (SBV.svLessThan sx sy) <$> (cmpEq sx sy k)
+  where sx = SBV.svSign x
+        sy = SBV.svSign y
+
 cmpLt, cmpGt :: SWord -> SWord -> Eval SBool -> Eval SBool
 cmpLt x y k = SBV.svOr (SBV.svLessThan x y) <$> (cmpEq x y k)
 cmpGt x y k = SBV.svOr (SBV.svGreaterThan x y) <$> (cmpEq x y k)
@@ -543,65 +518,62 @@ cmpLtEq, cmpGtEq :: SWord -> SWord -> Eval SBool -> Eval SBool
 cmpLtEq x y k = SBV.svAnd (SBV.svLessEq x y) <$> (cmpNotEq x y k)
 cmpGtEq x y k = SBV.svAnd (SBV.svGreaterEq x y) <$> (cmpNotEq x y k)
 
+cmpMod ::
+  (SInteger -> SInteger -> Eval SBool -> Eval SBool) ->
+  (Integer -> SInteger -> SInteger -> Eval SBool -> Eval SBool)
+cmpMod cmp modulus x y k = cmp (SBV.svRem x m) (SBV.svRem y m) k
+  where m = integerLit modulus
+
+cmpModEq :: Integer -> SInteger -> SInteger -> Eval SBool -> Eval SBool
+cmpModEq m x y k = SBV.svAnd (svDivisible m (SBV.svMinus x y)) <$> k
+
+cmpModNotEq :: Integer -> SInteger -> SInteger -> Eval SBool -> Eval SBool
+cmpModNotEq m x y k = SBV.svOr (SBV.svNot (svDivisible m (SBV.svMinus x y))) <$> k
+
+svDivisible :: Integer -> SInteger -> SBool
+svDivisible m x = SBV.svEqual (SBV.svRem x (integerLit m)) (integerLit 0)
+
 cmpBinary :: (SBool -> SBool -> Eval SBool -> Eval SBool)
           -> (SWord -> SWord -> Eval SBool -> Eval SBool)
-          -> SBool -> Binary SBool SWord
-cmpBinary fb fw b _ty v1 v2 = VBit <$> cmpValue fb fw v1 v2 (return b)
+          -> (SInteger -> SInteger -> Eval SBool -> Eval SBool)
+          -> (Integer -> SInteger -> SInteger -> Eval SBool -> Eval SBool)
+          -> SBool -> Binary SBool SWord SInteger
+cmpBinary fb fw fi fz b ty v1 v2 = VBit <$> cmpValue fb fw fi fz ty v1 v2 (return b)
 
--- Polynomials -----------------------------------------------------------------
+-- Signed arithmetic -----------------------------------------------------------
 
--- TODO: Data.SBV.BitVectors.Polynomials should export ites, addPoly,
--- and mdp (the following definitions are copied from that module)
+signedQuot :: SWord -> SWord -> SWord
+signedQuot x y = SBV.svUnsign (SBV.svQuot (SBV.svSign x) (SBV.svSign y))
 
--- | Add two polynomials
-addPoly :: [SBool] -> [SBool] -> [SBool]
-addPoly xs    []      = xs
-addPoly []    ys      = ys
-addPoly (x:xs) (y:ys) = SBV.svXOr x y : addPoly xs ys
+signedRem :: SWord -> SWord -> SWord
+signedRem x y = SBV.svUnsign (SBV.svRem (SBV.svSign x) (SBV.svSign y))
 
-ites :: SBool -> [SBool] -> [SBool] -> [SBool]
-ites s xs ys
- | Just t <- SBV.svAsBool s
- = if t then xs else ys
- | True
- = go xs ys
- where go [] []         = []
-       go []     (b:bs) = SBV.svIte s SBV.svFalse b : go [] bs
-       go (a:as) []     = SBV.svIte s a SBV.svFalse : go as []
-       go (a:as) (b:bs) = SBV.svIte s a b : go as bs
 
--- conservative over-approximation of the degree
-degree :: [SBool] -> Int
-degree xs = walk (length xs - 1) $ reverse xs
-  where walk n []     = n
-        walk n (b:bs)
-         | Just t <- SBV.svAsBool b
-         = if t then n else walk (n-1) bs
-         | True
-         = n -- over-estimate
+sshrV :: Value
+sshrV =
+  nlam $ \_n ->
+  nlam $ \_k ->
+  wlam $ \x -> return $
+  wlam $ \y ->
+   case SBV.svAsInteger y of
+     Just i ->
+       let z = SBV.svUnsign (SBV.svShr (SBV.svSign x) (fromInteger i))
+        in return . VWord (toInteger (SBV.intSizeOf x)) . ready . WordVal $ z
+     Nothing ->
+       let z = SBV.svUnsign (SBV.svShiftRight (SBV.svSign x) y)
+        in return . VWord (toInteger (SBV.intSizeOf x)) . ready . WordVal $ z
 
-mdp :: [SBool] -> [SBool] -> ([SBool], [SBool])
-mdp xs ys = go (length ys - 1) (reverse ys)
-  where degTop  = degree xs
-        go _ []     = error "SBV.Polynomial.mdp: Impossible happened; exhausted ys before hitting 0"
-        go n (b:bs)
-         | n == 0   = (reverse qs, rs)
-         | True     = let (rqs, rrs) = go (n-1) bs
-                      in (ites b (reverse qs) rqs, ites b rs rrs)
-         where degQuot = degTop - n
-               ys' = replicate degQuot SBV.svFalse ++ ys
-               (qs, rs) = divx (degQuot+1) degTop xs ys'
+carry :: SWord -> SWord -> Eval Value
+carry x y = return $ VBit c
+ where
+  c = SBV.svLessThan (SBV.svPlus x y) x
 
--- return the element at index i; if not enough elements, return false
--- N.B. equivalent to '(xs ++ repeat false) !! i', but more efficient
-nth :: [SBool] -> Int -> SBool
-nth []     _ = SBV.svFalse
-nth (x:_)  0 = x
-nth (_:xs) i = nth xs (i-1)
-
-divx :: Int -> Int -> [SBool] -> [SBool] -> ([SBool], [SBool])
-divx n _ xs _ | n <= 0 = ([], xs)
-divx n i xs ys'        = (q:qs, rs)
-  where q        = xs `nth` i
-        xs'      = ites q (xs `addPoly` ys') xs
-        (qs, rs) = divx (n-1) (i-1) xs' (tail ys')
+scarry :: SWord -> SWord -> Eval Value
+scarry x y = return $ VBit sc
+ where
+  n = SBV.intSizeOf x
+  z = SBV.svPlus (SBV.svSign x) (SBV.svSign y)
+  xsign = SBV.svTestBit x (n-1)
+  ysign = SBV.svTestBit y (n-1)
+  zsign = SBV.svTestBit z (n-1)
+  sc = SBV.svAnd (SBV.svEqual xsign ysign) (SBV.svNotEqual xsign zsign)
