@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Transform.MonoValues
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -78,9 +78,10 @@
 {-# LANGUAGE OverloadedStrings #-}
 module Cryptol.Transform.MonoValues (rewModule) where
 
-import Cryptol.ModuleSystem.Name (SupplyT,liftSupply,Supply,mkDeclared)
+import Cryptol.ModuleSystem.Name
+        (SupplyT,liftSupply,Supply,mkDeclared,NameSource(..))
 import Cryptol.Parser.Position (emptyRange)
-import Cryptol.TypeCheck.AST
+import Cryptol.TypeCheck.AST hiding (splitTApp) -- XXX: just use this one
 import Cryptol.TypeCheck.TypeMap
 import Cryptol.Utils.Ident (ModName)
 import Data.List(sortBy,groupBy)
@@ -145,7 +146,7 @@ type RO = ModName
 newName :: M Name
 newName  =
   do ns <- ask
-     liftSupply (mkDeclared ns "$mono" Nothing emptyRange)
+     liftSupply (mkDeclared ns SystemName "$mono" Nothing emptyRange)
 
 newTopOrLocalName :: M Name
 newTopOrLocalName  = newName
