@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Transform.Specialize
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -62,13 +62,13 @@ modify f = get >>= (set . f)
 -- type-specialized versions of all functions called (transitively) by
 -- the body of the expression.
 specialize :: Expr -> M.ModuleCmd Expr
-specialize expr modEnv = run $ do
+specialize expr (ev,modEnv) = run $ do
   let extDgs = allDeclGroups modEnv
   let (tparams, expr') = destETAbs expr
   spec' <- specializeEWhere expr' extDgs
   return (foldr ETAbs spec' tparams)
   where
-  run = M.runModuleT modEnv . fmap fst . runSpecT Map.empty
+  run = M.runModuleT (ev,modEnv) . fmap fst . runSpecT Map.empty
 
 specializeExpr :: Expr -> SpecM Expr
 specializeExpr expr =
@@ -163,8 +163,8 @@ specializeEWhere e dgs = do
 specializeDeclGroups :: [DeclGroup] -> SpecM ([DeclGroup], Map Name (TypesMap Name))
 specializeDeclGroups dgs = do
   let decls = concatMap groupDecls dgs
-  let isMono s = null (sVars s) && null (sProps s)
-  let monos = [ EVar (dName d) | d <- decls, isMono (dSignature d) ]
+  let isMonoType s = null (sVars s) && null (sProps s)
+  let monos = [ EVar (dName d) | d <- decls, isMonoType (dSignature d) ]
   (_, dgs', names) <- withDeclGroups dgs $ mapM specializeExpr monos
   return (dgs', names)
 
@@ -243,8 +243,8 @@ destETAbs = go []
 freshName :: Name -> [Type] -> SpecM Name
 freshName n _ =
   case nameInfo n of
-    Declared ns -> liftSupply (mkDeclared ns ident fx loc)
-    Parameter   -> liftSupply (mkParameter ident loc)
+    Declared ns s -> liftSupply (mkDeclared ns s ident fx loc)
+    Parameter     -> liftSupply (mkParameter ident loc)
   where
   fx    = nameFixity n
   ident = nameIdent n
