@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.TypeCheck.Infer
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -13,21 +13,24 @@
 {-# LANGUAGE ViewPatterns #-}
 {-# LANGUAGE RecursiveDo #-}
 {-# LANGUAGE Safe #-}
-module Cryptol.TypeCheck.Infer where
+module Cryptol.TypeCheck.Infer(checkSigB,inferModule,inferBinds,inferDs) where
 
 import           Cryptol.ModuleSystem.Name (asPrim,lookupPrimDecl)
 import           Cryptol.Parser.Position
 import qualified Cryptol.Parser.AST as P
-import qualified Cryptol.Parser.Names as P
+import qualified Cryptol.ModuleSystem.Exports as P
 import           Cryptol.TypeCheck.AST hiding (tSub,tMul,tExp)
 import           Cryptol.TypeCheck.Monad
+import           Cryptol.TypeCheck.Error
 import           Cryptol.TypeCheck.Solve
-import           Cryptol.TypeCheck.SimpType(tSub,tMul,tExp)
+import           Cryptol.TypeCheck.SimpType(tSub,tMul,tExp,tAdd)
 import           Cryptol.TypeCheck.Kind(checkType,checkSchema,checkTySyn,
-                                          checkNewtype)
+                                        checkPropSyn,checkNewtype,
+                                        checkParameterType,
+                                        checkParameterConstraints)
 import           Cryptol.TypeCheck.Instantiate
 import           Cryptol.TypeCheck.Depends
-import           Cryptol.TypeCheck.Subst (listSubst,apSubst,(@@),emptySubst)
+import           Cryptol.TypeCheck.Subst (listSubst,apSubst,(@@),isEmptySubst)
 import           Cryptol.TypeCheck.Solver.InfNat(genLog)
 import           Cryptol.Utils.Ident
 import           Cryptol.Utils.Panic(panic)
@@ -36,61 +39,80 @@ import           Cryptol.Utils.PP
 import qualified Data.Map as Map
 import           Data.Map (Map)
 import qualified Data.Set as Set
+import           Data.List(foldl',sortBy)
 import           Data.Either(partitionEithers)
 import           Data.Maybe(mapMaybe,isJust, fromMaybe)
 import           Data.List(partition,find)
 import           Data.Graph(SCC(..))
 import           Data.Traversable(forM)
-import           Control.Monad(when,zipWithM)
+import           Control.Monad(zipWithM,unless)
+
+
 
 inferModule :: P.Module Name -> InferM Module
 inferModule m =
   inferDs (P.mDecls m) $ \ds1 ->
-    do simplifyAllConstraints
+    do proveModuleTopLevel
        ts <- getTSyns
        nts <- getNewtypes
-       return Module { mName    = thing (P.mName m)
-                     , mExports = P.modExports m
-                     , mImports = map thing (P.mImports m)
-                     , mTySyns  = Map.mapMaybe onlyLocal ts
-                     , mNewtypes = Map.mapMaybe onlyLocal nts
-                     , mDecls   = ds1
+       pTs <- getParamTypes
+       pCs <- getParamConstraints
+       pFuns <- getParamFuns
+       return Module { mName      = thing (P.mName m)
+                     , mExports   = P.modExports m
+                     , mImports   = map thing (P.mImports m)
+                     , mTySyns    = Map.mapMaybe onlyLocal ts
+                     , mNewtypes  = Map.mapMaybe onlyLocal nts
+                     , mParamTypes = pTs
+                     , mParamConstraints = pCs
+                     , mParamFuns = pFuns
+                     , mDecls     = ds1
                      }
   where
   onlyLocal (IsLocal, x)    = Just x
   onlyLocal (IsExternal, _) = Nothing
 
 
+
 -- | Construct a primitive in the parsed AST.
 mkPrim :: String -> InferM (P.Expr Name)
 mkPrim str =
+  do nm <- mkPrim' str
+     return (P.EVar nm)
+
+-- | Construct a primitive in the parsed AST.
+mkPrim' :: String -> InferM Name
+mkPrim' str =
   do prims <- getPrimMap
-     return (P.EVar (lookupPrimDecl (packIdent str) prims))
+     return (lookupPrimDecl (packIdent str) prims)
+
+
 
 desugarLiteral :: Bool -> P.Literal -> InferM (P.Expr Name)
 desugarLiteral fixDec lit =
   do l <- curRange
-     demotePrim <- mkPrim "demote"
+     numberPrim <- mkPrim "number"
      let named (x,y)  = P.NamedInst
-                        P.Named { name = Located l (packIdent x), value = P.TNum y }
-         demote fs    = P.EAppT demotePrim (map named fs)
+                        P.Named { name = Located l (packIdent x), value = y }
+         number fs    = P.EAppT numberPrim (map named fs)
+         tBits n = P.TSeq (P.TNum n) P.TBit
 
      return $ case lit of
 
        P.ECNum num info ->
-         demote $ [ ("val", num) ] ++ case info of
-           P.BinLit n    -> [ ("bits", 1 * toInteger n) ]
-           P.OctLit n    -> [ ("bits", 3 * toInteger n) ]
-           P.HexLit n    -> [ ("bits", 4 * toInteger n) ]
-           P.CharLit     -> [ ("bits", 8 :: Integer) ]
+         number $ [ ("val", P.TNum num) ] ++ case info of
+           P.BinLit n    -> [ ("rep", tBits (1 * toInteger n)) ]
+           P.OctLit n    -> [ ("rep", tBits (3 * toInteger n)) ]
+           P.HexLit n    -> [ ("rep", tBits (4 * toInteger n)) ]
+           P.CharLit     -> [ ("rep", tBits (8 :: Integer)) ]
            P.DecLit
             | fixDec     -> if num == 0
-                              then [ ("bits", 0)]
+                              then [ ("rep", tBits 0)]
                               else case genLog num 2 of
-                                     Just (x,_) -> [ ("bits", x + 1) ]
+                                     Just (x,_) -> [ ("rep", tBits (x + 1)) ]
                                      _          -> []
             | otherwise  -> [ ]
-           P.PolyLit _n  -> [ ]
+           P.PolyLit _n  -> [ ("rep", P.TSeq P.TWild P.TBit) ]
 
        P.ECString s ->
           P.ETyped (P.EList [ P.ELit (P.ECNum (fromIntegral (fromEnum c))
@@ -105,8 +127,9 @@ appTys expr ts tGoal =
     P.EVar x ->
       do res <- lookupVar x
          (e',t) <- case res of
-           ExtVar s   -> instantiateWith (EVar x) s ts
-           CurSCC e t -> instantiateWith e (Forall [] [] t) ts
+           ExtVar s   -> instantiateWith x (EVar x) s ts
+           CurSCC e t -> do checkNoParams ts
+                            return (e,t)
 
          checkHasType t tGoal
          return e'
@@ -114,6 +137,11 @@ appTys expr ts tGoal =
     P.ELit l -> do e <- desugarLiteral False l
                    appTys e ts tGoal
 
+
+    P.ENeg {} -> panic "appTys" ["[bug] renamer bug", "unexpected ENeg" ]
+    P.EComplement {} ->
+      panic "appTys" ["[bug] renamer bug", "unexpected EComplement" ]
+
     P.EAppT e fs ->
       do ps <- mapM inferTyParam fs
          appTys e (ps ++ ts) tGoal
@@ -144,12 +172,22 @@ appTys expr ts tGoal =
     P.EParens e       -> appTys e ts tGoal
     P.EInfix a op _ b -> appTys (P.EVar (thing op) `P.EApp` a `P.EApp` b) ts tGoal
 
-  where mono = do e'     <- checkE expr tGoal
-                  (ie,t) <- instantiateWith e' (Forall [] [] tGoal) ts
-                  -- XXX seems weird to need to do this, as t should be the same
-                  -- as tGoal
-                  checkHasType t tGoal
-                  return ie
+  where mono = do e' <- checkE expr tGoal
+                  checkNoParams ts
+                  return e'
+
+checkNoParams :: [Located (Maybe Ident,Type)] -> InferM ()
+checkNoParams ts =
+  case pos of
+    p : _ -> inRange (srcRange p) (recordError TooManyPositionalTypeParams)
+    []    -> mapM_ badNamed named
+  where
+  badNamed l =
+    case fst (thing l) of
+      Just i -> recordError (UndefinedTypeParameter l { thing = i })
+      Nothing -> return ()
+
+  (named,pos) = partition (isJust . fst . thing) ts
 
 
 inferTyParam :: P.TypeInst Name -> InferM (Located (Maybe Ident, Type))
@@ -169,14 +207,6 @@ checkTypeOfKind :: P.Type Name -> Kind -> InferM Type
 checkTypeOfKind ty k = checkType ty (Just k)
 
 
--- | We use this when we want to ensure that the expr has exactly
--- (syntactically) the given type.
-inferE :: Doc -> P.Expr Name -> InferM (Expr, Type)
-inferE desc expr =
-  do t  <- newType desc KType
-     e1 <- checkE expr t
-     return (e1,t)
-
 -- | Infer the type of an expression, and translate it to a fully elaborated
 -- core term.
 checkE :: P.Expr Name -> Type -> InferM Expr
@@ -185,12 +215,16 @@ checkE expr tGoal =
     P.EVar x ->
       do res <- lookupVar x
          (e',t) <- case res of
-           ExtVar s   -> instantiateWith (EVar x) s []
-           CurSCC e t -> return (e, t)
+                     ExtVar s   -> instantiateWith x (EVar x) s []
+                     CurSCC e t -> return (e, t)
 
          checkHasType t tGoal
          return e'
 
+    P.ENeg {} -> panic "checkE" ["[bug] renamer bug", "unexpected ENeg" ]
+    P.EComplement {} ->
+      panic "checkE" ["[bug] renamer bug", "unexpected EComplement" ]
+
     P.ELit l -> (`checkE` tGoal) =<< desugarLiteral False l
 
     P.ETuple es ->
@@ -205,10 +239,11 @@ checkE expr tGoal =
 
     P.ESel e l ->
       do let src = case l of
-                     RecordSel _ _ -> text "type of record"
-                     TupleSel _ _  -> text "type of tuple"
-                     ListSel _ _   -> text "type of sequence"
-         (e',t) <- inferE src e
+                     RecordSel la _ -> TypeOfRecordField la
+                     TupleSel n _   -> TypeOfTupleField n
+                     ListSel _ _    -> TypeOfSeqElement
+         t <- newType src KType
+         e' <- checkE e t
          f <- newHasGoal l t tGoal
          return (f e')
 
@@ -225,15 +260,19 @@ checkE expr tGoal =
 
     P.EFromTo t1 Nothing Nothing ->
       do rng <- curRange
-         bit <- newType (text "bit-width of enumeration sequnce") KNum
+         fromThenPrim <- mkPrim' "fromThen"
+         let src = TypeParamInstNamed fromThenPrim (packIdent "bits")
+         bit <- newType src KNum
          fstT <- checkTypeOfKind t1 KNum
-         let totLen = tExp (tNum (2::Int)) bit
-             lstT   = tSub totLen (tNum (1::Int))
+         let nextT = tAdd fstT (tNum (1::Int))
+             lenT  = tSub (tExp (tNum (2::Int)) bit) fstT
 
-         fromToPrim <- mkPrim "fromTo"
-         appTys fromToPrim
+         appTys (P.EVar fromThenPrim)
            [ Located rng (Just (packIdent x), y)
-           | (x,y) <- [ ("first",fstT), ("last", lstT), ("bits", bit) ]
+           | (x,y) <- [ ("first", fstT)
+                      , ("next", nextT)
+                      , ("len",   lenT)
+                      , ("bits",  bit) ]
            ] tGoal
 
     P.EFromTo t1 mbt2 mbt3 ->
@@ -293,7 +332,7 @@ checkE expr tGoal =
            checkE (P.EApp fun newArg) tGoal
 
     P.EApp e1 e2 ->
-      do t1  <- newType (text "argument to function") KType
+      do t1  <- newType (TypeOfArg Nothing) KType
          e1' <- checkE e1 (tFun t1 tGoal)
          e2' <- checkE e2 t1
          return (EApp e1' e2')
@@ -316,10 +355,11 @@ checkE expr tGoal =
 
     P.ETypeVal t ->
       do l <- curRange
-         prim <- mkPrim "demote"
+         prim <- mkPrim "number"
          checkE (P.EAppT prim
                   [P.NamedInst
-                   P.Named { name = Located l (packIdent "val"), value = t }]) tGoal
+                   P.Named { name = Located l (packIdent "val")
+                           , value = t }]) tGoal
 
     P.EFun ps e -> checkFun (text "anonymous function") ps e tGoal
 
@@ -351,8 +391,8 @@ expectSeq ty =
          return tys
   where
   genTys =
-    do a <- newType (text "size of the sequence") KNum
-       b <- newType (text "type of sequence elements") KType
+    do a <- newType LenOfSeq KNum
+       b <- newType TypeOfSeqElement KType
        return (a,b)
 
 
@@ -377,11 +417,7 @@ expectTuple n ty =
          return tys
 
   where
-  genTys =forM [ 0 .. n - 1 ] $ \ i ->
-              let desc = text "type of"
-                     <+> ordinal i
-                     <+> text "tuple field"
-               in newType desc KType
+  genTys =forM [ 0 .. n - 1 ] $ \ i -> newType (TypeOfTupleField i) KType
 
 expectRec :: [P.Named a] -> Type -> InferM [(Ident,a,Type)]
 expectRec fs ty =
@@ -409,7 +445,7 @@ expectRec fs ty =
   genTys =
     do res <- forM fs $ \ f ->
              do let field = thing (P.name f)
-                t <- newType (text "type of field" <+> quotes (pp field)) KType
+                t <- newType (TypeOfRecordField field) KType
                 return (field, P.value f, t)
 
        let (ls,_,ts) = unzip3 res
@@ -445,7 +481,7 @@ expectFun  = go []
 
         _ ->
           do args <- genArgs arity
-             res  <- newType (text "result of function") KType
+             res  <- newType TypeOfRes KType
              case ty of
                TVar TVFree{} -> do ps <- unify ty (foldr tFun res args)
                                    newGoals CtExactType  ps
@@ -455,8 +491,8 @@ expectFun  = go []
     | otherwise =
       return (reverse tys, ty)
 
-  genArgs arity = forM [ 1 .. arity ] $ \ ix ->
-                      newType (text "argument" <+> ordinal ix) KType
+  genArgs arity = forM [ 1 .. arity ] $
+                    \ ix -> newType (TypeOfArg (Just ix)) KType
 
 
 checkHasType :: Type -> Type -> InferM ()
@@ -485,9 +521,9 @@ checkFun desc ps e tGoal =
 
 {-| The type the is the smallest of all -}
 smallest :: [Type] -> InferM Type
-smallest []   = newType (text "length of list comprehension") KNum
+smallest []   = newType LenOfSeq KNum
 smallest [t]  = return t
-smallest ts   = do a <- newType (text "length of list comprehension") KNum
+smallest ts   = do a <- newType LenOfSeq KNum
                    newGoals CtComprehension [ a =#= foldr1 tMin ts ]
                    return a
 
@@ -508,7 +544,7 @@ inferP desc pat =
   case pat of
 
     P.PVar x0 ->
-      do a   <- newType desc KType
+      do a   <- inRange (srcRange x0) (newType (DefinitionOf (thing x0)) KType)
          return (thing x0, x0 { thing = a })
 
     P.PTyped p t ->
@@ -524,13 +560,14 @@ inferP desc pat =
 inferMatch :: P.Match Name -> InferM (Match, Name, Located Type, Type)
 inferMatch (P.Match p e) =
   do (x,t) <- inferP (text "a value bound by a generator in a comprehension") p
-     n     <- newType (text "the length of a generator in a comprehension") KNum
+     n     <- newType LenOfCompGen KNum
      e'    <- checkE e (tSeq n (thing t))
      return (From x n (thing t) e', x, t, n)
 
 inferMatch (P.MatchLet b)
   | P.bMono b =
-  do a <- newType (text "`let` binding in comprehension") KType
+  do let rng = srcRange (P.bName b)
+     a <- inRange rng (newType (DefinitionOf (thing (P.bName b))) KType)
      b1 <- checkMonoB b a
      return (Let b1, dName b1, Located (srcRange (P.bName b)) a, tNum (1::Int))
 
@@ -544,22 +581,16 @@ inferCArm :: Int -> [P.Match Name] -> InferM
               , Type                   -- length of sequence
               )
 
-inferCArm _ [] = do n <- newType (text "lenght of empty comprehension") KNum
-                                                    -- shouldn't really happen
-                    return ([], Map.empty, n)
+inferCArm _ [] = panic "inferCArm" [ "Empty comprahension arm" ]
 inferCArm _ [m] =
   do (m1, x, t, n) <- inferMatch m
      return ([m1], Map.singleton x t, n)
 
 inferCArm armNum (m : ms) =
   do (m1, x, t, n)  <- inferMatch m
-     (ms', ds, n') <- withMonoType (x,t) (inferCArm armNum ms)
-     -- XXX: Well, this is just the lenght of this sub-sequence
-     let src = text "length of" <+> ordinal armNum <+>
-                                  text "arm of list comprehension"
-     sz <- newType src KNum
-     newGoals CtComprehension [ sz =#= tMul n n' ]
-     return (m1 : ms', Map.insertWith (\_ old -> old) x t ds, sz)
+     (ms', ds, n')  <- withMonoType (x,t) (inferCArm armNum ms)
+     newGoals CtComprehension [ pFin n' ]
+     return (m1 : ms', Map.insertWith (\_ old -> old) x t ds, tMul n n')
 
 -- | @inferBinds isTopLevel isRec binds@ performs inference for a
 -- strongly-connected component of 'P.Bind's. If @isTopLevel@ is true,
@@ -569,43 +600,80 @@ inferCArm armNum (m : ms) =
 -- be unaffected.
 inferBinds :: Bool -> Bool -> [P.Bind Name] -> InferM [Decl]
 inferBinds isTopLevel isRec binds =
-  mdo let dExpr (DExpr e) = e
-          dExpr DPrim     = panic "[TypeCheck]" [ "primitive in a recursive group" ]
-
-          exprMap = Map.fromList [ (x,inst (EVar x) (dExpr (dDefinition b)))
-                                 | b <- genBs, let x = dName b ] -- REC.
-
-          inst e (ETAbs x e1)     = inst (ETApp e (TVar (tpVar x))) e1
-          inst e (EProofAbs _ e1) = inst (EProofApp e) e1
-          inst e _                = e
-
-      -- when mono-binds is enabled, and we're not checking top-level
-      -- declarations, mark all bindings lacking signatures as monomorphic
-      monoBinds <- getMonoBinds
-      let binds' | monoBinds && not isTopLevel = sigs ++ monos
-                 | otherwise                   = binds
-
-          (sigs,noSigs) = partition (isJust . P.bSignature) binds
-          monos         = [ b { P.bMono = True } | b <- noSigs ]
+  do -- when mono-binds is enabled, and we're not checking top-level
+     -- declarations, mark all bindings lacking signatures as monomorphic
+     monoBinds <- getMonoBinds
+     let (sigs,noSigs) = partition (isJust . P.bSignature) binds
+         monos         = [ b { P.bMono = True } | b <- noSigs ]
+         binds' | monoBinds && not isTopLevel = sigs ++ monos
+                | otherwise                   = binds
 
 
-      ((doneBs, genCandidates), cs) <-
-        collectGoals $
 
+         check exprMap =
         {- Guess type is here, because while we check user supplied signatures
            we may generate additional constraints. For example, `x - y` would
            generate an additional constraint `x >= y`. -}
-        do (newEnv,todos) <- unzip `fmap` mapM (guessType exprMap) binds'
-           let extEnv = if isRec then withVarTypes newEnv else id
+           do (newEnv,todos) <- unzip `fmap` mapM (guessType exprMap) binds'
+              let otherEnv = filter isExt newEnv
+
+              let (sigsAndMonos,noSigGen) = partitionEithers todos
+
+              let prepGen = collectGoals
+                          $ do bs <- sequence noSigGen
+                               simplifyAllConstraints
+                               return bs
+
+              if isRec
+                then
+                  -- First we check the bindings with no signatures
+                  -- that need to be generalized.
+                  do (bs1,cs) <- withVarTypes newEnv prepGen
+
+                     -- We add these to the environment, so their fvs are
+                     -- not generalized.
+                     genCs <- withVarTypes otherEnv (generalize bs1 cs)
+
+                     -- Then we do all the rest,
+                     -- using the newly inferred poly types.
+                     let newEnv' = map toExt bs1 ++ otherEnv
+                     done <- withVarTypes newEnv' (sequence sigsAndMonos)
+                     return (done,genCs)
+
+                else
+                  do done      <- sequence sigsAndMonos
+                     (bs1, cs) <- prepGen
+                     genCs     <- generalize bs1 cs
+                     return (done,genCs)
+
+     rec
+       let exprMap = Map.fromList (map monoUse genBs)
+       (doneBs, genBs) <- check exprMap
+
+     simplifyAllConstraints
+
+     return (doneBs ++ genBs)
+
+  where
+  toExt d = (dName d, ExtVar (dSignature d))
+  isExt (_,y) = case y of
+                  ExtVar _ -> True
+                  _        -> False
+
+  monoUse d = (x, withQs)
+    where
+    x  = dName d
+    as = sVars (dSignature d)
+    qs = sProps (dSignature d)
+
+    appT e a = ETApp e (TVar (tpVar a))
+    appP e _ = EProofApp e
+
+    withTys  = foldl' appT (EVar x) as
+    withQs   = foldl' appP withTys  qs
+
+
 
-           extEnv $
-             do let (sigsAndMonos,noSigGen) = partitionEithers todos
-                genCs <- sequence noSigGen
-                done  <- sequence sigsAndMonos
-                simplifyAllConstraints
-                return (done, genCs)
-      genBs <- generalize genCandidates cs -- RECURSION
-      return (doneBs ++ genBs)
 
 
 {- | Come up with a type for recursive calls to a function, and decide
@@ -624,18 +692,18 @@ guessType exprMap b@(P.Bind { .. }) =
   case bSignature of
 
     Just s ->
-      do s1 <- checkSchema s
+      do s1 <- checkSchema AllowWildCards s
          return ((name, ExtVar (fst s1)), Left (checkSigB b s1))
 
     Nothing
       | bMono ->
-         do t <- newType (text "definition of" <+> quotes (pp name)) KType
+         do t <- newType (DefinitionOf name) KType
             let schema = Forall [] [] t
             return ((name, ExtVar schema), Left (checkMonoB b t))
 
       | otherwise ->
 
-        do t <- newType (text "definition of" <+> quotes (pp name)) KType
+        do t <- newType (DefinitionOf name) KType
            let noWay = tcPanic "guessType" [ "Missing expression for:" ,
                                                                 show name ]
                expr  = Map.findWithDefault noWay name exprMap
@@ -646,8 +714,8 @@ guessType exprMap b@(P.Bind { .. }) =
 
 
 
--- | The inputs should be declarations with monomorphic types
--- (i.e., of the form `Forall [] [] t`).
+{- | The inputs should be declarations with monomorphic types
+(i.e., of the form `Forall [] [] t`). -}
 generalize :: [Decl] -> [Goal] -> InferM [Decl]
 
 {- This may happen because we have monomorphic bindings.
@@ -659,52 +727,56 @@ generalize [] gs0 =
 
 
 generalize bs0 gs0 =
-  do gs <- forM gs0 $ \g -> applySubst g
-
-     -- XXX: Why would these bindings have signatures??
-     bs  <- forM bs0 $ \b -> do s <- applySubst (dSignature b)
-                                return b { dSignature = s }
+  do {- First, we apply the accumulating substitution to the goals
+        and the inferred types, to ensure that we have the most up
+        to date information. -}
+     gs <- applySubstGoals gs0
+     bs <- forM bs0 $ \b -> do s <- applySubst (dSignature b)
+                               return b { dSignature = s }
 
+     -- Next, we figure out which of the free variables need to be generalized
+     -- Variables apearing in the types of monomorphic bindings should
+     -- not be generalizedr.
      let goalFVS g  = Set.filter isFreeTV $ fvs $ goal g
          inGoals    = Set.unions $ map goalFVS gs
          inSigs     = Set.filter isFreeTV $ fvs $ map dSignature bs
-         candidates = Set.union inGoals inSigs
+         candidates = (Set.union inGoals inSigs)
 
      asmpVs <- varsWithAsmps
 
-
-
      let gen0          = Set.difference candidates asmpVs
          stays g       = any (`Set.member` gen0) $ Set.toList $ goalFVS g
          (here0,later) = partition stays gs
+     addGoals later   -- these ones we keep around for to solve later
 
-     -- Figure out what might be ambigious
-     let (maybeAmbig, ambig) = partition ((KNum ==) . kindOf)
-                             $ Set.toList
-                             $ Set.difference gen0 inSigs
-
-     when (not (null ambig)) $ recordError $ AmbiguousType $ map dName bs
-
-
-     solver <- getSolver
-     (as0,here1,mb_defSu,ws) <- io $ improveByDefaultingWith solver maybeAmbig here0
-     defSu <- case mb_defSu of
-               Nothing -> do recordError $ UnsolvedGoals True here0
-                             return emptySubst
-               Just s  -> return s
-
+     let maybeAmbig = Set.toList (Set.difference gen0 inSigs)
 
+     {- See if we might be able to default some of the potentially ambiguous
+        variables using the constraints that will be part of the newly
+        generalized schema.  -}
+     let (as0,here1,defSu,ws) = defaultAndSimplify maybeAmbig here0
 
+     extendSubst defSu
      mapM_ recordWarning ws
      let here = map goal here1
 
-     let as   = as0 ++ Set.toList (Set.difference inSigs asmpVs)
-         asPs = [ TParam { tpUnique = x, tpKind = k, tpName = Nothing }
-                                                   | TVFree x k _ _ <- as ]
+
+     {- This is the variables we'll be generalizing:
+          * any ones that survived the defaulting
+          * and vars in the inferred types that do not appear anywhere else. -}
+     let as   = sortBy numFst
+              $ as0 ++ Set.toList (Set.difference inSigs asmpVs)
+         asPs = [ TParam { tpUnique = x, tpKind = k, tpFlav = TPOther Nothing
+                         , tpInfo = i  } | TVFree x k _ i <- as ]
+
+     {- Finally, we replace free variables with bound ones, and fix-up
+        the definitions as needed to reflect that we are now working
+        with polymorphic things. For example, apply each occurrence to the
+        type parameters. -}
      totSu <- getSubst
      let
-         su     = listSubst (zip as (map (TVar . tpVar) asPs)) @@ defSu @@ totSu
-         qs     = map (apSubst su) here
+         su     = listSubst (zip as (map (TVar . tpVar) asPs)) @@ totSu
+         qs     = concatMap (pSplitAnd . apSubst su) here
 
          genE e = foldr ETAbs (foldr EProofAbs (apSubst su e) qs) asPs
          genB d = d { dDefinition = case dDefinition d of
@@ -714,25 +786,22 @@ generalize bs0 gs0 =
                                   $ apSubst su $ sType $ dSignature d
                     }
 
-     addGoals later
      return (map genB bs)
 
+  where
+  numFst x y = case (kindOf x, kindOf y) of
+                 (KNum, KNum) -> EQ
+                 (KNum, _)    -> LT
+                 (_,KNum)     -> GT
+                 _            -> EQ
 
-
+-- | Check a monomorphic binding.
 checkMonoB :: P.Bind Name -> Type -> InferM Decl
 checkMonoB b t =
   inRangeMb (getLoc b) $
   case thing (P.bDef b) of
 
-    P.DPrim ->
-         return Decl { dName = thing (P.bName b)
-                     , dSignature = Forall [] [] t
-                     , dDefinition = DPrim
-                     , dPragmas = P.bPragmas b
-                     , dInfix = P.bInfix b
-                     , dFixity = P.bFixity b
-                     , dDoc = P.bDoc b
-                     }
+    P.DPrim -> panic "checkMonoB" ["Primitive with no signature?"]
 
     P.DExpr e ->
       do e1 <- checkFun (pp (thing (P.bName b))) (P.bParams b) e t
@@ -769,51 +838,29 @@ checkSigB b (Forall as asmps0 t0, validSchema) = case thing (P.bDef b) of
                    addGoals validSchema
                    () <- simplifyAllConstraints  -- XXX: using `asmps` also?
                    return e1
-     cs <- applySubst cs0
+     cs <- applySubstGoals cs0
 
-     let letGo qs c  = Set.null (qs `Set.intersection` fvs (goal c))
+     let findKeep vs keep todo =
+          let stays (_,cvs)    = not $ Set.null $ Set.intersection vs cvs
+              (yes,perhaps)    = partition stays todo
+              (stayPs,newVars) = unzip yes
+          in case stayPs of
+               [] -> (keep,map fst todo)
+               _  -> findKeep (Set.unions (vs:newVars)) (stayPs ++ keep) perhaps
 
-         splitPreds qs n ps =
-           let (l,n1) = partition (letGo qs) ps
-           in if null n1
-                then (l,n)
-                else splitPreds (fvs (map goal n1) `Set.union` qs) (n1 ++ n) l
+     let (stay,leave) = findKeep (Set.fromList (map tpVar as)) []
+                            [ (c, fvs c) | c <- cs ]
 
-         (later0,now) = splitPreds (Set.fromList (map tpVar as)) [] cs
+     addGoals leave
 
-     asmps1 <- applySubst asmps0
+     asmps1 <- applySubstPreds asmps0
 
-     defSu1 <- proveImplication (thing (P.bName b)) as asmps1 now
-     let later = apSubst defSu1 later0
-         asmps = apSubst defSu1 asmps1
+     su <- proveImplication (Just (thing (P.bName b))) as asmps1 stay
+     extendSubst su
 
-     -- Now we check for any remaining variables that are not mentioned
-     -- in the environment.  The plan is to try to default these to something
-     -- reasonable.
-     do let laterVs = fvs (map goal later)
-        asmpVs <- varsWithAsmps
-        let genVs   = laterVs `Set.difference` asmpVs
-            (maybeAmbig,ambig) = partition ((== KNum) . kindOf)
-                                           (Set.toList genVs)
-        when (not (null ambig)) $ recordError
-                                $ AmbiguousType [ thing (P.bName b) ]
-
-        solver <- getSolver
-        (_,_,mb_defSu2,ws) <-
-            io $ improveByDefaultingWith solver maybeAmbig later
-        defSu2 <- case mb_defSu2 of
-                   Nothing -> do recordError $ UnsolvedGoals True later
-                                 return emptySubst
-                   Just s -> return s
-        mapM_ recordWarning ws
-        extendSubst defSu2
-
-     addGoals later
-
-     su <- getSubst
-     let su' = defSu1 @@ su
-         t   = apSubst su' t0
-         e2  = apSubst su' e1
+     let asmps  = concatMap pSplitAnd (apSubst su asmps1)
+     t      <- applySubst t0
+     e2     <- applySubst e1
 
      return Decl
         { dName       = thing (P.bName b)
@@ -830,18 +877,43 @@ inferDs ds continue = checkTyDecls =<< orderTyDecls (mapMaybe toTyDecl ds)
   where
   isTopLevel = isTopDecl (head ds)
 
-  checkTyDecls (TS t : ts) =
-    do t1 <- checkTySyn t
+  checkTyDecls (AT t mbD : ts) =
+    do t1 <- checkParameterType t mbD
+       withParamType t1 (checkTyDecls ts)
+
+  checkTyDecls (TS t mbD : ts) =
+    do t1 <- checkTySyn t mbD
        withTySyn t1 (checkTyDecls ts)
 
-  checkTyDecls (NT t : ts) =
-    do t1 <- checkNewtype t
+  checkTyDecls (PS t mbD : ts) =
+    do t1 <- checkPropSyn t mbD
+       withTySyn t1 (checkTyDecls ts)
+
+  checkTyDecls (NT t mbD : ts) =
+    do t1 <- checkNewtype t mbD
        withNewtype t1 (checkTyDecls ts)
 
   -- We checked all type synonyms, now continue with value-level definitions:
-  checkTyDecls [] = checkBinds [] $ orderBinds $ mapMaybe toBind ds
+  checkTyDecls [] =
+    do cs <- checkParameterConstraints (concatMap toParamConstraints ds)
+       withParameterConstraints cs $
+         do xs <- mapM checkParameterFun (mapMaybe toParamFun ds)
+            withParamFuns xs $ checkBinds [] $ orderBinds $ mapMaybe toBind ds
 
 
+  checkParameterFun x =
+    do (s,gs) <- checkSchema NoWildCards (P.pfSchema x)
+       su <- proveImplication (Just (thing (P.pfName x)))
+                              (sVars s) (sProps s) gs
+       unless (isEmptySubst su) $
+         panic "checkParameterFun" ["Subst not empty??"]
+       let n = thing (P.pfName x)
+       return ModVParam { mvpName = n
+                        , mvpType = s
+                        , mvpDoc  = P.pfDoc x
+                        , mvpFixity = P.pfFixity x
+                        }
+
   checkBinds decls (CyclicSCC bs : more) =
      do bs1 <- inferBinds isTopLevel True bs
         foldr (\b m -> withVar (dName b) (dSignature b) m)
