@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.TypeCheck.Monad
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -11,22 +11,27 @@
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE RecursiveDo #-}
 {-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE OverloadedStrings #-}
 module Cryptol.TypeCheck.Monad
   ( module Cryptol.TypeCheck.Monad
   , module Cryptol.TypeCheck.InferTypes
   ) where
 
-import           Cryptol.ModuleSystem.Name (FreshM(..),Supply)
+import           Cryptol.ModuleSystem.Name
+                    (FreshM(..),Supply,mkParameter
+                    , nameInfo, NameInfo(..),NameSource(..))
 import           Cryptol.Parser.Position
 import qualified Cryptol.Parser.AST as P
-import           Cryptol.Prelude (writeTcPreludeContents)
 import           Cryptol.TypeCheck.AST
 import           Cryptol.TypeCheck.Subst
-import           Cryptol.TypeCheck.Unify(mgu, Result(..), UnificationError(..))
+import           Cryptol.TypeCheck.Unify(mgu, runResult, UnificationError(..))
 import           Cryptol.TypeCheck.InferTypes
+import           Cryptol.TypeCheck.Error(Warning(..),Error(..),cleanupErrors)
+import           Cryptol.TypeCheck.PP (brackets, commaSep)
 import qualified Cryptol.TypeCheck.SimpleSolver as Simple
-import qualified Cryptol.TypeCheck.Solver.CrySAT as CrySAT
-import           Cryptol.Utils.PP(pp, (<+>), Doc, text, quotes)
+import qualified Cryptol.TypeCheck.Solver.SMT as SMT
+import           Cryptol.Utils.PP(pp, (<+>), text, quotes)
+import           Cryptol.Utils.Ident(Ident)
 import           Cryptol.Utils.Panic(panic)
 
 import qualified Control.Applicative as A
@@ -35,15 +40,12 @@ import qualified Data.Map as Map
 import qualified Data.Set as Set
 import           Data.Map (Map)
 import           Data.Set (Set)
-import           Data.List(find, minimumBy, groupBy, sortBy, foldl')
-import           Data.Maybe(mapMaybe)
-import           Data.Function(on)
+import           Data.List(find, foldl')
+import           Data.Maybe(mapMaybe,fromMaybe)
 import           MonadLib hiding (mapM)
 
 import           Data.IORef
 
-import           System.FilePath((</>))
-import           System.Directory(doesFileExist)
 
 
 import GHC.Generics (Generic)
@@ -59,6 +61,15 @@ data InferInput = InferInput
   , inpVars      :: Map Name Schema   -- ^ Variables that are in scope
   , inpTSyns     :: Map Name TySyn    -- ^ Type synonyms that are in scope
   , inpNewtypes  :: Map Name Newtype  -- ^ Newtypes in scope
+
+    -- When typechecking a module these start off empty.
+    -- We need them when type-checking an expression at the command
+    -- line, for example.
+  , inpParamTypes       :: !(Map Name ModTParam)  -- ^ Type parameters
+  , inpParamConstraints :: !([Located Prop])      -- ^ Constraints on parameters
+  , inpParamFuns        :: !(Map Name ModVParam)  -- ^ Value parameters
+
+
   , inpNameSeeds :: NameSeeds         -- ^ Private state of type-checker
   , inpMonoBinds :: Bool              -- ^ Should local bindings without
                                       --   signatures be monomorphized?
@@ -67,9 +78,9 @@ data InferInput = InferInput
   , inpSearchPath :: [FilePath]
     -- ^ Where to look for Cryptol theory file.
 
-  , inpPrimNames :: !PrimMap          -- ^ The mapping from 'Ident' to 'Name',
-                                      -- for names that the typechecker
-                                      -- needs to refer to.
+  , inpPrimNames :: !PrimMap
+    -- ^ This is used when the type-checker needs to refer to a predefined
+    -- identifier (e.g., @number@).
 
   , inpSupply :: !Supply              -- ^ The supply for fresh name generation
   } deriving Show
@@ -82,6 +93,7 @@ data NameSeeds = NameSeeds
   } deriving (Show, Generic, NFData)
 
 -- | The initial seeds, used when checking a fresh program.
+-- XXX: why does this start at 10?
 nameSeeds :: NameSeeds
 nameSeeds = NameSeeds { seedTVar = 10, seedGoal = 0 }
 
@@ -102,20 +114,23 @@ bumpCounter = do RO { .. } <- IM ask
                  io $ modifyIORef' iSolveCounter (+1)
 
 runInferM :: TVars a => InferInput -> InferM a -> IO (InferOutput a)
-runInferM info (IM m) = CrySAT.withSolver (inpSolverConfig info) $ \solver ->
-  do loadCryTCPrel solver (inpSearchPath info)
-     coutner <- newIORef 0
+runInferM info (IM m) = SMT.withSolver (inpSolverConfig info) $ \solver ->
+  do coutner <- newIORef 0
      rec ro <- return RO { iRange     = inpRange info
-                     , iVars          = Map.map ExtVar (inpVars info)
-                     , iTVars         = []
-                     , iTSyns         = fmap mkExternal (inpTSyns info)
-                     , iNewtypes      = fmap mkExternal (inpNewtypes info)
-                     , iSolvedHasLazy = iSolvedHas finalRW     -- RECURSION
-                     , iMonoBinds     = inpMonoBinds info
-                     , iSolver        = solver
-                     , iPrimNames     = inpPrimNames info
-                     , iSolveCounter  = coutner
-                     }
+                         , iVars          = Map.map ExtVar (inpVars info)
+                         , iTVars         = []
+                         , iTSyns         = fmap mkExternal (inpTSyns info)
+                         , iNewtypes      = fmap mkExternal (inpNewtypes info)
+                         , iParamTypes    = inpParamTypes info
+                         , iParamFuns     = inpParamFuns info
+                         , iParamConstraints = inpParamConstraints info
+
+                         , iSolvedHasLazy = iSolvedHas finalRW     -- RECURSION
+                         , iMonoBinds     = inpMonoBinds info
+                         , iSolver        = solver
+                         , iPrimNames     = inpPrimNames info
+                         , iSolveCounter  = coutner
+                         }
 
          (result, finalRW) <- runStateT rw
                             $ runReaderT ro m  -- RECURSION
@@ -134,14 +149,13 @@ runInferM info (IM m) = CrySAT.withSolver (inpSolverConfig info) $ \solver ->
                                   (iSupply finalRW)
                                   (apSubst defSu result)
            (cts,has) -> return $ InferFailed warns
-                $ dropErrorsFromSameLoc
+                $ cleanupErrors
                 [ ( goalRange g
                   , UnsolvedGoals False [apSubst theSu g]
                   ) | g <- fromGoals cts ++ map hasGoal has
                 ]
        errs -> return $ InferFailed warns
-                      $ dropErrorsFromSameLoc
-                                  [(r,apSubst theSu e) | (r,e) <- errs]
+                      $ cleanupErrors [(r,apSubst theSu e) | (r,e) <- errs]
 
   where
   mkExternal x = (IsExternal, x)
@@ -159,24 +173,7 @@ runInferM info (IM m) = CrySAT.withSolver (inpSolverConfig info) $ \solver ->
           , iSupply     = inpSupply info
           }
 
-  dropErrorsFromSameLoc = map chooseBestError . groupBy ((==)    `on` fst)
-                                              . sortBy  (cmpRange `on` fst)
 
-  addErrorSize (r,e) = (length (show (pp e)), (r,e))
-  chooseBestError    = snd . minimumBy (compare `on` fst) . map addErrorSize
-
-  -- The actual order does not matter
-  cmpRange (Range x y z) (Range a b c) = compare (x,y,z) (a,b,c)
-
-  loadCryTCPrel s [] =
-    do file <- writeTcPreludeContents
-       CrySAT.loadFile s file
-
-  loadCryTCPrel s (p : ps) =
-    do let file = p </> "CryptolTC.z3"
-       yes <- doesFileExist file
-       if yes then CrySAT.loadFile s file
-              else loadCryTCPrel s ps
 
 
 
@@ -206,6 +203,16 @@ data RO = RO
    -- at the top-level, but then there can't be a newtype with the
    -- same name (this should be caught by the renamer).
 
+  , iParamTypes :: Map Name ModTParam
+    -- ^ Parameter types
+
+  , iParamConstraints :: [Located Prop]
+    -- ^ Constraints on the type parameters
+
+  , iParamFuns :: Map Name ModVParam
+    -- ^ Parameter functions
+
+
   , iSolvedHasLazy :: Map Int (Expr -> Expr)
     -- ^ NOTE: This field is lazy in an important way!  It is the
     -- final version of `iSolvedHas` in `RW`, and the two are tied
@@ -218,7 +225,7 @@ data RO = RO
     -- in where-blocks will never be generalized. Bindings with type
     -- signatures, and all bindings at top level are unaffected.
 
-  , iSolver :: CrySAT.Solver
+  , iSolver :: SMT.Solver
 
   , iPrimNames :: !PrimMap
 
@@ -306,10 +313,20 @@ recordError e =
 
 recordWarning :: Warning -> InferM ()
 recordWarning w =
-  do r <- curRange
+  unless ignore $
+  do r <- case w of
+            DefaultingTo d _ -> return (tvarSource d)
+            _ -> curRange
      IM $ sets_ $ \s -> s { iWarnings = (r,w) : iWarnings s }
+  where
+  ignore
+    | DefaultingTo d _ <- w
+    , Just n <- tvSourceName (tvarDesc d)
+    , Declared _ SystemName <- nameInfo n
+      = True
+    | otherwise = False
 
-getSolver :: InferM CrySAT.Solver
+getSolver :: InferM SMT.Solver
 getSolver =
   do RO { .. } <- IM ask
      return iSolver
@@ -336,7 +353,8 @@ newGoals src ps = addGoals =<< mapM (newGoal src) ps
 The substitution IS applied to them. -}
 getGoals :: InferM [Goal]
 getGoals =
-  do goals <- applySubst =<< IM (sets $ \s -> (iCts s, s { iCts = emptyGoals }))
+  do goals <- applySubst =<<
+                  IM (sets $ \s -> (iCts s, s { iCts = emptyGoals }))
      return (fromGoals goals)
 
 -- | Add a bunch of goals that need solving.
@@ -413,6 +431,11 @@ solveHasGoal n e =
 
 --------------------------------------------------------------------------------
 
+-- | Generate a fresh variable name to be used in a local binding.
+newParamName :: Ident -> InferM Name
+newParamName x =
+  do r <- curRange
+     liftSupply (mkParameter x r)
 
 newName :: (NameSeeds -> (a , NameSeeds)) -> InferM a
 newName upd = IM $ sets $ \s -> let (x,seeds) = upd (iNameSeeds s)
@@ -425,32 +448,39 @@ newGoalName = newName $ \s -> let x = seedGoal s
                               in (x, s { seedGoal = x + 1})
 
 -- | Generate a new free type variable.
-newTVar :: Doc -> Kind -> InferM TVar
+newTVar :: TVarSource -> Kind -> InferM TVar
 newTVar src k = newTVar' src Set.empty k
 
 -- | Generate a new free type variable that depends on these additional
 -- type parameters.
-newTVar' :: Doc -> Set TVar -> Kind -> InferM TVar
+newTVar' :: TVarSource -> Set TParam -> Kind -> InferM TVar
 newTVar' src extraBound k =
-  do bound <- getBoundInScope
+  do r <- curRange
+     bound <- getBoundInScope
      let vs = Set.union extraBound bound
+         msg = TVarInfo { tvarDesc = src, tvarSource = r }
      newName $ \s -> let x = seedTVar s
-                     in (TVFree x k vs src, s { seedTVar = x + 1 })
+                     in (TVFree x k vs msg, s { seedTVar = x + 1 })
 
 
 
 -- | Generate a new free type variable.
-newTParam :: Maybe Name -> Kind -> InferM TParam
-newTParam nm k = newName $ \s -> let x = seedTVar s
-                                 in (TParam { tpUnique = x
-                                            , tpKind   = k
-                                            , tpName   = nm
-                                            }
-                                 , s { seedTVar = x + 1 })
+newTParam :: P.TParam Name -> TPFlavor -> Kind -> InferM TParam
+newTParam nm flav k = newName $ \s ->
+  let x = seedTVar s
+  in (TParam { tpUnique = x
+            , tpKind   = k
+            , tpFlav   = flav
+            , tpInfo   = desc
+            }
+     , s { seedTVar = x + 1 })
+  where desc = TVarInfo { tvarDesc = TVFromSignature (P.tpName nm)
+                        , tvarSource = fromMaybe emptyRange (P.tpRange nm)
+                        }
 
 
 -- | Generate an unknown type.  The doc is a note about what is this type about.
-newType :: Doc -> Kind -> InferM Type
+newType :: TVarSource -> Kind -> InferM Type
 newType src k = TVar `fmap` newTVar src k
 
 
@@ -463,21 +493,21 @@ unify :: Type -> Type -> InferM [Prop]
 unify t1 t2 =
   do t1' <- applySubst t1
      t2' <- applySubst t2
-     case mgu t1' t2' of
-       OK (su1,ps) -> extendSubst su1 >> return ps
-       Error err   ->
-         do case err of
-              UniTypeLenMismatch _ _ -> recordError (TypeMismatch t1' t2')
-              UniTypeMismatch s1 s2 -> recordError (TypeMismatch s1 s2)
-              UniKindMismatch k1 k2 -> recordError (KindMismatch k1 k2)
-              UniRecursive x t -> recordError (RecursiveType (TVar x) t)
-              UniNonPolyDepends x vs -> recordError
-                                          (TypeVariableEscaped (TVar x) vs)
-              UniNonPoly x t -> recordError (NotForAll x t)
-            return []
-
-
-
+     let ((su1, ps), errs) = runResult (mgu t1' t2')
+     extendSubst su1
+     let toError :: UnificationError -> Error
+         toError err =
+           case err of
+             UniTypeLenMismatch _ _ -> TypeMismatch t1' t2'
+             UniTypeMismatch s1 s2  -> TypeMismatch s1 s2
+             UniKindMismatch k1 k2  -> KindMismatch k1 k2
+             UniRecursive x t       -> RecursiveType (TVar x) t
+             UniNonPolyDepends x vs -> TypeVariableEscaped (TVar x) vs
+             UniNonPoly x t         -> NotForAll x t
+     case errs of
+       [] -> return ps
+       _  -> do mapM_ (recordError . toError) errs
+                return []
 
 -- | Apply the accumulated substitution to something with free type variables.
 applySubst :: TVars t => t -> InferM t
@@ -485,13 +515,52 @@ applySubst t =
   do su <- getSubst
      return (apSubst su t)
 
+applySubstPreds :: [Prop] -> InferM [Prop]
+applySubstPreds ps =
+  do ps1 <- applySubst ps
+     return (concatMap pSplitAnd ps1)
+
+
+applySubstGoals :: [Goal] -> InferM [Goal]
+applySubstGoals gs =
+  do gs1 <- applySubst gs
+     return [ g { goal = p } | g <- gs1, p <- pSplitAnd (goal g) ]
+
 -- | Get the substitution that we have accumulated so far.
 getSubst :: InferM Subst
 getSubst = IM $ fmap iSubst get
 
--- | Add to the accumulated substitution.
+-- | Add to the accumulated substitution, checking that the datatype
+-- invariant for `Subst` is maintained.
 extendSubst :: Subst -> InferM ()
-extendSubst su = IM $ sets_ $ \s -> s { iSubst = su @@ iSubst s }
+extendSubst su =
+  do mapM_ check (substToList su)
+     IM $ sets_ $ \s -> s { iSubst = su @@ iSubst s }
+  where
+    check :: (TVar, Type) -> InferM ()
+    check (v, ty) =
+      case v of
+        TVBound _ ->
+          panic "Cryptol.TypeCheck.Monad.extendSubst"
+            [ "Substitution instantiates bound variable:"
+            , "Variable: " ++ show (pp v)
+            , "Type:     " ++ show (pp ty)
+            ]
+        TVFree _ _ tvs _ ->
+          do let bounds tv =
+                   case tv of
+                     TVBound tp -> Set.singleton tp
+                     TVFree _ _ tps _ -> tps
+             let vars = Set.unions (map bounds (Set.elems (fvs ty)))
+                 -- (Set.filter isBoundTV (fvs ty))
+             let escaped = Set.difference vars tvs
+             if Set.null escaped then return () else
+               panic "Cryptol.TypeCheck.Monad.extendSubst"
+                 [ "Escaped quantified variables:"
+                 , "Substitution:  " ++ show (pp v <+> text ":=" <+> pp ty)
+                 , "Vars in scope: " ++ show (brackets (commaSep (map pp (Set.toList tvs))))
+                 , "Escaped:       " ++ show (brackets (commaSep (map pp (Set.toList escaped))))
+                 ]
 
 
 -- | Variables that are either mentioned in the environment or in
@@ -524,14 +593,17 @@ lookupVar x =
          do mbNT <- lookupNewtype x
             case mbNT of
               Just nt -> return (ExtVar (newtypeConType nt))
-              Nothing -> do recordError $ UndefinedVariable x
-                            a <- newType (text "type of" <+> pp x) KType
-                            return $ ExtVar $ Forall [] [] a
+              Nothing ->
+                do mbParamFun <- lookupParamFun x
+                   case mbParamFun of
+                     Just pf -> return (ExtVar (mvpType pf))
+                     Nothing -> panic "lookupVar" [ "Undefined type variable"
+                                                  , show x]
 
 -- | Lookup a type variable.  Return `Nothing` if there is no such variable
 -- in scope, in which case we must be dealing with a type constant.
-lookupTVar :: Name -> InferM (Maybe Type)
-lookupTVar x = IM $ asks $ fmap (TVar . tpVar) . find this . iTVars
+lookupTParam :: Name -> InferM (Maybe TParam)
+lookupTParam x = IM $ asks $ find this . iTVars
   where this tp = tpName tp == Just x
 
 -- | Lookup the definition of a type synonym.
@@ -542,6 +614,14 @@ lookupTSyn x = fmap (fmap snd . Map.lookup x) getTSyns
 lookupNewtype :: Name -> InferM (Maybe Newtype)
 lookupNewtype x = fmap (fmap snd . Map.lookup x) getNewtypes
 
+-- | Lookup the kind of a parameter type
+lookupParamType :: Name -> InferM (Maybe ModTParam)
+lookupParamType x = Map.lookup x <$> getParamTypes
+
+-- | Lookup the schema for a parameter function.
+lookupParamFun :: Name -> InferM (Maybe ModVParam)
+lookupParamFun x = Map.lookup x <$> getParamFuns
+
 -- | Check if we already have a name for this existential type variable and,
 -- if so, return the definition.  If not, try to create a new definition,
 -- if this is allowed.  If not, returns nothing.
@@ -554,14 +634,12 @@ existVar x k =
        Nothing ->
          case scopes of
            [] ->
-              do recordError $ ErrorMsg $
-                    text "Undefined type" <+> quotes (pp x)
-                 newType (text "undefined existential type varible" <+>
-                                                            quotes (pp x)) k
+              do recordError $ ErrorMsg
+                             $ text "Undefined type" <+> quotes (pp x)
+                 newType TypeErrorPlaceHolder k
 
            sc : more ->
-             do ty <- newType (text "existential type variable"
-                                                        <+> quotes (pp x)) k
+             do ty <- newType TypeErrorPlaceHolder k
                 IM $ sets_ $ \s -> s{ iExistTVars = Map.insert x ty sc : more }
                 return ty
 
@@ -574,13 +652,29 @@ getTSyns = IM $ asks iTSyns
 getNewtypes :: InferM (Map Name (DefLoc,Newtype))
 getNewtypes = IM $ asks iNewtypes
 
+-- | Returns the parameter functions declarations
+getParamFuns :: InferM (Map Name ModVParam)
+getParamFuns = IM $ asks iParamFuns
+
+-- | Returns the abstract function declarations
+getParamTypes :: InferM (Map Name ModTParam)
+getParamTypes = IM $ asks iParamTypes
+
+-- | Constraints on the module's parameters.
+getParamConstraints :: InferM [Located Prop]
+getParamConstraints = IM $ asks iParamConstraints
+
 -- | Get the set of bound type variables that are in scope.
 getTVars :: InferM (Set Name)
 getTVars = IM $ asks $ Set.fromList . mapMaybe tpName . iTVars
 
 -- | Return the keys of the bound variables that are in scope.
-getBoundInScope :: InferM (Set TVar)
-getBoundInScope = IM $ asks $ Set.fromList . map tpVar . iTVars
+getBoundInScope :: InferM (Set TParam)
+getBoundInScope =
+  do ro <- IM ask
+     let params = Set.fromList (map mtpParam (Map.elems (iParamTypes ro)))
+         bound  = Set.fromList (iTVars ro)
+     return $! Set.union params bound
 
 -- | Retrieve the value of the `mono-binds` option.
 getMonoBinds :: InferM Bool
@@ -638,6 +732,11 @@ withNewtype t (IM m) =
   IM $ mapReader
         (\r -> r { iNewtypes = Map.insert (ntName t) (IsLocal,t)
                                                      (iNewtypes r) }) m
+withParamType :: ModTParam -> InferM a -> InferM a
+withParamType a (IM m) =
+  IM $ mapReader
+        (\r -> r { iParamTypes = Map.insert (mtpName a) a (iParamTypes r) })
+        m
 
 -- | The sub-computation is performed with the given variable in scope.
 withVarType :: Name -> VarType -> InferM a -> InferM a
@@ -650,6 +749,18 @@ withVarTypes xs m = foldr (uncurry withVarType) m xs
 withVar :: Name -> Schema -> InferM a -> InferM a
 withVar x s = withVarType x (ExtVar s)
 
+-- | The sub-computation is performed with the given abstract function in scope.
+withParamFuns :: [ModVParam] -> InferM a -> InferM a
+withParamFuns xs (IM m) =
+  IM $ mapReader (\r -> r { iParamFuns = foldr add (iParamFuns r) xs }) m
+  where
+  add x = Map.insert (mvpName x) x
+
+-- | Add some assumptions for an entire module
+withParameterConstraints :: [Located Prop] -> InferM a -> InferM a
+withParameterConstraints ps (IM m) =
+  IM $ mapReader (\r -> r { iParamConstraints = ps ++ iParamConstraints r }) m
+
 
 -- | The sub-computation is performed with the given variables in scope.
 withMonoType :: (Name,Located Type) -> InferM a -> InferM a
@@ -684,10 +795,13 @@ inNewScope m =
 
 newtype KindM a = KM { unKM :: ReaderT KRO (StateT KRW InferM)  a }
 
-data KRO = KRO { lazyTVars  :: Map Name Type -- ^ lazy map, with tyvars.
-               , allowWild  :: Bool           -- ^ are type-wild cards allowed?
+data KRO = KRO { lazyTParams :: Map Name TParam -- ^ lazy map, with tparams.
+               , allowWild   :: AllowWildCards  -- ^ are type-wild cards allowed?
                }
 
+-- | Do we allow wild cards in the given context.
+data AllowWildCards = AllowWildCards | NoWildCards
+
 data KRW = KRW { typeParams :: Map Name Kind -- ^ kinds of (known) vars.
                , kCtrs      :: [(ConstraintSource,[Prop])]
                }
@@ -704,45 +818,48 @@ instance Monad KindM where
   fail x        = KM (fail x)
   KM m >>= k    = KM (m >>= unKM . k)
 
+
+
+
 {- | The arguments to this function are as follows:
 
-(type param. name, kind signature (opt.), a type representing the param)
+(type param. name, kind signature (opt.), type parameter)
 
-The type representing the parameter is just a thunk that we should not force.
-The reason is that the type depnds on the kind of parameter, that we are
+The type parameter is just a thunk that we should not force.
+The reason is that the parameter depends on the kind that we are
 in the process of computing.
 
 As a result we return the value of the sub-computation and the computed
 kinds of the type parameters. -}
-runKindM :: Bool                          -- Are type-wild cards allowed?
-         -> [(Name, Maybe Kind, Type)]   -- ^ See comment
+runKindM :: AllowWildCards               -- Are type-wild cards allowed?
+         -> [(Name, Maybe Kind, TParam)] -- ^ See comment
          -> KindM a -> InferM (a, Map Name Kind, [(ConstraintSource,[Prop])])
 runKindM wildOK vs (KM m) =
   do (a,kw) <- runStateT krw (runReaderT kro m)
      return (a, typeParams kw, kCtrs kw)
   where
-  tys  = Map.fromList [ (x,t) | (x,_,t)      <- vs ]
-  kro  = KRO { allowWild = wildOK, lazyTVars = tys }
+  tps  = Map.fromList [ (x,t) | (x,_,t)      <- vs ]
+  kro  = KRO { allowWild = wildOK, lazyTParams = tps }
   krw  = KRW { typeParams = Map.fromList [ (x,k) | (x,Just k,_) <- vs ]
              , kCtrs = []
              }
 
 -- | This is what's returned when we lookup variables during kind checking.
-data LkpTyVar = TLocalVar Type (Maybe Kind) -- ^ Locally bound variable.
-              | TOuterVar Type              -- ^ An outer binding.
+data LkpTyVar = TLocalVar TParam (Maybe Kind) -- ^ Locally bound variable.
+              | TOuterVar TParam              -- ^ An outer binding.
 
 -- | Check if a name refers to a type variable.
 kLookupTyVar :: Name -> KindM (Maybe LkpTyVar)
 kLookupTyVar x = KM $
-  do vs <- lazyTVars `fmap` ask
+  do vs <- lazyTParams `fmap` ask
      ss <- get
      case Map.lookup x vs of
        Just t  -> return $ Just $ TLocalVar t $ Map.lookup x $ typeParams ss
-       Nothing -> lift $ lift $ do t <- lookupTVar x
+       Nothing -> lift $ lift $ do t <- lookupTParam x
                                    return (fmap TOuterVar t)
 
 -- | Are type wild-cards OK in this context?
-kWildOK :: KindM Bool
+kWildOK :: KindM AllowWildCards
 kWildOK = KM $ fmap allowWild ask
 
 -- | Reports an error.
@@ -756,10 +873,10 @@ kRecordWarning w = kInInferM $ recordWarning w
 -- NOTE:  We do not simplify these, because we end up with bottom.
 -- See `Kind.hs`
 -- XXX: Perhaps we can avoid the recursion?
-kNewType :: Doc -> Kind -> KindM Type
+kNewType :: TVarSource -> Kind -> KindM Type
 kNewType src k =
-  do tps <- KM $ do vs <- asks lazyTVars
-                    return $ Set.fromList [ tv | TVar tv <- Map.elems vs ]
+  do tps <- KM $ do vs <- asks lazyTParams
+                    return $ Set.fromList (Map.elems vs)
      kInInferM $ TVar `fmap` newTVar' src tps k
 
 -- | Lookup the definition of a type synonym.
@@ -770,6 +887,9 @@ kLookupTSyn x = kInInferM $ lookupTSyn x
 kLookupNewtype :: Name -> KindM (Maybe Newtype)
 kLookupNewtype x = kInInferM $ lookupNewtype x
 
+kLookupParamType :: Name -> KindM (Maybe ModTParam)
+kLookupParamType x = kInInferM (lookupParamType x)
+
 kExistTVar :: Name -> Kind -> KindM Type
 kExistTVar x k = kInInferM $ existVar x k
 
