@@ -36,4 +36,4 @@ instance PP Solved where
     case res of
       SolvedIf ps  -> text "solved" $$ nest 2 (vcat (map pp ps))
       Unsolved     -> text "unsolved"
-      Unsolvable e -> text "unsolvable" <> colon <+> text (tcErrorMessage e)
+      Unsolvable e -> text "unsolvable" <.> colon <+> text (tcErrorMessage e)
