@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.TypeCheck.Unify
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -13,8 +13,8 @@ module Cryptol.TypeCheck.Unify where
 
 import Cryptol.TypeCheck.AST
 import Cryptol.TypeCheck.Subst
-import Cryptol.Utils.Panic (panic)
 
+import Control.Monad.Writer (Writer, writer, runWriter)
 import Data.Ord(comparing)
 import Data.List(sortBy)
 import qualified Data.Set as Set
@@ -26,34 +26,21 @@ import Prelude.Compat
 -- on bound variables.
 type MGU = (Subst,[Prop])
 
-data Result a = OK a | Error UnificationError
-                deriving (Functor)
+type Result a = Writer [UnificationError] a
+
+runResult :: Result a -> (a, [UnificationError])
+runResult = runWriter
 
 data UnificationError
   = UniTypeMismatch Type Type
   | UniKindMismatch Kind Kind
   | UniTypeLenMismatch Int Int
   | UniRecursive TVar Type
-  | UniNonPolyDepends TVar [TVar]
+  | UniNonPolyDepends TVar [TParam]
   | UniNonPoly TVar Type
 
-instance Applicative Result where
-  pure = OK
-
-  OK f     <*> OK x    = OK (f x)
-  OK _     <*> Error e = Error e
-  Error e  <*> _       = Error e
-
-instance Monad Result where
-  return a      = OK a
-
-  OK a >>= k    = k a
-  Error x >>= _ = Error x
-
-  fail x        = panic "Cryptol.TypeCheck.Unify.fail" [x]
-
-uniError :: UnificationError -> Result a
-uniError e = Error e
+uniError :: UnificationError -> Result MGU
+uniError e = writer (emptyMGU, [e])
 
 
 emptyMGU :: MGU
@@ -116,20 +103,29 @@ bindVar x (tNoUser -> TVar y)
 
 bindVar v@(TVBound {}) (tNoUser -> TVar v1@(TVFree {})) = bindVar v1 (TVar v)
 
-bindVar v@(TVBound _ k) t
+bindVar v@(TVBound {}) t
   | k == kindOf t = if k == KNum
                        then return (emptySubst, [TVar v =#= t])
                        else uniError $ UniNonPoly v t
   | otherwise     = uniError $ UniKindMismatch k (kindOf t)
+  where k = kindOf v
 
+bindVar x@(TVFree _ _ xscope _) (TVar y@(TVFree _ _ yscope _))
+  | xscope `Set.isProperSubsetOf` yscope = return (singleSubst y (TVar x), [])
 
 bindVar x@(TVFree _ k inScope _d) t
   | not (k == kindOf t)     = uniError $ UniKindMismatch k (kindOf t)
   | recTy && k == KType     = uniError $ UniRecursive x t
-  | not (Set.null escaped)  = uniError $ UniNonPolyDepends x$ Set.toList escaped
+  | not (Set.null escaped)  = uniError $ UniNonPolyDepends x $ Set.toList escaped
   | recTy                   = return (emptySubst, [TVar x =#= t])
   | otherwise               = return (singleSubst x t, [])
     where
-    vs      = fvs t
-    escaped = Set.filter isBoundTV vs `Set.difference` inScope
-    recTy   = x `Set.member` vs
+    escaped = freeParams t `Set.difference` inScope
+    recTy   = x `Set.member` fvs t
+
+
+freeParams :: FVS t => t -> Set.Set TParam
+freeParams x = Set.unions (map params (Set.toList (fvs x)))
+  where
+    params (TVFree _ _ tps _) = tps
+    params (TVBound tp) = Set.singleton tp
