@@ -1,14 +1,42 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Utils.Ident
 -- Copyright   :  (c) 2015-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
 -- Stability   :  provisional
 -- Portability :  portable
 
-{-# LANGUAGE DeriveGeneric #-}
+{-# LANGUAGE DeriveGeneric, OverloadedStrings #-}
 
-module Cryptol.Utils.Ident where
+module Cryptol.Utils.Ident
+  ( -- * Module names
+    ModName
+  , modNameToText
+  , textToModName
+  , modNameChunks
+  , packModName
+  , preludeName
+  , interactiveName
+  , noModuleName
+  , exprModName
+
+  , isParamInstModName
+  , paramInstModName
+  , notParamInstModName
+
+    -- * Identifiers
+  , Ident
+  , packIdent
+  , packInfix
+  , unpackIdent
+  , mkIdent
+  , mkInfix
+  , isInfixIdent
+  , nullIdent
+  , identText
+  , modParamIdent
+
+  ) where
 
 import           Control.DeepSeq (NFData)
 import           Data.Char (isSpace)
@@ -19,24 +47,70 @@ import           GHC.Generics (Generic)
 
 
 -- | Module names are just text.
-type ModName = T.Text
+data ModName = ModName T.Text
+  deriving (Eq,Ord,Show,Generic)
 
-unpackModName :: ModName -> [String]
-unpackModName  = unfoldr step
+instance NFData ModName
+
+modNameToText :: ModName -> T.Text
+modNameToText (ModName x) = x
+
+textToModName :: T.Text -> ModName
+textToModName = ModName
+
+modNameChunks :: ModName -> [String]
+modNameChunks  = unfoldr step . modNameToText . notParamInstModName
   where
   step str
     | T.null str = Nothing
     | otherwise  = case T.breakOn modSep str of
                      (a,b) -> Just (T.unpack a,T.drop (T.length modSep) b)
 
-packModName :: [String] -> ModName
-packModName strs = T.intercalate modSep (map (trim . T.pack) strs)
+isParamInstModName :: ModName -> Bool
+isParamInstModName (ModName x) = modInstPref `T.isPrefixOf` x
+
+-- | Convert a parameterized module's name to the name of the module
+-- containing the same definitions but with explicit parameters on each
+-- definition.
+paramInstModName :: ModName -> ModName
+paramInstModName (ModName x)
+  | modInstPref `T.isPrefixOf` x = ModName x
+  | otherwise = ModName (T.append modInstPref x)
+
+
+notParamInstModName :: ModName -> ModName
+notParamInstModName (ModName x)
+  | modInstPref `T.isPrefixOf` x = ModName (T.drop (T.length modInstPref) x)
+  | otherwise = ModName x
+
+
+packModName :: [T.Text] -> ModName
+packModName strs = textToModName (T.intercalate modSep (map trim strs))
   where
   -- trim space off of the start and end of the string
   trim str = T.dropWhile isSpace (T.dropWhileEnd isSpace str)
 
 modSep :: T.Text
-modSep  = T.pack "::"
+modSep  = "::"
+
+modInstPref :: T.Text
+modInstPref = "`"
+
+
+preludeName :: ModName
+preludeName  = packModName ["Cryptol"]
+
+interactiveName :: ModName
+interactiveName  = packModName ["<interactive>"]
+
+noModuleName :: ModName
+noModuleName = packModName ["<none>"]
+
+exprModName :: ModName
+exprModName = packModName ["<expr>"]
+
+
+--------------------------------------------------------------------------------
 
 -- | Identifiers, along with a flag that indicates whether or not they're infix
 -- operators. The boolean is present just as cached information from the lexer,
@@ -80,14 +154,8 @@ nullIdent (Ident _ t) = T.null t
 identText :: Ident -> T.Text
 identText (Ident _ t) = t
 
+modParamIdent :: Ident -> Ident
+modParamIdent (Ident x t) = Ident x (T.append (T.pack "module parameter ") t)
 
--- Frequently Used Names -------------------------------------------------------
 
-preludeName :: ModName
-preludeName  = packModName ["Cryptol"]
 
-preludeExtrasName :: ModName
-preludeExtrasName = packModName ["Cryptol", "Extras"]
-
-interactiveName :: ModName
-interactiveName  = packModName ["<interactive>"]
